(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory MatrixCarrier imports Matrix Rational
begin

definition rat_gt :: "rat \<Rightarrow> rat \<Rightarrow> rat \<Rightarrow> bool" where 
  "rat_gt \<delta> x y \<equiv> x - y \<ge> (if \<delta> > 0 then min \<delta> 1 else 1)"


lemma rat_gt_SN:  "SN {(x,y). 0 \<le> y \<and> rat_gt \<delta> x y}"
  proof (rule ccontr, unfold SN_defs, clarify)
  fix x f
  let ?R = "{(x,y). 0 \<le> y \<and> rat_gt \<delta> x y}"
  let ?S = "\<lambda> d. {(x,y). (0 :: rat) \<le> y \<and> x - y \<ge> d}" 
  assume Rsteps: "\<forall> i. (f i, f (Suc i)) \<in> ?R"
  have "\<exists> d. 0 < d \<and> ?R = ?S d"
    by (cases "\<delta> > 0", simp add: rat_gt_def, rule exI[of _ "min \<delta> 1"], simp, simp add: rat_gt_def, rule exI[of _ 1], simp)
  then obtain d :: rat where d: "d > 0" and R: "?R = ?S d" by auto
  let ?Rd = "?S d"
  from Rsteps R have steps:  "\<forall> i. (f i, f (Suc i)) \<in> ?Rd" by auto
  from spec[OF steps, of 0] d have fz_0: "f 0 > 0" by auto
  {
    fix n :: nat
    have "f n \<le> f 0 - (of_nat n) * d"
    proof (induct n, simp)
      case (Suc n)
      with spec[OF steps, of n] show ?case by (auto simp: field_simps)
    qed
  } note main1 = this
  { 
    fix n :: nat
    have "f 0 \<ge> of_nat (Suc n) * d" using main1[of "Suc n"] spec[OF steps, of "n"] by auto
  } note main2 = this
  obtain numeratord denomd where d_id: "d = Fract numeratord denomd" by (cases "d", auto)
  with d have denomd: "denomd \<noteq> 0" by (cases "denomd = 0", simp_all add: rat_number_collapse)
  obtain numerator0 denom0 where fz: "f 0 = Fract numerator0 denom0" by (cases "f 0", auto)
  with fz_0 have denom0: "denom0 \<noteq> 0" by (cases "denom0 = 0", simp_all add: rat_number_collapse)
  from fz d_id main2 have contra: "\<And> n. Fract numerator0 denom0 \<ge> Fract (int (Suc n) * numeratord) denomd" by (simp add: of_nat_rat)
  from denom0 have square: "denom0 * denom0 > 0" by  (cases "denom0 > 0", simp add: Ring_and_Field.ordered_semiring_strict_class.mult_pos_pos, simp add: Ring_and_Field.ordered_ring_strict_class.mult_neg_neg)
  from d d_id denomd have pos: "numeratord * denomd  > 0" 
    by (auto simp: sgn_times mult_less_0_iff zero_less_mult_iff less_rat' zero_less_Fract_iff, cases "numeratord = 0", auto simp: rat_number_collapse, cases "denomd < 0", simp_all,
      simp only: minus_rat_cancel[symmetric, of numeratord denomd],
      simp only: zero_less_Fract_iff[of "- denomd" "- numeratord"])
  let ?p = "numeratord * denom0 * denomd * denom0"
  have id: "?p = (denom0 * denom0) * (numeratord * denomd)" by auto
  have posp: "?p > 0" using Ring_and_Field.ordered_semiring_strict_class.mult_pos_pos[OF square pos] by (simp add: id)
  have "\<exists> p . ?p = p + 1 \<and> p \<ge> 0"
    by (rule exI[of _ "?p - 1"], simp add: posp)
  from this obtain p where p: "?p = p + 1" and ppos: "0 \<le> p" by auto
  have id2: "\<And> n. n * numeratord * denom0 * (denomd * denom0) = n * ?p" by auto
  from contra square denomd denom0 show False 
  proof (simp add: p id2, simp add: right_distrib left_distrib)
    let ?r = "numerator0 * denomd * (denomd * denom0)"
    let ?rr = "?r - p - 1"
    assume ass: "\<And> n. p + int n * p + (1 + int n) \<le> ?r"
    have nppos: "\<And> n. int n * p \<ge> 0" by  (rule Ring_and_Field.pordered_cancel_semiring_class.mult_nonneg_nonneg[OF _ ppos], simp)
    {
      fix n
      have "int n = 0 + int n" by auto
      also have "\<dots> \<le> int n * p + int n" using nppos[of n] by auto
      finally have "p + 1 + int n \<le> p + 1 + int n * p + int n" by auto
      with ass[of n] have "int n \<le> ?rr" by auto
    } note contra2 = this
    show False 
    proof (cases "?rr \<ge> 0")
      case False
      with contra2[of 0] show False by auto
    next
      case True
      from zero_le_imp_eq_int[OF True] obtain n where id2: "?rr = int n" by auto
      from contra2[of "Suc n"] show False by (simp only: id2)
    qed
  qed
qed


instantiation nat :: ordered_semiring_1
begin
fun ge_nat :: "nat \<Rightarrow> nat \<Rightarrow> bool" where "ge_nat x y = (y \<le> x)"
instance by (intro_classes, auto)
end

instantiation nat :: max_ordered_monoid_add
begin
fun max0_nat :: "nat \<Rightarrow> nat" where "max0_nat x = x"
instance by (intro_classes, auto)
end

definition nat_mono :: "nat \<Rightarrow> bool" where "nat_mono x \<equiv> x \<noteq> 0"
definition int_mono :: "int \<Rightarrow> bool" where "int_mono x \<equiv> x \<ge> 1"
definition rat_mono :: "rat \<Rightarrow> bool" where "rat_mono x \<equiv> x \<ge> 1"

interpretation nat_SN : SN_strict_mono_ordered_semiring_1 "op > :: nat \<Rightarrow> nat \<Rightarrow> bool" nat_mono
proof (unfold_locales, auto simp: nat_mono_def)
  show "SN {(x,y). (y :: nat) < x}" (is "SN ?gt")
  proof (rule ccontr, unfold SN_defs, clarify)
    fix x f
    assume steps: "\<forall> i. (f i, f (Suc i)) \<in> ?gt"
    have "\<forall> i. f i + i \<le> f 0"
    proof 
      fix i
      show "f i + i \<le> f 0"
      proof (induct i, simp)
	case (Suc i)
	with spec[OF steps, of i] show ?case by auto
      qed
    qed
    hence "f (Suc (f 0)) + Suc (f 0) \<le> f 0" by blast
    thus False by auto
  qed
qed

instantiation int :: ordered_semiring_1
begin
fun ge_int :: "int \<Rightarrow> int \<Rightarrow> bool" where "ge_int x y = (y \<le> x)"
instance by (intro_classes, auto simp: mult_right_mono mult_left_mono)
end

instantiation int :: max_ordered_monoid_add
begin
fun max0_int :: "int \<Rightarrow> int" where "max0_int x = max 0 x"
instance by (intro_classes, auto)
end

interpretation int_SN : SN_strict_mono_ordered_semiring_1 "op > :: int \<Rightarrow> int \<Rightarrow> bool" int_mono
proof (unfold_locales, auto simp: Ring_and_Field.mult_strict_left_mono int_mono_def)
  show "SN {(x,y). 0 \<le> y \<and> (y :: int) < x}" (is "SN ?gt")
  proof (rule ccontr, unfold SN_defs, clarify)
    fix x f
    assume steps: "\<forall> i. (f i, f (Suc i)) \<in> ?gt"
    have main: "\<forall> i. f i + int i \<le> f 0"
    proof 
      fix i
      show "f i + int i \<le> f 0"
      proof (induct i, simp)
	case (Suc i)
	with spec[OF steps, of i] show ?case by auto
      qed
    qed
    have contra: "\<forall> i. int (Suc i) \<le> f 0" 
    proof
      fix i
      from spec[OF main, of "Suc i"] spec[OF steps, of i] 
      show "int (Suc i) \<le> f 0" by auto
    qed
    show False 
    proof (cases "0 \<le> f 0")
      case False
      with spec[OF contra, of 0] show False by auto
    next
      case True
      obtain n where "f 0 = int n" using zero_le_imp_eq_int[OF True] ..
      with spec[OF contra, of n] show False by auto
    qed
  qed
qed


instantiation rat :: ordered_semiring_1
begin
fun ge_rat :: "rat \<Rightarrow> rat \<Rightarrow> bool" where "ge_rat x y = (y \<le> x)"
instance by (intro_classes, auto simp: mult_right_mono mult_left_mono)
end

instantiation rat :: max_ordered_monoid_add
begin
fun max0_rat :: "rat \<Rightarrow> rat" where "max0_rat x = max 0 x"
instance by (intro_classes, auto)
end

interpretation rat_SN : SN_strict_mono_ordered_semiring_1 "rat_gt \<delta>" rat_mono
proof (unfold_locales, (auto simp: rat_gt_def)[4], simp, rule rat_gt_SN, simp_all add: rat_mono_def)
  fix x y z :: rat
  assume x: "1 \<le> x" and yz: "rat_gt \<delta> y z"
  have "\<exists> d > 0. rat_gt \<delta> = (\<lambda> x y. d \<le> x - y)" 
  proof (cases "\<delta> > 0")
    case True
    show ?thesis 
      by (rule exI[of _ "min \<delta> 1"], simp add: True, intro ext, auto simp: rat_gt_def True)
  next
    case False
    show ?thesis 
      by (rule exI[of _ "1"], simp, intro ext, auto simp: rat_gt_def False)
  qed
  from this obtain d where d: "0 < d" and rat: "rat_gt \<delta> = (\<lambda> x y. d \<le> x - y)" by auto
  from yz have yzd: "d \<le> y - z" by (simp add: rat)
  show "rat_gt \<delta> (x * y) (x * z)"
  proof (simp only: rat)
    let ?p = "(x - 1) * (y - z)"
    from x have x1: "0 \<le> x - 1" by auto
    from yzd d have yz0: "0 \<le> y - z" by auto
    have "0 \<le> ?p" 
      by (rule Ring_and_Field.pordered_cancel_semiring_class.mult_nonneg_nonneg[OF x1 yz0])
    have "x * y - x * z = x * (y - z)" using Ring_and_Field.ring_class.right_diff_distrib[of x y z] by auto
    also have "\<dots> = ((x - 1) + 1) * (y - z)" by auto
    also have "\<dots> = ?p + 1 * ( y - z)" by (rule Ring_and_Field.ring_class.ring_distribs(2))
    also have "\<dots> = ?p + (y - z)" by simp
    also have "greater_eq \<dots> (0 + d)" using yzd `0 \<le> ?p` by auto
    finally 
    show "d \<le> x * y - x * z" by auto
  qed
qed



section {* the artic semiring *}
text {* plus is interpreted as max, times is interpreted as plus, 0 is -infinity, 1 is 0 *}
datatype arctic = MinInfty | Num_arc int


instantiation arctic :: ordered_semiring_1
begin
fun plus_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> arctic"
where "plus_arctic MinInfty y = y"
   |  "plus_arctic x MinInfty = x"
   |  "plus_arctic (Num_arc x) (Num_arc y) = (Num_arc (max x y))"
fun times_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> arctic"
where "times_arctic MinInfty y = MinInfty"
   |  "times_arctic x MinInfty = MinInfty"
   |  "times_arctic (Num_arc x) (Num_arc y) = (Num_arc (x + y))"
fun ge_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> bool"
where "ge_arctic x MinInfty = True"
   |  "ge_arctic MinInfty (Num_arc _) = False"
   |  "ge_arctic (Num_arc x) (Num_arc y) = (y \<le> x)"
definition zero_arctic :: arctic
where "zero_arctic = MinInfty"
definition one_arctic :: arctic
where "one_arctic = Num_arc 0"
instance
proof
  fix x y :: arctic
  show "x + y = y + x"
    by (cases x, cases y, auto, cases y, auto)
next
  fix x y z :: arctic
  show "(x + y) + z = x + (y + z)"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x y z :: arctic
  show "(x * y) * z = x * (y * z)"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x :: arctic
  show "x * 0 = 0"
    by (cases x, auto simp: zero_arctic_def)
next
  fix x y z :: arctic
  show "x * (y + z) = x * y + x * z"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x y z :: arctic
  show "(x + y) * z = x * z + y * z"
    by (cases x, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z :: arctic
  assume "x \<ge> y" 
  thus "x + z \<ge> y + z"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z :: arctic
  assume "x \<ge> y" 
  thus "x * z \<ge> y * z"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z :: arctic
  assume "y \<ge> z"
  thus "x * y \<ge> x * z"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x :: arctic
  show "x \<ge> x" 
    by (cases x, auto)
next
  fix x y z :: arctic
  assume "x \<ge> y" and "y \<ge> z"
  thus "x \<ge> z"
    by (cases x, cases y, auto, cases y, cases z, auto, cases z, auto)
next
  fix x :: arctic
  show "0 + x = x"
    by (simp add: zero_arctic_def)
next
  fix x :: arctic
  show "0 * x = 0"
    by (simp add: zero_arctic_def)
next
  fix x :: arctic
  show "1 * x = x"
    by (cases x, simp_all add: one_arctic_def)
next
  fix x :: arctic
  show "x * 1 = x"
    by (cases x, simp_all add: one_arctic_def)
next
  show "(0 :: arctic) \<noteq> 1"
    by (simp add: zero_arctic_def one_arctic_def)
next
  show "(1 :: arctic) \<ge> 0"
    by (simp add: zero_arctic_def one_arctic_def)
qed
end


text {* x > y is interpreted as y = -inf or (x,y != -inf and x > y on the integers) *}
fun gt_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> bool"
where "gt_arctic _ MinInfty = True"
   |  "gt_arctic MinInfty (Num_arc _) = False"
   |  "gt_arctic (Num_arc x) (Num_arc y) = (y < x)"


fun get_arctic_num :: "arctic \<Rightarrow> int"
where "get_arctic_num (Num_arc n) = n"

fun pos_arctic :: "arctic \<Rightarrow> bool"
where "pos_arctic MinInfty = False"
    | "pos_arctic (Num_arc n) = (0 <= n)"

interpretation arctic_SN : SN_both_mono_ordered_semiring_1 gt_arctic pos_arctic
proof 
  fix x y z :: arctic
  assume "x \<ge> y" and "gt_arctic y z"
  thus "gt_arctic x z"
    by (cases z, simp, cases y, simp, cases x, auto)
next 
  fix x y z :: arctic
  assume "gt_arctic x y" and "y \<ge> z"
  thus "gt_arctic x z"
    by (cases z, simp, cases y, simp, cases x, auto)
next
  fix x y z u
  assume "gt_arctic x y" and "gt_arctic z u"
  thus "gt_arctic (x + z) (y + u)"
    by (cases y, cases u, simp, cases z, auto, cases x, auto, cases u, auto, cases z, auto, cases x, auto, cases x, auto, cases z, auto, cases x, auto)
next
  fix x y z
  assume "gt_arctic x y"
  thus "gt_arctic (x * z) (y * z)"
    by (cases y, simp, cases z, simp, cases x, auto)
next
  fix x
  assume "gt_arctic 0 x"
  thus "x = 0"
    by (cases x, auto simp: zero_arctic_def)
next
  show "pos_arctic 1" unfolding one_arctic_def by simp
next
  fix x
  show "gt_arctic x 0" unfolding zero_arctic_def by simp
next
  fix x y
  assume "pos_arctic x" 
  thus "pos_arctic (x + y)" by (cases x, simp, cases y, auto)
next
  fix x y
  assume "pos_arctic x" and "pos_arctic y"
  thus "pos_arctic (x * y)" by (cases x, simp, cases y, auto)
next
  show "SN {(x,y). pos_arctic y \<and> gt_arctic x y}" (is "SN ?rel")
    unfolding SN_defs
  proof (rule allI, rule notI)
    fix x
    assume "\<exists> f . f 0 = x \<and> (\<forall> i. (f i, f (Suc i)) \<in> ?rel)"
    from this obtain f where "f 0 = x" and seq: "\<forall> i. (f i, f (Suc i)) \<in> ?rel" by auto
    from seq have steps: "\<forall> i. gt_arctic (f i) (f (Suc i)) \<and> pos_arctic (f (Suc i)) " by auto
    let ?g = "\<lambda> i. get_arctic_num (f i)"
    have "\<forall> i. ?g (Suc i) \<ge> 0 \<and> ?g i  > ?g (Suc i)"
    proof
      fix i
      from steps have i: "gt_arctic (f i) (f (Suc i)) \<and> pos_arctic (f (Suc i))" by auto
      from i obtain n where fi: "f i = Num_arc n" by (cases "f (Suc i)", simp, cases "f i", auto)
      from i obtain m where fsi: "f (Suc i) = Num_arc m" by (cases "f (Suc i)", auto)
      with i have gz: "0 \<le> m" by simp
      from i fi fsi have "n > m" by auto
      with fi fsi gz
      show "?g (Suc i) \<ge> 0 \<and> ?g i > ?g (Suc i)" by auto
    qed
    from this obtain g where "\<forall> i. g (Suc i) \<ge> 0 \<and> (op > :: int \<Rightarrow> int \<Rightarrow> bool) (g i) (g (Suc i))" by auto
    hence "\<exists> f. f 0 = g 0 \<and> (\<forall> i. (f i, f (Suc i)) \<in> {(x,y). y \<ge> 0 \<and> x > y})" by auto
    with int_SN.SN show False unfolding SN_defs by auto
  qed 
next
  fix x :: arctic
  show "x \<ge> 0" unfolding zero_arctic_def by simp
qed


end




