(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RPO imports TermOrder ArgumentFilter
begin

fun lex_ext :: "(bool \<times> bool)list \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> bool \<times> bool"
where "lex_ext [] n m = (n > m, n \<ge> m)"
    | "lex_ext ( (stri,nstri) # bs ) n m = (if stri then (True,True) else if nstri then lex_ext bs n m else (False,False))"

lemma lex_ext_iff : "(lex_ext xs n m) = (
  (\<exists> i < length xs. (\<forall> j < i. snd (xs ! j)) \<and> fst (xs ! i)) \<or> 
  (\<forall> i < length xs. snd (xs ! i)) \<and> n > m,
  (\<exists> i < length xs. (\<forall> j < i. snd (xs ! j)) \<and> fst (xs ! i)) \<or> 
  (\<forall> i < length xs. snd (xs ! i)) \<and> n \<ge> m)
  "
proof (induct xs, simp)
  case (Cons pair xs)
  show ?case
  proof (cases pair)
    case (Pair a b)
    show ?thesis
    proof (cases a)
      case True
      with Pair Cons show ?thesis by auto
    next
      case False
      show ?thesis
      proof (cases b)
	case False 
	with Pair `\<not> a` show ?thesis by (force)
      next
	case True
	let ?c1 = "\<lambda> i xs. snd (xs ! i)"
	have id1: "(\<exists>i<Suc (length xs).
          (\<forall>j<i. snd (((False, True) # xs) ! j)) \<and>
          fst (((False, True) # xs) ! i)) =
	  (\<exists>i<length xs. (\<forall>j<i. snd (xs ! j)) \<and> fst (xs ! i))" (is "?l = ?r")
	proof
	  assume ?l
	  from this obtain i where i: "i < Suc (length xs)" and all: "\<forall> j < i. snd (((False,True) # xs) ! j)" and fst: "fst (((False, True) # xs) ! i)" by auto
	  from fst obtain j where i2: "i = Suc j" by (cases i, auto)
	  from i2 i have j: "j < length xs" by auto
	  from fst i2 have fst: "fst (xs ! j)" by auto
	  show ?r 
	  proof (rule exI[of _ j], simp add: j fst, intro allI impI)
	    fix k
	    assume "k < j"
	    with i i2 have "Suc k < i" by auto
	    with spec[OF all, of "Suc k"]
	    show "snd (xs ! k)" by auto
	  qed
	next
	  assume ?r
	  from this obtain i where i: "i < length xs" and all: "\<forall> j < i. snd (xs ! j)" and fst: "fst (xs ! i)" by auto
	  show ?l
	  proof (rule exI[of _ "Suc i"], simp add: i fst, intro allI impI)
	    fix j
	    assume "j < Suc i"
	    thus "snd (((False,True) # xs) ! j)"
	      by (cases j, auto simp: all)	    	  
	  qed
	qed	  
	have id2: " (\<forall>i<Suc (length xs). snd (((False, True) # xs) ! i)) =  (\<forall>i<length xs. snd (xs ! i))" (is "?l = ?r")
	proof
	  assume ?l
	  show ?r
	  proof (intro allI impI)
	    fix i
	    assume "i < length xs"
	    hence "Suc i < Suc (length xs)" by simp
	    with spec[OF `?l`, of "Suc i"] show "snd (xs ! i)" by auto
	  qed
	next
	  assume ?r
	  show ?l
	  proof (intro allI impI)
	    fix i
	    assume i: "i < Suc (length xs)"
	    show "snd (((False,True) # xs) ! i)"
	    proof (cases i, simp)
	      case (Suc j)
	      with i have "j < length xs" by simp
	      with spec[OF `?r`, of j] Suc show ?thesis by auto
	    qed
	  qed
	qed
	show ?thesis by (simp add: True False Cons Pair, simp only: id1, simp only: id2, auto)
      qed
    qed
  qed
qed    

lemma lex_ext_stri_imp_nstri : 
  assumes "fst (lex_ext bs n m)" 
  shows "snd (lex_ext bs n m)"
using assms
by (simp only: lex_ext_iff, simp, auto)

lemma all_nstri_imp_lex_nstri : assumes "\<forall> (a,b) \<in> set bs. b" and "n \<ge> m"
  shows "snd (lex_ext bs n m)"
using assms proof (induct bs)
  case Nil thus ?case by simp
next
  case (Cons b bs)
  obtain b1 b2 where "fst b = b1" and "snd b = b2" by auto
  hence b: "b = (b1,b2)" by auto
  from Cons have IH: "snd(lex_ext bs n m)" and "m \<le> n" and "b2" unfolding b by auto
  thus ?case unfolding b by simp
qed

lemma one_stri_imp_lex_stri : assumes "\<forall> (a,b) \<in> set bs. b"
  shows "fst (lex_ext (bs @ (True,b) # cs) n m)"
using assms proof (induct bs)
  case Nil thus ?case by simp
next
  case (Cons x bs)
   obtain x1 x2 where "fst x = x1" and "snd x = x2" by auto
   hence x: "x = (x1,x2)" by auto
  from Cons have IH: "fst (lex_ext (bs @ (True,b) # cs) n m)" and "x2" unfolding x by auto
  thus ?case unfolding x by simp
qed

lemma lex_ext_mono: fixes f :: "'a \<times> 'a \<Rightarrow> bool \<times> bool"
  assumes "\<forall> (x,y) \<in> set (zip xs ys). (fst (f (x,y)) \<longrightarrow> fst (f (g x,g y))) \<and> (snd (f (x,y)) \<longrightarrow> snd (f (g x,g y)))"
  shows "(fst (lex_ext (map f (zip xs ys)) n m) \<longrightarrow> fst (lex_ext (map f (zip (map g xs) (map g ys))) n m))
      \<and> (snd (lex_ext (map f (zip xs ys)) n m) \<longrightarrow> snd (lex_ext (map f (zip (map g xs) (map g ys))) n m))"
using assms
proof (induct xs arbitrary: ys, simp)
  case (Cons x xs) note oCons = this
  thus ?case
  proof (cases ys, simp)
    case (Cons y ys)
    show ?thesis 
    proof (cases "f (x,y)")
      case (Pair stri nstri) note oPair = this
      show ?thesis 
      proof (cases "f (g x, g y)")
	case (Pair gstri gnstri)
	show ?thesis 
	proof (cases stri)
	  case True
	  with Pair oPair have "fst (f (x,y))" by auto
	  with Pair oPair oCons Cons have "fst (f (g x, g y))" by auto
	  with Pair oPair oCons Cons have gstri by auto
	  with Cons Pair show ?thesis by auto
	next
	  case False note oFalse = this
	  with Cons `f (x, y) = (stri,nstri)` show ?thesis 
	  proof (cases "\<not> nstri", simp)
	    case False note iFalse = this
	    hence nstri by simp
	    with oFalse Pair oPair  oCons Cons have gnstri by auto
	    from Pair Cons show ?thesis 
	    proof (cases gstri, simp)
	      case False note iiFalse = this
	      with oCons Cons  False have ind: " (fst (lex_ext (map f (zip xs ys)) n m) \<longrightarrow> fst (lex_ext (map f (zip (map g xs) (map g ys))) n m)) \<and>
                                  (snd (lex_ext (map f (zip xs ys)) n m) \<longrightarrow> snd (lex_ext (map f (zip (map g xs) (map g ys))) n m))" (is "(?stri \<longrightarrow> ?gstri) \<and> (?nstri \<longrightarrow> ?gnstri)") 
		by auto
	      show ?thesis 
	      proof (cases ?nstri)
		case False
		hence "\<not> ?stri" using lex_ext_stri_imp_nstri by auto
		with Pair oPair Cons oCons show ?thesis by (simp)
	      next
		case True
		with ind have ?gnstri by auto
		with Pair oPair Cons oCons show ?thesis by auto
	      qed
	    qed
	  qed
	qed
      qed
    qed
  qed
qed

declare lex_ext.simps[simp del]

(* precedence, max arity, and two terms are given, result is pair (strict, non-strict) *)
function lpo :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool \<times> bool" where
  "lpo _ _ (Var x) (Var y) = (False, x = y)"
| "lpo pr _ (Var x) (Fun g ts) = (False, ts = [] \<and> pr g 0 = 0)"
| "lpo pr n (Fun f ss) (Var y) = (let con = any (\<lambda> s. snd (lpo pr n s (Var y))) ss in (con,con))" 
| "lpo pr n (Fun f ss) (Fun g ts) = (
    if any (\<lambda> s. snd (lpo pr n s (Fun g ts))) ss 
       then (True,True)
       else (let ls = length ss; lt = length ts; prl = pr f ls; prr = pr g lt in 
         if prl \<ge> prr \<and> (forall (\<lambda> t. fst (lpo pr n (Fun f ss) t)) ts)
          then if prl > prr then (True,True) 
            else if ls = lt \<or> lt \<le> n then (let comps = map (\<lambda> (s,t). lpo pr n s t) (zip ss ts) in lex_ext comps ls lt ) 
            else (False,False) 
          else (False,False)))"
by pat_completeness auto

termination lpo 
proof (relation "measures [\<lambda> (pr,n,s,t). size s + size t]", (auto)[4])
  fix "pr" n and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and prl prr  pair s t
  assume "pair \<in> set (zip ss ts)" and "(s,t) = pair"
  hence s: "s \<in> set ss" and t: "t \<in> set ts" using set_zip_leftD set_zip_rightD by auto
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 show "((pr, n, s, t), pr, n, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,n,s,t). size s + size t]"
    by auto
qed

lemma lpo_stri_imp_nstri[rule_format] :  "fst (lpo pr n s t) \<longrightarrow>
  snd (lpo pr n s t)"
using assms
proof (induct rule: lpo.induct[of "\<lambda> pr n s t. fst (lpo pr n s t) \<longrightarrow> snd (lpo pr n s t)"], (auto simp: Let_def)[3])
  case (4 "pr" n f ss g ts)
  show ?case 
    by (cases "\<not> (length ss = length ts \<or> length ts \<le> n)", simp add: Let_def, simp add: Let_def, 
      auto simp: lex_ext_stri_imp_nstri)
qed

lemma lpo_nstri_refl : "snd (lpo pr n s s)"
proof (induct s, simp)
  case (Fun f ss)
  have "\<forall> m. snd (lex_ext (map (\<lambda> (s,t). lpo pr n s t) (zip ss ss)) m m)"
  proof
    fix m
    show "snd (lex_ext (map (\<lambda> (s,t). lpo pr n s t) (zip ss ss)) m m)"
    proof (rule all_nstri_imp_lex_nstri)
      from Fun show "\<forall> (a,b) \<in> set (map (\<lambda> (s,t). lpo pr n s t) (zip ss ss)). b" 
	by (induct ss, simp, auto, force)
    next
      show "m \<le> m" by simp
    qed
  qed
  hence rec1: "snd (lex_ext (map (\<lambda> (s,t). lpo pr n s t) (zip ss ss)) (length ss) (length ss))" ..
  have rec2: "\<forall> s \<in> set ss. fst (lpo pr n (Fun f ss) s)"
  proof
    fix s
    assume s: "s \<in> set ss"
    with Fun have rec: "snd (lpo pr n s s)" by auto
    with s show "fst (lpo pr n (Fun f ss) s)"
    proof (cases s)
      case (Fun g ts)
      from s rec have "any (\<lambda> t. snd (lpo pr n t s)) ss" by (auto simp: any_mem_conv mem_iff)
      with Fun show ?thesis by auto
    next
      case (Var x)
      from s rec have "any (\<lambda> t. snd (lpo pr n t s)) ss" by (auto simp: any_mem_conv mem_iff)
      with Var show ?thesis by auto
    qed
  qed
  from rec1 rec2 show ?case by (auto simp: Let_def)
qed

lemma supt_imp_lpo_stri : assumes "s \<rhd> t" 
  shows "fst (lpo pr n s t)"
  using assms
proof (induct)
  case (arg s ss f)
  have refl: "snd (lpo pr n s s)" by (rule lpo_nstri_refl)
  from arg refl have fact: "any (\<lambda> t. snd (lpo pr n t s)) ss" by (auto simp: any_mem_conv mem_iff)
  thus ?case 
    by (cases s, auto)
next
  case (subt_of_arg s ss t f)
  from subt_of_arg lpo_stri_imp_nstri[of "pr" n s t] have "\<exists> s \<in> set ss. snd (lpo pr n s t)" by auto
  hence "any (\<lambda> s. snd (lpo pr n s t)) ss" by (auto simp: any_mem_conv mem_iff)
  thus ?case
    by (cases t, auto)
qed

lemma lpo_nstri_mono: assumes rel: "snd (lpo pr n s t)"
       shows "snd (lpo pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (lpo pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (lpo pr n ?s b)"
      proof (rule supt_imp_lpo_stri, cases "b \<in> set bef", (auto)[1], cases "b \<in> set aft", (auto)[1])
	assume "b \<notin> set bef" and "b \<notin> set aft"
	thus "?s \<rhd> b" using elem by blast
      qed
  qed
  from rel have "any (\<lambda> s. snd (lpo pr n s t)) ?ss" by (auto simp: any_mem_conv mem_iff)
  hence t: "fst (lpo pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). lpo pr n s t)"
  have lex_ext: "snd (lex_ext (?map (zip ?ss ?ts)) (length ?ss) (length ?ts))"
  proof (rule all_nstri_imp_lex_nstri, auto simp: zip_same)
    fix a b assume "lpo pr n s t = (a,b)" with rel show b by auto
  next
    fix a b u assume "(a,b) = lpo pr n u u" from this[symmetric] lpo_nstri_refl[of "pr" n u] show b by simp
  next
    fix a b u assume "(a,b) = lpo pr n u u" from this[symmetric] lpo_nstri_refl[of "pr" n u] show b by simp
  qed
  show ?thesis by (simp only: lpo.simps, cases "any (\<lambda> s. snd (lpo pr n s (Fun f ?ts))) ?ss", simp, simp add: forall_set_conv t aft_bef lex_ext[simplified])
qed

lemma lpo_stri_mono: assumes rel: "fst (lpo pr n s t)"
       shows "fst (lpo pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (lpo pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (lpo pr n ?s b)"
      proof (rule supt_imp_lpo_stri, cases "b \<in> set bef", (auto)[1], cases "b \<in> set aft", (auto)[1])
	assume "b \<notin> set bef" and "b \<notin> set aft"
	thus "?s \<rhd> b" using elem by blast
      qed
  qed
  from lpo_stri_imp_nstri[OF rel] have "any (\<lambda> s. snd (lpo pr n s t)) ?ss" by (auto simp: any_mem_conv mem_iff)
  hence t: "fst (lpo pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). lpo pr n s t)"
  obtain stri nstri where st: "lpo pr n s t = (stri,nstri)" by force
  with rel have stri: "stri = True" by simp
  have lex_ext: "fst (lex_ext (?map (zip ?ss ?ts)) (length ?ss) (length ?ts))" 
  proof (simp add: st stri, rule one_stri_imp_lex_stri, auto simp: zip_same)
    fix a b u assume "lpo pr n u u = (a,b)" with lpo_nstri_refl[of "pr" n u] show b by simp
  qed
  show ?thesis by (simp only: lpo.simps, cases "any (\<lambda> s. snd (lpo pr n s (Fun f ?ts))) ?ss", simp, simp add: forall_set_conv t aft_bef lex_ext[simplified])
qed


lemma lpo_stable : 
   "(fst (lpo pr n s t) \<longrightarrow> fst (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (lpo pr n s t) \<longrightarrow> snd (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"
proof (induct rule: lpo.induct[of "\<lambda> pr n s t. (fst (lpo pr n s t) \<longrightarrow> fst (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (lpo pr n s t) \<longrightarrow> snd (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"])
  case (1 "pr" n x y)
  show ?case by (auto simp: lpo_nstri_refl)
next
  case (2 "pr" n x g ts)
  show ?case by (cases "\<sigma>", cases "get_sub \<sigma> x", auto simp: Let_def all_nstri_imp_lex_nstri[of "[]", simplified])
next
  case (3 "pr" n f ss y)
  {
    assume "any (\<lambda> s. snd (lpo pr n s (Var y))) ss"
    from this obtain s where s: "s \<in> set ss" and "snd (lpo pr n s (Var y))" by (force simp: any_mem_conv mem_iff)
    with 3 have "snd (lpo pr n (s \<cdot> \<sigma>) (Var y \<cdot> \<sigma>))" by auto
    with s have "\<exists> s' \<in> set (map (\<lambda> s'. s' \<cdot> \<sigma>) ss). snd (lpo pr n s' (Var y \<cdot> \<sigma>))" by auto
    hence "any (\<lambda> s'. snd (lpo pr n s' (Var y \<cdot> \<sigma>))) (map (\<lambda> s'. s' \<cdot> \<sigma>) ss)" by (force simp: any_mem_conv mem_iff)
  }
  thus ?case by (cases "Var y \<cdot> \<sigma>", auto simp: Let_def)
next
  case (4 "pr" n f ss g ts)
  let ?map = "map (\<lambda> s'. s' \<cdot> \<sigma>)"
  let ?t = "Fun g ts"
  let ?s = "Fun f ss"
  let ?ssig = "Fun f ss \<cdot> \<sigma>"
  let ?tsig = "Fun g ts \<cdot> \<sigma>"
  show ?case
  proof (cases "\<exists> s \<in> set ss. snd (lpo pr n s ?t)")
    case True
    from this obtain s where s: "s \<in> set ss" and "snd (lpo pr n s ?t)" by auto
    with 4 have "snd (lpo pr n (s \<cdot> \<sigma>) ?tsig)" by auto
    with s have "\<exists> s' \<in> set (map (\<lambda> s'. s' \<cdot> \<sigma>) ss). snd (lpo pr n s' ?tsig)" by auto
    hence "any (\<lambda> s'. snd (lpo pr n s' ?tsig)) (map (\<lambda> s'. s' \<cdot> \<sigma>) ss)" by (force simp: any_mem_conv mem_iff)
    thus ?thesis by auto
  next
    case False
    hence not1: "\<not> (any (\<lambda> s. snd (lpo pr n s ?t)) ss)" by (force simp: any_mem_conv mem_iff)
    show ?thesis 
    proof (cases "snd (lpo pr n ?s ?t)")
      case False
      with lpo_stri_imp_nstri have "\<not> (fst (lpo pr n ?s ?t))" by blast
      with False show ?thesis by auto
    next
      case True
      with not1 have smallTs: "\<forall> t \<in> set ts. fst (lpo pr n ?s t)" by (cases ?thesis, auto)
      with 4 not1 have "\<forall> t \<in> set ts. fst (lpo pr n ?ssig (t \<cdot> \<sigma>))" by auto
      hence smallTssig: "\<forall> t \<in> set (?map ts). fst (lpo pr n ?ssig t)" by auto
      from True smallTs have ge: "pr f (length ss) \<ge> pr g (length ts)" by (cases ?thesis, auto simp: Let_def not1 simp del: any_set_conv)
      show ?thesis 
      proof (cases "pr f (length ss) > pr g (length ts)")
	case True
	with smallTssig not1 show ?thesis by auto
      next
	case False
	let ?mmap = "map (\<lambda> (s,t). lpo pr n s t)"
	from False ge have eq: "pr f (length ss) = pr g (length ts)" by auto
	with not1 smallTs True have snd_lex: "snd (lex_ext (?mmap (zip ss ts)) (length ss) (length ts)) \<and> (length ss = length ts \<or> length ts \<le> n)"
	  by (cases "length ss = length ts \<or> length ts \<le> n", auto)
	let ?ind = "\<lambda> s t. (fst (lpo pr n s t) \<longrightarrow> fst (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (lpo pr n s t) \<longrightarrow> snd (lpo pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))" 
	have ind: "\<forall> (s,t) \<in> set (zip ss ts). ?ind s t"
	proof (intro ballI, simp add: split_paired_all)
	  fix s t
	  assume mem: "(s,t) \<in> set (zip ss ts)"
	  show "?ind s t"
	    by (rule 4, rule not1, simp add: smallTs eq, (rule HOL.refl)+, simp add: eq False snd_lex, rule smallTs, rule False, simp add: eq snd_lex, rule mem, simp)
	qed
	have lex: "(fst (lex_ext (?mmap (zip ss ts)) (length ss) (length ts)) \<longrightarrow>
               fst (lex_ext (?mmap (zip (?map ss) (?map ts))) (length ss) (length ts))) \<and>
              (snd (lex_ext (?mmap (zip ss ts)) (length ss) (length ts)) \<longrightarrow>
               snd (lex_ext (?mmap (zip (?map ss) (?map ts))) (length ss) (length ts)))" (is "(?stri \<longrightarrow> ?mstri) \<and> (?nstri \<longrightarrow> ?mnstri)")
	  by (rule lex_ext_mono, auto simp: ind)
	with snd_lex have ?mnstri by auto
	have nstri: "snd (lpo pr n ?ssig ?tsig)" by (simp add: Let_def eq snd_lex smallTssig[simplified] `?mnstri`[simplified])
	{
	  assume "fst (lpo pr n ?s ?t)"
	  with not1 smallTs eq snd_lex have "?stri" by auto
	  with lex have "?mstri" by auto
	  have "fst (lpo pr n ?ssig ?tsig)" by (simp add: Let_def eq snd_lex smallTssig[simplified] `?mstri`[simplified])
	}
	with nstri show ?thesis by simp
      qed
    qed
  qed
qed

lemma lpo_least_1: assumes "pr f 0 = 0" 
  shows "snd (lpo pr n t (Fun f []))"
proof (cases t, simp add: assms)
  case (Fun g ts)
  with assms show ?thesis by (simp add: Let_def lex_ext.simps)
qed

lemma lpo_least_2: assumes "pr f 0 = 0" 
  shows "\<not> fst (lpo pr n (Fun f []) t)"
proof (cases t, simp)
  case (Fun g ts)
  with assms show ?thesis by (simp add: Let_def lex_ext.simps)
qed

lemma lpo_least_3: assumes "pr f 0 = 0" 
  and "snd (lpo pr n (Fun f []) t)"
  shows "snd (lpo pr n u t)"
using assms
proof (cases t, simp)
  case (Fun g ts)
  show ?thesis
  proof (cases ts)
    case Nil
    with Fun assms have "pr g 0 = 0" by (cases ?thesis, simp, simp add: Let_def)
    thus ?thesis by (simp only: Fun Nil, rule lpo_least_1)
  next
    case (Cons s tts)
    have "\<not> fst (lpo pr n (Fun f []) s)" by (rule lpo_least_2, rule `pr f 0 = 0`)
    with `snd (lpo pr n (Fun f []) t)` Fun Cons have False by (simp del: forall_set_conv add: Let_def forall.simps)
    thus ?thesis ..
  qed
qed
	
lemma lpo_compat[rule_format]: 
  fixes s t u :: "('f,'v)term"
  shows "(snd (lpo pr n s t) \<and> fst (lpo pr n t u) \<longrightarrow> fst (lpo pr n s u)) \<and> (fst (lpo pr n s t) \<and> snd (lpo pr n t u) \<longrightarrow> fst (lpo pr n s u)) \<and> (snd (lpo pr n s t) \<and> snd (lpo pr n t u) \<longrightarrow> snd (lpo pr n s u))" (is "?tran s t u")
proof (rule
    wf_induct[OF wf_measures[of "[\<lambda> (s,t,u). size s, \<lambda> (s,t,u). size t, \<lambda> (s,t,u). size u]"], 
    of "\<lambda> (s,t,u). ?tran s t u" "(s,t,u)", simplified], clarify)
  fix s t u :: "('f,'v)term"
  assume ind: "\<forall> s' t' u'. (size s' < size s \<longrightarrow> ?tran s' t' u') \<and> (size s' = size s \<and> (size t' < size t \<or> size t' = size t \<and> size u' < size u) \<longrightarrow> ?tran s' t' u')"
  let ?GR = "\<lambda> s t. fst (lpo pr n s t)"
  let ?GE = "\<lambda> s t. snd (lpo pr n s t)"
  let ?map = "map (\<lambda> (s,t). lpo pr n s t)"
  show "?tran s t u"
  proof (cases t)
    case (Var x)
    { 
      assume "?GR t u"
      with Var have False by (cases u, auto)
    } 
    moreover
    {
      assume gr: "?GR s t" and ge: "?GE t u"
      have "?GR s u"
      proof (cases u)
	case (Var y)
	with ge `t = Var x` have "t = u" by auto
	with gr show ?thesis by auto
      next
	case (Fun h us)
	with ge Var have us: "us = []" and pri: "pr h 0 = 0" by auto
	from gr Var obtain f ss where s: "s = Fun f ss" by (cases s, auto)
	with gr Var obtain s' where s': "s' \<in> set ss" by (auto simp: any_set_conv Let_def)
	have "?GE s' u" by (simp only: Fun us, rule lpo_least_1, rule pri)
	with s' have "any (\<lambda> s'. ?GE s' u) ss" by (auto simp: any_set_conv)
	with s Fun show ?thesis by simp
      qed
    }
    moreover
    {
      assume ge1: "?GE s t" and ge2: "?GE t u"
      have "?GE s u"
      proof (cases u)
	case (Var y)
	with ge2 `t = Var x` have "t = u" by auto
	with ge1 show ?thesis by auto
      next
	case (Fun h us)
	with ge2 Var have us: "us = []" and pri: "pr h 0 = 0" by auto
	show ?thesis by (simp only: Fun us, rule lpo_least_1, rule pri) 
      qed
    }
    ultimately show ?thesis by blast
  next
    case (Fun g ts)
    let ?t = "Fun g ts"
    from Fun have t: "t = ?t" .
    show ?thesis 
    proof (cases s)
      case (Var x)
      {
	assume gr: "?GR s t"
	with Var Fun have False by auto
      }
      moreover
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	with Var Fun have pri: "pr g 0 = 0" and "ts = []" by auto
	with gr Fun have False using lpo_least_2[of "pr" g n u] by auto
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u"
	with Var Fun have pri: "pr g 0 = 0" and empty: "ts = []" by auto
	have "?GE s u" by (rule lpo_least_3[of "pr" g n u], rule pri, simp add: empty[symmetric] Fun[symmetric], rule ge2)
      }
      ultimately show ?thesis by blast
    next
      case (Fun f ss)
      let ?s = "Fun f ss"
      from Fun have s: "s = ?s" .
      let ?s1 = "\<exists> s' \<in> set ss. ?GE s' t"
      let ?t1 = "\<exists> t' \<in> set ts. ?GE t' u"
      let ?ls = "length ss"
      let ?lt = "length ts"
      let ?pf = "pr f ?ls"
      let ?pg = "pr g ?lt"
      from s have "\<forall> s' \<in> set ss. size s' < size s" by auto
      with ind have ind2: "\<forall> s' \<in> set ss. \<forall> t' u'. ?tran s' t' u'" by blast
      let ?ex = "\<lambda> xs ys. (\<exists>i< min (length xs) (length ys).
	(\<forall>j<i. snd (?map (zip xs ys) ! j)) \<and>
	fst (?map (zip xs ys) ! i))"
      let ?all = "\<lambda> xs ys. (\<forall>i< min (length xs) (length ys).
	snd (?map (zip xs ys) ! i))"
      let ?lens = "\<lambda> xs ys. length xs > length ys"
      let ?lenn = "\<lambda> xs ys. length xs \<ge> length ys"
      have lex_ext: "\<And> us p q. \<lbrakk>?ex ss ts \<or> ?all ss ts \<and> p ss ts; ?ex ts us \<or> ?all ts us \<and> q ts us; p ss ts \<longrightarrow> ?ls \<ge> ?lt; q ts us \<longrightarrow> ?lt \<ge> length us\<rbrakk> \<Longrightarrow> ?ex ss us \<or> ?all ss us \<and> (p ss ts \<and> q ts us \<or> ?ls > length us)"
      proof -
	fix us p q
	assume st: "?ex ss ts \<or> ?all ss ts \<and> p ss ts"
	assume tu: "?ex ts us \<or> ?all ts us \<and> q ts us"
	assume plen: "p ss ts \<longrightarrow> ?ls \<ge> ?lt"
	assume qlen: "q ts us \<longrightarrow> ?lt \<ge> length us"
	let ?st = "zip ss ts"
	let ?su = "zip ss us"
	let ?tu = "zip ts us"
	from st
	show "?ex ss us \<or> ?all ss us \<and> (p ss ts \<and> q ts us \<or> ?ls > length us)"
	proof
	  assume "?ex ss ts"
	  from this obtain i where il: "i < min ?ls ?lt" and "fst (?map ?st ! i)" and "\<forall> i' < i. snd (?map ?st ! i')" by blast
	  hence iset: "ss ! i \<in> set ss" and i: "?GR (ss ! i) (ts ! i)" and i': "\<forall> i' < i. ?GE (ss ! i') (ts ! i')" by auto
	  from tu show ?thesis
	  proof
	    assume "?ex ts us"
	    from this obtain j where jl: "j < min ?lt (length us)" and "fst (?map ?tu ! j)" and "\<forall> j' < j. snd (?map ?tu ! j')" by blast
	    hence j: "?GR (ts ! j) (us ! j)" and j': "\<forall> j' < j. ?GE (ts ! j') (us ! j')" by auto
	    show ?thesis 
	    proof (cases "i < j")
	      case True
	      with j' have "?GE (ts ! i) (us ! i)" by auto
	      from i this spec[OF spec[OF bspec[OF ind2, OF iset]], of "ts ! i" "us ! i"]
	      have i: "?GR (ss ! i) (us ! i)" by simp
	      have i': "\<forall> i' < i. ?GE (ss ! i') (us ! i')"
	      proof (intro allI impI)
		fix i'
		assume i'l: "i' < i"
		from i'l i' have one: "?GE (ss ! i') (ts ! i')" by auto
		from i'l True j' have two: "?GE (ts ! i') (us ! i')" by auto
		from i'l il have "ss ! i' \<in> set ss" by auto
		from one two spec[OF spec[OF bspec[OF ind2, OF this]], of "ts ! i'" "us ! i'"]
		show "?GE (ss ! i') (us ! i')" by simp
	      qed
	      from True il jl have "i < min ?ls (length us)" by auto
	      with i i' show ?thesis by auto
	    next
	      case False
	      hence ij: "i \<ge> j" by simp
	      show ?thesis 
	      proof (cases "i > j")
		case True
		with il have jset: "ss ! j \<in> set ss" by auto
		from True i' have "?GE (ss ! j) (ts ! j)" by auto		
		from j this spec[OF spec[OF bspec[OF ind2, OF jset]], of "ts ! j" "us ! j"]
		have j: "?GR (ss ! j) (us ! j)" by simp
		have j': "\<forall> j' < j. ?GE (ss ! j') (us ! j')"
		proof (intro allI impI)
		  fix j'
		  assume j'l: "j' < j"
		  from j'l i' True have one: "?GE (ss ! j') (ts ! j')" by auto
		  from j'l j' have two: "?GE (ts ! j') (us ! j')" by auto
		  from j'l il True  have "ss ! j' \<in> set ss" by auto
		  from one two spec[OF spec[OF bspec[OF ind2, OF this]], of "ts ! j'" "us ! j'"]
		  show "?GE (ss ! j') (us ! j')" by simp
		qed
		from True il jl have "j < min ?ls (length us)" by auto
		with j j' show ?thesis by auto
	      next
		case False
		with ij have ji: "j = i" by simp
		with j have "?GR (ts ! i) (us ! i)" by auto
		with lpo_stri_imp_nstri have "?GE (ts ! i) (us ! i)" by auto
		from i this spec[OF spec[OF bspec[OF ind2, OF iset]], of "ts ! i" "us ! i"]
		have i: "?GR (ss ! i) (us ! i)" by simp
		have i': "\<forall> i' < i. ?GE (ss ! i') (us ! i')"
		proof (intro allI impI)
		  fix i'
		  assume i'l: "i' < i"
		  from i'l i' have one: "?GE (ss ! i') (ts ! i')" by auto
		  from i'l ji j' have two: "?GE (ts ! i') (us ! i')" by auto
		  from i'l il have "ss ! i' \<in> set ss" by auto
		  from one two spec[OF spec[OF bspec[OF ind2, OF this]], of "ts ! i'" "us ! i'"]
		  show "?GE (ss ! i') (us ! i')" by simp
		qed
		from ji il jl have "i < min ?ls (length us)" by auto
		with i i' show ?thesis by auto
	      qed
	    qed
	  next
	    assume "?all ts us \<and> q ts us"
	    hence j': "\<forall> j < min ?lt (length us). ?GE (ts ! j) (us ! j)" by auto
	    show ?thesis 
	    proof (cases "i < min ?lt (length us)")
	      case True
	      with j' have "?GE (ts ! i) (us ! i)" by auto
	      from i this spec[OF spec[OF bspec[OF ind2, OF iset]], of "ts ! i" "us ! i"]
	      have i: "?GR (ss ! i) (us ! i)" by simp
	      have i': "\<forall> i' < i. ?GE (ss ! i') (us ! i')"
	      proof (intro allI impI)
		fix i'
		assume i'l: "i' < i"
		from i'l i' have one: "?GE (ss ! i') (ts ! i')" by auto
		from i'l True j' have two: "?GE (ts ! i') (us ! i')" by auto
		from i'l il have "ss ! i' \<in> set ss" by auto
		from one two spec[OF spec[OF bspec[OF ind2, OF this]], of "ts ! i'" "us ! i'"]
		show "?GE (ss ! i') (us ! i')" by simp
	      qed
	      from True il have "i < min ?ls (length us)" by auto
	      with i i' show ?thesis by auto
	    next
	      case False
	      have i': "\<forall> i' < min ?ls (length us). ?GE (ss ! i') (us ! i')"
	      proof (intro allI impI)
		fix i'
		assume i'l: "i' < min ?ls (length us)"
		from i'l i' il False have one: "?GE (ss ! i') (ts ! i')" by auto
		from i'l j' il False have two: "?GE (ts ! i') (us ! i')" by auto 
		from i'l i' il have "ss ! i' \<in> set ss" by auto
		from one two spec[OF spec[OF bspec[OF ind2, OF this]], of "ts ! i'" "us ! i'"]
		show "?GE (ss ! i') (us ! i')" by simp
	      qed
	      from il False have "?ls > length us" by auto
	      with i' show ?thesis by auto
	    qed
	  qed
	next
	  assume "?all ss ts \<and> p ss ts"
	  with plen have i': "\<forall> i < min ?ls ?lt. ?GE (ss ! i) (ts ! i)" and p: "p ss ts" and stl: "?ls \<ge> ?lt"
	  and i'set: "\<forall> i < min ?ls ?lt. ss ! i \<in> set ss" by auto
	  from tu 
	  show ?thesis 
	  proof
	    assume "?ex ts us"
	    from this obtain j where jl: "j < min ?lt (length us)" and "fst (?map ?tu ! j)" and "\<forall> j' < j. snd (?map ?tu ! j')" by blast
	    hence j: "?GR (ts ! j) (us ! j)" and j': "\<forall> j' < j. ?GE (ts ! j') (us ! j')" by auto
	    from jl stl have len: "j < min ?ls ?lt" by auto
	    with i' i'set have "?GE (ss ! j) (ts ! j)" and jset: "ss ! j \<in> set ss" by auto
	    from j this spec[OF spec[OF bspec[OF ind2, OF jset]], of "ts ! j" "us ! j"]
	    have j: "?GR (ss ! j) (us ! j)" by simp
	    have j': "\<forall> j' < j. ?GE (ss ! j') (us ! j')"
	    proof (intro allI impI)
	      fix j'
	      assume j'l: "j' < j"
	      from j'l i' len have one: "?GE (ss ! j') (ts ! j')" by auto
	      from j'l j' have two: "?GE (ts ! j') (us ! j')" by auto
	      from j'l len i'set have "ss ! j' \<in> set ss" by auto
	      from one two spec[OF spec[OF bspec[OF ind2, OF this]], of "ts ! j'" "us ! j'"]
	      show "?GE (ss ! j') (us ! j')" by simp
	    qed
	    from len jl have "j < min ?ls (length us)" by auto
	    with j j' show ?thesis by auto	  
	  next
	    assume "?all ts us \<and> q ts us"
	    with qlen have j': "\<forall> i < min ?lt (length us). ?GE (ts ! i) (us ! i)" and q: "q ts us" and tul: "?lt \<ge> length us" 
	      by auto
	    have "\<forall> i < min ?ls (length us). ?GE (ss ! i) (us ! i)"
	    proof (intro allI impI)
	      fix i
	      assume i: "i < min ?ls (length us)"
	      hence iset: "ss ! i \<in> set ss" by auto
	      from i stl tul i' have one: "?GE (ss ! i) (ts ! i)" by auto
	      from i stl tul j' have two: "?GE (ts ! i) (us ! i)" by auto
	      from one two spec[OF spec[OF bspec[OF ind2, OF iset]], of "ts ! i" "us ! i"]
	      show "?GE (ss ! i) (us ! i)" by simp
	    qed
	    with p q show ?thesis by auto
	  qed
	qed
      qed
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' spec[OF spec[OF bspec[OF ind2]], of s' t u, simplified] ges gr have "?GR s' u" by auto
	  hence "?GE s' u" by (rule lpo_stri_imp_nstri)
	  with s' s show ?thesis by (cases u, auto simp: any_set_conv Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    from t' t False ge s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto simp: any_set_conv)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" by force
	  next
	    case False
	    from t this gr obtain h us where u: "u = Fun h us" by (cases u, auto simp: any_set_conv Let_def)
	    let ?u = "Fun h us"	    
	    from s t u ge gr have ge: "?GE ?s ?t" and gr: "?GR ?t ?u" by auto
	    from `\<not> ?s1` t ge have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto simp: any_set_conv)
	    from `\<not> ?t1` u gr have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto simp: any_set_conv)
	    let ?lu = "length us"
	    let ?ph = "pr h ?lu"
	    from `\<not> ?s1` t ge st' have fg: "?pf \<ge> ?pg" by (cases ?thesis, auto simp: Let_def any_set_conv) 
	    from `\<not> ?t1` u gr tu' have fh: "?pg \<ge> ?ph" by (cases ?thesis, auto simp: Let_def any_set_conv) 
	    from tu' s t ge spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" by force
	    show ?thesis
	    proof (cases "?pf > ?pg \<or> ?pg > ?ph")
	      case True
	      with fg fh have "?pf > ?ph" by auto
	      with su' s u show ?thesis by (auto)
	    next
	      case False
	      from False fg fh have fg: "?pf = ?pg" and gh: "?pg = ?ph" and fh: "?pf = ?ph" by auto
	      from fg `\<not> ?s1` t ge st' have "snd (lex_ext (?map (zip ss ts)) ?ls ?lt) \<and> (?ls = ?lt \<or> ?lt \<le> n)" by (cases "?ls = ?lt \<or> ?lt \<le> n", auto simp: Let_def any_set_conv)
	      hence lex1: "?ex ss ts \<or> ?all ss ts \<and> ?lenn ss ts" and disj1: "?ls = ?lt \<or> ?lt \<le> n" by (simp only: lex_ext_iff, simp+)
	      from gh `\<not> ?t1` u gr tu' have "fst (lex_ext (?map (zip ts us)) ?lt ?lu) \<and> (?lt = ?lu \<or> ?lu \<le> n)" by (cases "?lt = ?lu \<or> ?lu \<le> n", auto simp: Let_def any_set_conv)
	      hence lex2: "?ex ts us \<or> ?all ts us \<and> ?lens ts us" and disj2: "?lt = ?lu \<or> ?lu \<le> n" by (simp only: lex_ext_iff, simp+)	      
	      from disj1 disj2 have disj: "?ls = ?lu \<or> ?lu \<le> n" by arith
	      from lex_ext[of ?lenn us ?lens, OF lex1 lex2, simplified]	      
	      have "fst (lex_ext (?map (zip ss us)) ?ls ?lu)"
		by (simp only: lex_ext_iff, auto)	      
	      with s u fh su' disj show ?thesis by auto
	    qed
	  qed
	qed
      }
      moreover
      {
	assume gr: "?GR s t" and ge: "?GE t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' spec[OF spec[OF bspec[OF ind2]], of s' t u, simplified] ges ge have "?GE s' u" by auto
	  with s' s show ?thesis by (cases u, auto simp: any_set_conv Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    have tt': "?GE t t'" by (rule lpo_stri_imp_nstri[OF supt_imp_lpo_stri], auto simp: t t')
	    from t' t False gr s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto simp: any_set_conv)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" by force
	  next
	    case False
	    from t this ge obtain h us where u: "u = Fun h us" by (cases u, auto simp: any_set_conv Let_def)
	    let ?u = "Fun h us"	    
	    from s t u ge gr have gr: "?GR ?s ?t" and ge: "?GE ?t ?u" by auto
	    from `\<not> ?s1` t gr have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto simp: any_set_conv)
	    from `\<not> ?t1` u ge have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto simp: any_set_conv)
	    let ?lu = "length us"
	    let ?ph = "pr h ?lu"
	    from `\<not> ?s1` t gr st' have fg: "?pf \<ge> ?pg" by (cases ?thesis, auto simp: Let_def any_set_conv) 
	    from `\<not> ?t1` u ge tu' have fh: "?pg \<ge> ?ph" by (cases ?thesis, auto simp: Let_def any_set_conv) 
	    from tu' s t lpo_stri_imp_nstri[OF gr] spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" by force
	    show ?thesis
	    proof (cases "?pf > ?pg \<or> ?pg > ?ph")
	      case True
	      with fg fh have "?pf > ?ph" by auto
	      with su' s u show ?thesis by (auto)
	    next
	      case False
	      with fg fh have fg: "?pf = ?pg" and gh: "?pg = ?ph" and fh: "?pf = ?ph" by auto
	      from fg `\<not> ?s1` t gr st' have "fst (lex_ext (?map (zip ss ts)) ?ls ?lt) \<and> (?ls = ?lt \<or> ?lt \<le> n)" by (cases "?ls = ?lt \<or> ?lt \<le> n", auto simp: Let_def any_set_conv)
	      hence lex1: "?ex ss ts \<or> ?all ss ts \<and> ?lens ss ts" and disj1: "?ls = ?lt \<or> ?lt \<le> n" by (simp only: lex_ext_iff, simp+)
	      from gh `\<not> ?t1` u ge tu' have "snd (lex_ext (?map (zip ts us)) ?lt ?lu)  \<and> (?lt = ?lu \<or> ?lu \<le> n)" by (cases "?lt = ?lu \<or> ?lu \<le> n", auto simp: Let_def any_set_conv)	     
	      hence lex2: "?ex ts us \<or> ?all ts us \<and> ?lenn ts us" and disj2: "?lt = ?lu \<or> ?lu \<le> n" by (simp only: lex_ext_iff, simp+)
	      from disj1 disj2 have disj: "?ls = ?lu \<or> ?lu \<le> n" by arith
	      from lex_ext[of ?lens us ?lenn, OF lex1 lex2, simplified]
	      have "fst (lex_ext (?map (zip ss us)) ?ls ?lu)"
		by (simp only: lex_ext_iff, auto)
	      with s u fh su' disj show ?thesis by (auto)
	    qed
	  qed
	qed
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u" and ngt1: "\<not> ?GR s t" and ngt2: "\<not> ?GR t u"
	have "\<not> ?s1"
	proof
	  assume ?s1
	  with s have "?GR s t" by (cases t, auto simp: any_set_conv)
	  with ngt1 show False ..
	qed
	have "\<not> ?t1"
	proof
	  assume ?t1
	  with t have "?GR t u" by (cases u, auto simp: any_set_conv)
	  with ngt2 show False ..
	qed	
	from t `\<not> ?t1` ge2 obtain h us where u: "u = Fun h us" by (cases u, auto simp: any_set_conv Let_def)
	let ?u = "Fun h us"	    
	from s t u ge1 ge2 have ge1: "?GE ?s ?t" and ge2: "?GE ?t ?u" by auto
	from `\<not> ?s1` t ge1 have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto simp: any_set_conv)
	from `\<not> ?t1` u ge2 have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto simp: any_set_conv)
	let ?lu = "length us"
	let ?ph = "pr h ?lu"
	from tu' s t ge1 spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" by force
	from `\<not> ?s1` st' ge1 ngt1 s t have fg: "?pf = ?pg" by (simp add: any_set_conv Let_def, cases ?thesis, simp, cases "?pf > ?pg", auto) 
	from `\<not> ?t1` tu' ge2 ngt2 t u have gh: "?pg = ?ph" by (simp add: any_set_conv Let_def, cases ?thesis, simp, cases "?pg > ?ph", auto)
	from `\<not> ?s1` st' ge1 s t fg have "snd (lex_ext (?map (zip ss ts)) ?ls ?lt)  \<and> (?ls = ?lt \<or> ?lt \<le> n)" by (cases "?ls = ?lt \<or> ?lt \<le> n", auto simp: Let_def any_set_conv)
	hence lex1: "?ex ss ts \<or> ?all ss ts \<and> ?lenn ss ts" and disj1: "?ls = ?lt \<or> ?lt \<le> n" by (simp only: lex_ext_iff, simp+)
	from `\<not> ?t1` tu' ge2 t u gh have "snd (lex_ext (?map (zip ts us)) ?lt ?lu) \<and> (?lt = ?lu \<or> ?lu \<le> n)" by (cases "?lt = ?lu \<or> ?lu \<le> n", auto simp: any_set_conv Let_def)
	hence lex2: "?ex ts us \<or> ?all ts us \<and> ?lenn ts us" and disj2: "?lt = ?lu \<or> ?lu \<le> n" by (simp only: lex_ext_iff, simp+)
	from disj1 disj2 have disj: "?ls = ?lu \<or> ?lu \<le> n" by arith
	from lex_ext[of ?lenn us ?lenn, OF lex1 lex2, simplified]
	have "snd (lex_ext (?map (zip ss us)) ?ls ?lu)"
	  by (simp only: lex_ext_iff, auto)
	with fg gh su' s u disj have "?GE s u" by auto
      } 
      ultimately
      show ?thesis using lpo_stri_imp_nstri[of "pr" n s u] by auto
    qed
  qed
qed


lemma lpo_trans: assumes one: "fst (lpo pr n s t)" and two: "fst (lpo pr n t u)" shows "fst (lpo pr n s u)"
  using assms lpo_compat lpo_stri_imp_nstri[OF one] by blast

definition LPO_S where "LPO_S pr n = {(s,t) | s t. fst (lpo pr n s t)}"
definition LPO_NS where "LPO_NS pr n = {(s,t) | s t. snd (lpo pr n s t)}"

lemma LPO_NS_refl: "refl (LPO_NS pr n)"
unfolding refl_on_def LPO_NS_def
 by (simp add: lpo_nstri_refl)

lemma LPO_NS_trans: "trans (LPO_NS pr n)"
unfolding trans_def LPO_NS_def 
using lpo_compat by blast

lemma LPO_S_trans: "trans (LPO_S pr n)"
unfolding trans_def LPO_S_def 
using lpo_trans by blast

abbreviation lpo_prop where "lpo_prop pr n X \<equiv> \<forall> t. (\<forall> s. fst (lpo pr n t s) \<longrightarrow> s \<in> X) \<longrightarrow> t \<in> X"

(* note that the properties of W W1 W2 do not depend on the definition of lpo, they can
   be generalized. However, due to lazyness, this is currently not done *)
definition W :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f,'v)term set" where "W pr n \<equiv> \<Inter> { X. lpo_prop pr n X}"

lemma W1 : "(t \<in> W pr n) = ( \<forall> s. fst (lpo pr n (t :: ('f,'v)term) s) \<longrightarrow> s \<in> W pr n)" (is "?r = ?l")
proof
  assume ?l
  show "t \<in> W pr n"
    unfolding W_def
  proof (clarify)
    fix X :: "('f,'v)term set"
    assume im: "lpo_prop pr n X"
    hence subset: "W pr n \<subseteq> X" unfolding W_def by blast
    from im have imp: "(\<forall> s. fst (lpo pr n t s) \<longrightarrow> s \<in> X) \<longrightarrow> t \<in> X" ..
    have "\<forall> s. fst (lpo pr n t s) \<longrightarrow> s \<in> X" using `?l` subset by blast
    with imp show "t \<in> X" by blast
  qed
next
  assume ?r
  show ?l 
  proof (intro allI impI)
    fix s :: "('f,'v)term"
    assume lpo: "fst (lpo pr n t s)"
    show "s \<in> W pr n" unfolding W_def
    proof (clarify)
      fix X :: "('f,'v)term set"
      assume imp: "lpo_prop pr n X"
      show "s \<in> X" 
      proof (rule ccontr)
	assume "s \<notin> X"
	hence "lpo_prop pr n (X - {t})" using imp lpo by blast
	with `?r` show False unfolding W_def by blast
      qed
    qed
  qed
qed

	
fun W2_fun :: "(('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
  where "W2_fun P pr n 0 = (SOME x . x \<in> W pr n \<and> \<not> P x)"
  | "W2_fun P pr n (Suc m) = (SOME x. x \<in> W pr n \<and> \<not> P x \<and> fst (lpo pr n (W2_fun P pr n m) x))"

lemma W2[rule_format] : "\<forall> t \<in> W pr m. ( \<forall> s. fst (lpo pr m (t :: ('f,'v)term) s) \<longrightarrow> P s) \<longrightarrow> P t \<Longrightarrow> \<forall> t \<in> W pr m. P t"
proof (rule ccontr)
  assume one: "\<not> (\<forall> t \<in> W pr m. P t)"
    and other: "\<forall> (t :: ('f,'v)term) \<in> W pr m. (\<forall> s. fst (lpo pr m t s) \<longrightarrow> P s) \<longrightarrow> P t"
  have "\<forall> n. W2_fun P pr m n \<in> W pr m \<and> \<not> P (W2_fun P pr m n) \<and> (n > 0 \<longrightarrow> fst (lpo pr m (W2_fun P pr m (n - 1)) (W2_fun P pr m n)))" (is "\<forall> n. ?P n")
  proof
    fix n
    show "?P n"
    proof (induct n)
      case 0
      from one obtain t where tW: "t \<in> W pr m \<and> \<not> P t" by auto 
      show ?case by (simp, rule someI[of "\<lambda> t. t \<in> W pr m \<and> \<not> P t" t], rule tW)
    next
      case (Suc n)
      with other obtain s where lpo: "fst (lpo pr m (W2_fun P pr m n) s) \<and> \<not> P s" by blast
      from Suc W1[of "W2_fun P pr m n" "pr" m] lpo have "s \<in> W pr m" by blast
      with lpo have conj: "s \<in> W pr m \<and> \<not> P s \<and> fst (lpo pr m (W2_fun P pr m n) s)" by simp
      from someI[of "\<lambda> s. s \<in> W pr m \<and> \<not> P s \<and> fst (lpo pr m (W2_fun P pr m n) s)" s, OF conj]
      show ?case by simp
    qed
  qed
  hence all: "\<forall> n. W2_fun P pr m n \<in> W pr m \<and> fst (lpo pr m (W2_fun P pr m n) (W2_fun P pr m (Suc n)))"
    by auto
  from this obtain g :: "nat \<Rightarrow> ('f,'v)term" where g: "(\<forall> n. g n \<in> W pr m \<and> fst (lpo pr m (g n) (g (Suc n))))" by blast 
  from g have "lpo_prop pr m (W pr m - range g)" unfolding W_def by blast
  with g show False unfolding W_def by blast
qed

hide const W2_fun


lemma W_imp_W_next: assumes "t \<in> W pr m" and "fst (lpo pr m t s)"
  shows "s \<in> W pr m"
  using assms W1[of t] by blast


lemma W_imp_SN : fixes "pr" :: "'f \<Rightarrow> nat \<Rightarrow> nat" and t :: "('f,'v)term"
  assumes t: "t \<in> W pr m" shows "SN_elt (LPO_S pr m) t"
proof (rule ccontr)
  assume "\<not> ?thesis"
  from this obtain g where g0: "g 0 = t" and steps: "\<forall> n. fst (lpo pr m (g n) (g (Suc n)))" unfolding SN_elt_def[of "LPO_S pr m" t] LPO_S_def by force
  have g1: "\<forall> n. g n \<in> W pr m" 
  proof
    fix n
    show "g n \<in> W pr m"
    proof (induct n, simp add: g0 t)
      case (Suc n)
      thus ?case by (rule W_imp_W_next[OF _ spec[OF steps, of n]])
    qed
  qed
  with steps have "lpo_prop pr m (W pr m - range g)" unfolding W_def by blast
  with g1 show False unfolding W_def by blast
qed


lemma var_in_W : "Var x \<in> W pr n"
proof (simp only: W1[of "Var x" "pr" n], intro allI impI)
  fix s
  assume "fst (lpo pr n (Var x) s)"
  hence False by (cases s, auto)
  thus "s \<in> W pr n" ..
qed

lemma nstri_and_W : assumes t: "t \<in> W pr n" and nstri: "snd (lpo pr n t s)"
  shows "s \<in> W pr n"
proof (simp add: W1[of s], intro impI allI)
  fix u
  assume "fst (lpo pr n s u)"
  with lpo_compat nstri have u: "fst (lpo pr n t u)" by blast
  with t W1[of t] show "u \<in> W pr n" by auto
qed

definition W_list : "W_list pr m ts \<equiv> \<forall> t \<in> set ts. t \<in> W pr m"

lemma LPO_compat : "LPO_NS pr n O LPO_S pr n \<subseteq> LPO_S pr n"
  using lpo_compat[of "pr" n] 
  by (unfold LPO_S_def LPO_NS_def, auto)

  
lemma fun_in_W[rule_format] : fixes "pr" :: "'f \<Rightarrow> nat \<Rightarrow> nat" and ts :: "('f,'v)term list"
  assumes "W_list pr m ts"
  shows "Fun f ts \<in> W pr m"
proof -
  let ?cond = "\<lambda> ys xs :: ('f,'v)term list.(\<forall> y \<in> set ys. SN_elt (LPO_S pr m) y) \<and> (length ys = length xs \<or> length xs \<le> m) \<and> fst (lex_ext (map (\<lambda> (y,x). lpo pr m y x) (zip ys xs)) (length ys) (length xs))" 
  let ?r4 = "{(ys :: ('f,'v)term list,xs). ?cond ys xs}"
  let ?r3 = "{(xs :: ('f,'v)term list,ys). ?cond ys xs}"
  let ?r2 = "{(xs :: ('f,'v)term list,ys). W_list pr m ys \<and> (length ys = length xs \<or> length xs \<le> m) \<and> fst (lex_ext (map (\<lambda> (y,x). lpo pr m y x) (zip ys xs)) (length ys) (length xs))}"
  let ?r1 = "{(x :: nat,y). x < y} <*lex*> ?r2"
  let ?m = "\<lambda> (f,ts). (pr f (length ts), ts)"
  let ?r = "{(a,b). (?m a, ?m b) \<in> ?r1}"
  let ?gr = "\<lambda> t s. fst (lpo pr m t s)"
  let ?ge = "\<lambda> t s. snd (lpo pr m t s)"
  have supset: "?r3 \<supseteq> ?r2" unfolding W_list using W_imp_SN[of _ "pr" m] by auto
  have "wf ?r2" 
  proof (rule wf_subset[OF _ supset])
    have id: "?r4^-1 = ?r3" by auto
    have "(SN :: ('f,'v)term list brel \<Rightarrow> bool) ?r4" 
    proof (rule ccontr)
      assume "\<not> ?thesis"
      from this obtain f where f: "\<forall> n :: nat. ?cond (f n) (f (Suc n))" unfolding SN_defs by blast
      have m_imp_m: "\<forall> n. length (f n) \<le> m \<longrightarrow> length (f (Suc n)) \<le> m"
      proof (intro allI impI)
	fix n
	assume "length (f n) \<le> m"
	thus "length (f (Suc n)) \<le> m"
	  using spec[OF f, of n] by auto
      qed
      have lm_imp_m_or_eq: "\<forall> n. length (f n) > m \<longrightarrow> length (f (Suc n)) \<le> m \<or> length (f n) = length (f (Suc n))"
      proof (intro allI impI)
	fix n
	assume "length (f n) > m"
	hence "\<not> length (f n) \<le> m" by auto
	thus "length (f (Suc n)) \<le> m \<or> length (f n) = length (f (Suc n))"
	  using spec[OF f, of n] by blast
      qed
      let ?l0 = "max (length (f 0)) m"
      have "\<forall> n. length (f n) \<le> ?l0"
      proof (intro allI)
	fix n
	show "length (f n) \<le> ?l0"
	proof (induct n, simp)
	  case (Suc n)
	  show ?case
	  proof (cases "length (f n) \<le> m")
	    case True
	    with spec[OF m_imp_m, of n] show ?thesis by auto
	  next
	    case False
	    hence "length (f n) > m" by auto
	    with spec[OF lm_imp_m_or_eq, of n] 
	    have "length (f n) = length (f (Suc n)) \<or> length (f (Suc n)) \<le> m" by auto
	    with Suc show ?thesis by auto
	  qed
	qed
      qed
      from this obtain m' where len: "\<forall> n. length (f n) \<le> m'" by auto
      let ?lexgr = "\<lambda> ys xs. fst (lex_ext (map (\<lambda> (y,x). lpo pr m y x) (zip ys xs)) (length ys) (length xs))"
      let ?lexge = "\<lambda> ys xs. snd (lex_ext (map (\<lambda> (y,x). lpo pr m y x) (zip ys xs)) (length ys) (length xs))"
      let ?con = "\<lambda> ys (xs :: ('f,'v)term list) m'.(\<forall> y \<in> set ys. SN_elt (LPO_S pr m) y) \<and> length ys \<le> m' \<and> ?lexgr ys xs"
      let ?confn = "\<lambda> m' f n . ?con (f n) (f (Suc n)) m'"
      from f len have  "\<exists> f. \<forall> n. ?confn m' f n" by auto
      thus False
      proof (induct m')
	case 0
	from this obtain f where "?confn 0 f 0" by auto	
	hence "?lexgr [] (f (Suc 0))" by force
	thus False by (simp add: lex_ext.simps)
      next
	case (Suc m')
	from this obtain f where confn: "\<forall> n. ?confn (Suc m') f n" by auto
	have ne: "\<forall> n. f n \<noteq> []"
	proof
	  fix n
	  show "f n \<noteq> []"
	  proof (cases "f n")
	    case (Cons a b) thus ?thesis by auto
	  next
	    case Nil
	    with spec[OF confn, of n] show ?thesis by (simp add: lex_ext.simps)
	  qed
	qed
	let ?hf = "\<lambda> n. hd (f n)"
	have ge: "\<forall> n. ?ge (?hf n) (?hf (Suc n))"
	proof
	  fix n
	  from spec[OF ne, of n] obtain a as where n: "f n = a # as" by (cases "f n", auto)
	  from spec[OF ne, of "Suc n"] obtain b bs where sn: "f (Suc n) = b # bs" by (cases "f (Suc n)", auto)
	  from n sn lpo_stri_imp_nstri[of "pr" m a b] have "?ge a b" 
	  proof (cases "?gr a b", simp, cases "?ge a b", simp)
	    assume "\<not> ?gr a b" and "\<not> ?ge a b" 
	    hence "lpo pr m a b = (False,False)" by (cases "lpo pr m a b", auto)
	    with spec[OF confn, of n] n sn have False by (simp add: lex_ext.simps)
	    thus "?ge a b" ..
	  qed
	  with n sn show "?ge (?hf n) (?hf (Suc n))" by simp
	qed
	let ?S = "(LPO_S :: ('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f,'v)trs) pr m"
	let ?NS = "(LPO_NS :: ('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f,'v)trs) pr m"
	from ge have GE: "\<forall> n. (?hf n, ?hf (Suc n)) \<in> ?NS \<union> ?S" unfolding LPO_NS_def by auto
	from spec[OF confn, of 0] spec[OF ne, of 0] have SN_0: "SN_elt ?S (?hf 0)" by (cases "f 0", auto )
	from non_strict_ending[of ?hf, OF GE LPO_compat SN_0]
	obtain j where j: "\<forall> i \<ge> j. (?hf i, ?hf (Suc i)) \<in> ?NS - ?S" by auto
	let ?g = "\<lambda> n. tl (f (j + n))"
	obtain g where g: "g = ?g"  by auto
	have "\<forall> n. ?confn m' g n" 
	proof
	  fix n
	  let ?nj = "j + n"
	  from spec[OF j, of ?nj]
	  have ge_not_gr: "(?hf ?nj, ?hf (Suc ?nj)) \<in> ?NS - ?S" by simp
	  from spec[OF confn, of ?nj] have old: "?confn (Suc m') f ?nj" by simp
	  from spec[OF ne, of ?nj] obtain a as where n: "f ?nj = a # as" by (cases "f ?nj", auto)
	  from spec[OF ne, of "Suc ?nj"] obtain b bs where sn: "f (Suc ?nj) = b # bs" by (cases "f (Suc ?nj)", auto)
	  from old have one: "\<forall> y \<in> set (g n). SN_elt ?S y" 
	    by (simp add: g n)
	  from old have two: "length (g n) \<le> m'" by (simp add: g n)
	  from ge_not_gr have ge_not_gr2: "lpo pr m a b = (False, True)" by (simp add: n sn LPO_S_def LPO_NS_def, cases "lpo pr m a b", auto)
	  from old have three: "?lexgr (g n) (g (Suc n))" by (simp add: g n sn ge_not_gr2 lex_ext.simps, simp add: lex_ext_iff) 
	  from one two three show "?confn m' g n" by blast
	qed
	with Suc show ?thesis by blast
      qed
    qed
    thus "wf ?r3" by (simp only: SN_iff_wf[of ?r4] id)
  qed
  hence wf_r1: "wf ?r1" by (rule wf_lex_prod[OF wf_less, of ?r2])
  have wf_r: "wf ?r" using wf_inv_image[OF wf_r1, of ?m] unfolding inv_image_def .
  let ?W = "W pr m"
  let ?WL = "W_list pr m"
  let ?ind = "\<lambda> (f,ts).  ?WL ts \<longrightarrow> Fun f ts \<in> ?W"
  have "?ind (f,ts)" 
  proof (rule wf_induct[OF wf_r, of ?ind])
    fix fts
    assume ind: "\<forall> gss. (gss,fts) \<in> ?r \<longrightarrow> ?ind gss"
    show "?ind fts"
    proof (cases fts)
      case (Pair f ts)
      thus ?thesis
      proof (simp, intro impI)
	assume ts: "?WL ts"
	let ?t = "Fun f ts"	
	show "?t \<in> ?W"
	proof (simp only: W1[of ?t], intro allI impI)
	  fix s
	  assume "?gr ?t s"
	  thus "s \<in> ?W"
	  proof (induct s, simp add: var_in_W)
	    case (Fun g ss)
	    let ?s = "Fun g ss"
	    from Fun have t_gr_s: "?gr ?t ?s" by auto
	    show "?s \<in> ?W"
	    proof (cases "\<exists> t \<in> set ts. ?ge t ?s")
	      case True
	      then obtain t where "t \<in> set ts" and ge: "?ge t ?s" by auto	      
	      with ts have "t \<in> ?W" unfolding W_list by auto
	      show "?s \<in> ?W" by (rule nstri_and_W[OF `t \<in> ?W` ge])
	    next
	      case False
	      let ?ls = "length ss"
	      let ?lt = "length ts"
	      let ?pf = "pr f ?lt"
	      let ?pg = "pr g ?ls"
	      from False t_gr_s have f_ge_g: "?pf \<ge> ?pg" by (cases ?thesis, auto simp: Let_def any_set_conv)
	      from False t_gr_s f_ge_g have small_ss: "\<forall> s \<in> set ss. ?gr ?t s" by (cases ?thesis, auto simp: Let_def any_set_conv)
	      with Fun have ss_W: "?WL ss" unfolding W_list by auto
	      show ?thesis
	      proof (cases "?pf = ?pg")
		case False
		with f_ge_g have f_gt_g: "?pf > ?pg" by auto
		hence "((g,ss),(f,ts)) \<in> ?r" by (simp add: in_lex_prod)
		with ind have "?ind (g,ss)" using Pair by auto
		with ss_W show ?thesis by auto
	      next
		case True
		with False t_gr_s small_ss have disj: "?lt = ?ls \<or> ?ls \<le> m" by (cases ?thesis, auto simp: Let_def any_set_conv)
		from True False t_gr_s small_ss disj have lex: "fst (lex_ext (map (\<lambda> (s,t). lpo pr m s t) (zip ts ss)) ?lt ?ls)" by (cases ?thesis, auto simp: Let_def any_set_conv)
		from True lex ts disj have "((g,ss),(f,ts)) \<in> ?r" by (simp add: in_lex_prod)
		with ind have "?ind (g,ss)" using Pair by auto
		with ss_W show ?thesis by auto
	      qed
	    qed
	  qed
	qed
      qed
    qed
  qed
  with assms show ?thesis by auto
qed

lemma W_is_univ: "t \<in> W pr m"
  by (induct t, auto simp: var_in_W fun_in_W W_list)


lemma LPO_S_SN: "SN (LPO_S pr n)" 
  unfolding SN_def by (intro allI, rule W_imp_SN, rule W_is_univ)

hide const W W_list

interpretation LPO_redpair : redpair_order "LPO_S pr n" "LPO_NS pr n"
proof (unfold_locales, rule LPO_S_SN, rule LPO_compat)
  show "sub_closed (LPO_NS pr n)"
    unfolding sub_closed_def 
  proof (rule subsetI2)
    fix s t
    assume "(s,t) \<in> CS (LPO_NS pr n)"
    then obtain u v \<sigma> where "(u,v) \<in> LPO_NS pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: closure_sub.cases) auto
    thus "(s,t) \<in> LPO_NS pr n" unfolding LPO_NS_def using lpo_stable[of "pr" n] by auto
  qed
next
  show "sub_closed (LPO_S pr n)"
    unfolding sub_closed_def 
  proof (rule subsetI2)
    fix s t
    assume "(s,t) \<in> CS (LPO_S pr n)"
    then obtain u v \<sigma> where "(u,v) \<in> LPO_S pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: closure_sub.cases) auto
    thus "(s,t) \<in> LPO_S pr n" unfolding LPO_S_def using lpo_stable[of "pr" n] by auto
  qed
next
  show "one_step_ctxt_closed (LPO_NS pr n)"
    unfolding one_step_ctxt_closed_def LPO_NS_def
    by (intro allI impI, clarify, simp del: lpo.simps, rule lpo_nstri_mono)
next
  show "refl (LPO_NS pr n)" by (rule LPO_NS_refl)
next
  show "trans (LPO_NS pr n)" by (rule LPO_NS_trans)
next
  show "trans (LPO_S pr n)" by (rule LPO_S_trans)
qed 

definition lpo_strict where "lpo_strict pr n \<equiv> \<lambda> (s,t). checkB (fst (lpo pr n s t)) (''could not orient '' +#+ shows s +@+ '' >LPO '' +#+ shows t +@+ shows_nl)"
definition lpo_nstrict where "lpo_nstrict pr n \<equiv> \<lambda> (s,t). checkB (snd (lpo pr n s t)) (''could not orient '' +#+ shows s +@+ '' >=LPO '' +#+ shows t +@+ shows_nl)"

interpretation LPO_redpair_impl : redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n"
  by (unfold_locales, unfold LPO_S_def LPO_NS_def lpo_strict_def lpo_nstrict_def, simp_all add: split_paired_all) 

interpretation LPO_ce_redpair : ce_redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n" m c d
proof
  show "ce_trs (c,d,m) \<subseteq> LPO_NS pr n"
    unfolding LPO_NS_def 
    by (simp add: ce_trs.simps, auto simp: lpo_stri_imp_nstri[OF supt_imp_lpo_stri])
qed

interpretation LPO_ce_mono_redpair : ce_mono_redpair_impl "LPO_S pr n" "LPO_NS pr n" m c d "lpo_strict pr n" "lpo_nstrict pr n" ch
proof
  show "ce_trs (c,d,m) \<subseteq> LPO_S pr n"
    unfolding LPO_S_def 
    by (simp add: ce_trs.simps, auto simp: supt_imp_lpo_stri)
next
  show "one_step_ctxt_closed (LPO_S pr n)"
    unfolding one_step_ctxt_closed_def LPO_S_def
    by (intro allI impI, clarify, simp del: lpo.simps, rule lpo_stri_mono)
qed

definition full_af :: "'f af" where "full_af f n i \<equiv> True"

interpretation LPO_ce_af_redpair : ce_af_redpair_impl "LPO_S pr n" "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n" m c d full_af
  by  (unfold_locales, unfold af_compatible_def full_af_def, auto)

locale LPO_afs_impl = 
  fixes "pr" :: "('f::shows) filtered \<Rightarrow> nat \<Rightarrow> nat" and n :: nat and c d :: 'f  and pi :: "'f afs_list"
  assumes check: "isOK(check_wf_afs pi)"


sublocale LPO_afs_impl \<subseteq> afs_ce_redpair_impl pi "LPO_S pr n"  "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n" c d 
proof
  from check show "wf_afs (afs_list_to_afs pi)" using check_wf_afs_sound afs.wf_afs by blast
qed

sublocale LPO_afs_impl \<subseteq> afs_ce_mono_redpair_impl pi "LPO_S pr n"  "LPO_NS pr n" "lpo_strict pr n" "lpo_nstrict pr n" c d cm .. 

end
