module Ceta = 
struct

let rec id x = (fun xa -> xa) x;;

type 'a eq = {eq : 'a -> 'a -> bool};;
let eq _A = _A.eq;;

type nat = Zero_nat | Suc of nat;;

type ('a, 'b) scg = Null | Scg of 'a * 'a * ('b * 'b) list * ('b * 'b) list;;

type 'a shp = Sharp of 'a | Plain of 'a;;

type xml =
  Xml of
    char list * (char list * char list) list * xml list * (char list) option;;

let rec eqa _A a b = eq _A a b;;

let rec leta s f = f s;;

type 'a ord = {less_eq : 'a -> 'a -> bool; less : 'a -> 'a -> bool};;
let less_eq _A = _A.less_eq;;
let less _A = _A.less;;

let rec nat_aux
  i n = (if Big_int.le_big_int i (Big_int.big_int_of_int 0) then n
          else nat_aux (Big_int.sub_big_int i (Big_int.big_int_of_int 1))
                 (Suc n));;

let rec nat i = nat_aux i Zero_nat;;

let rec hd (x :: xs) = x;;

let rec tl = function [] -> []
             | x :: xs -> xs;;

let rec less_eq_nat
  x0 n = match x0, n with Suc m, n -> less_nat m n
    | Zero_nat, n -> true
and less_nat
  m x1 = match m, x1 with m, Suc n -> less_eq_nat m n
    | n, Zero_nat -> false;;

let rec lex_ext
  x0 n m = match x0, n, m with [], n, m -> (less_nat m n, less_eq_nat m n)
    | (stri, nstri) :: bs, n, m ->
        (if stri then (true, true)
          else (if nstri then lex_ext bs n m else (false, false)));;

let rec list_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, a :: lista -> f2 a lista
    | f1, f2, [] -> f1;;

let rec zip
  xs x1 = match xs, x1 with xs, [] -> []
    | xs, y :: ys -> (match xs with [] -> [] | z :: zs -> (z, y) :: zip zs ys);;

let rec split f (a, b) = f a b;;

let rec map
  f x1 = match f, x1 with f, [] -> []
    | f, x :: xs -> f x :: map f xs;;

let rec fst (a, b) = a;;

let rec forall
  p x1 = match p, x1 with p, [] -> true
    | p, x :: xs -> p x && forall p xs;;

let rec plus_nat
  x0 n = match x0, n with Suc m, n -> plus_nat m (Suc n)
    | Zero_nat, n -> n;;

let rec size_list
  = function [] -> Zero_nat
    | a :: lista -> plus_nat (size_list lista) (Suc Zero_nat);;

let rec snd (a, b) = b;;

let rec any
  uu x1 = match uu, x1 with uu, [] -> false
    | p, x :: xs -> p x || any p xs;;

let rec eq_nat
  x0 x1 = match x0, x1 with Zero_nat, Zero_nat -> true
    | Suc nat, Suc nat' -> eq_nat nat nat'
    | Zero_nat, Suc nat' -> false
    | Suc nat', Zero_nat -> false;;

let rec null = function [] -> true
               | x :: xs -> false;;

type ('a, 'b) term = Var of 'b | Fun of 'a * ('a, 'b) term list;;

let rec lpo _B
  uu uv x2 x3 = match uu, uv, x2, x3 with
    uu, uv, Var x, Var y -> (false, eqa _B x y)
    | pr, uw, Var x, Fun (g, ts) ->
        (false, null ts && eq_nat (pr g Zero_nat) Zero_nat)
    | pr, n, Fun (f, ss), Var y ->
        let con = any (fun s -> snd (lpo _B pr n s (Var y))) ss in
        (con, con)
    | pr, n, Fun (f, ss), Fun (g, ts) ->
        (if any (fun s -> snd (lpo _B pr n s (Fun (g, ts)))) ss
          then (true, true)
          else let ls = size_list ss in
               let lt = size_list ts in
               let prl = pr f ls in
               let prr = pr g lt in
               (if less_eq_nat prr prl &&
                     forall (fun t -> fst (lpo _B pr n (Fun (f, ss)) t)) ts
                 then (if less_nat prr prl then (true, true)
                        else (if eq_nat ls lt || less_eq_nat lt n
                               then let comps =
                                      map (fun (a, b) -> lpo _B pr n a b)
(zip ss ts)
                                      in
                                    lex_ext comps ls lt
                               else (false, false)))
                 else (false, false)));;

type ('a, 'b) sub = Sub of ('b -> ('a, 'b) term);;

let rec foldl
  f a x2 = match f, a, x2 with f, a, [] -> a
    | f, a, x :: xs -> foldl f (f a x) xs;;

let one_nat : nat = Suc Zero_nat;;

let rec sz
  = function Var x -> one_nat
    | Fun (f, ts) ->
        plus_nat one_nat (foldl (fun a x -> plus_nat a (sz x)) Zero_nat ts);;

type 'a poly = PVar of nat | PNum of 'a | PSum of 'a poly list |
  PMult of 'a poly list;;

let rec option_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Some a -> f2 a
    | f1, f2, None -> f1;;

let rec decomp
  x0 x1 = match x0, x1 with [], [] -> Some []
    | uu :: uv, [] -> None
    | [], uw :: ux -> None
    | s :: ss, t :: ts ->
        (match decomp ss ts with None -> None
          | Some pairs -> Some ((s, t) :: pairs));;

let rec append
  x0 ys = match x0, ys with [], ys -> ys
    | x :: xs, ys -> x :: append xs ys;;

type ('a, 'b) hterm = HHole | HFun of 'a * ('a, 'b) hterm list;;

let rec gcMatchList _A
  = function [] -> Some []
    | (HHole, uu) :: ps -> gcMatchList _A ps
    | (HFun (f, hs), Fun (g, ss)) :: ps ->
        (if eqa _A f g
          then (match decomp hs ss with None -> None
                 | Some pairs -> gcMatchList _A (append pairs ps))
          else None)
    | (HFun (f, hs), Var x) :: ps ->
        (match gcMatchList _A ps with None -> None
          | Some pairs -> Some ((HFun (f, hs), x) :: pairs));;

let rec mergeHTerms _A
  x0 x1 = match x0, x1 with [], [] -> Some []
    | [], uu :: uv -> None
    | uw :: ux, [] -> None
    | HHole :: hsl, h :: hsr ->
        (match mergeHTerms _A hsl hsr with None -> None
          | Some res -> Some (h :: res))
    | HFun (f, hs) :: hsl, HHole :: hsr ->
        (match mergeHTerms _A hsl hsr with None -> None
          | Some res -> Some (HFun (f, hs) :: res))
    | HFun (f, fs) :: hsl, HFun (g, gs) :: hsr ->
        (if eqa _A f g
          then (match mergeHTerms _A fs gs with None -> None
                 | Some mhs ->
                   (match mergeHTerms _A hsl hsr with None -> None
                     | Some res -> Some (HFun (f, mhs) :: res)))
          else None);;

let rec nth
  x0 x1 = match x0, x1 with x :: xs, Suc n -> nth xs n
    | x :: xs, Zero_nat -> x;;

let rec merge _A
  hl hr =
    (match mergeHTerms _A [hl] [hr] with None -> None
      | Some res -> Some (nth res Zero_nat));;

let rec mergeX _A _B
  x h xa2 = match x, h, xa2 with x, h, [] -> Some []
    | x, h, (ht, y) :: ps ->
        (if eqa _A x y
          then (match merge _B h ht with None -> None
                 | Some m -> mergeX _A _B x m ps)
          else (match mergeX _A _B x h ps with None -> None
                 | Some pairs -> Some ((ht, y) :: pairs)));;

let rec mergeAll _A _B
  = function [] -> true
    | (h, x) :: ps ->
        (match mergeX _B _A x h ps with None -> false
          | Some a -> mergeAll _A _B a);;

let rec gcMatchI _A _B
  h t = (match gcMatchList _A [(h, t)] with None -> false
          | Some a -> mergeAll _A _B a);;

let rec capI _A _B
  uu x1 = match uu, x1 with uu, Var uv -> HHole
    | r, Fun (f, ts) ->
        let h = HFun (f, map (capI _A _B r) ts) in
        (if any (fun ra -> gcMatchI _A _B h (fst ra)) r then HHole else h);;

let rec comp f g = (fun x -> f (g x));;

let rec rev xs = foldl (fun xsa x -> x :: xsa) [] xs;;

let rec upt i j = (if less_nat i j then i :: upt (Suc i) j else []);;

type pos = Empty | PCons of nat * pos;;

let rec concat_map
  f x1 = match f, x1 with f, [] -> []
    | f, x :: xs -> append (f x) (concat_map f xs);;

let rec poss_list
  = function Var x -> [Empty]
    | Fun (f, ss) ->
        Empty ::
          concat_map (fun (i, a) -> map (fun aa -> PCons (i, aa)) a)
            (zip (upt Zero_nat (size_list ss)) (map poss_list ss));;

let rec subt_at
  s x1 = match s, x1 with s, Empty -> s
    | Fun (f, ss), PCons (i, p) -> subt_at (nth ss i) p;;

let rec decompa
  x0 x1 = match x0, x1 with [], [] -> Some []
    | uu :: uv, [] -> None
    | [], uw :: ux -> None
    | s :: ss, t :: ts ->
        (match decompa ss ts with None -> None
          | Some pairs -> Some ((s, t) :: pairs));;

let rec decompose_one _A
  x0 t = match x0, t with Var x, t -> Some [(Var x, t)]
    | Fun (v, va), Var y -> None
    | Fun (f, ts), Fun (g, ss) -> (if eqa _A f g then decompa ts ss else None);;

let rec decompose_n _A
  x0 s t = match x0, s, t with Zero_nat, s, t -> decompose_one _A s t
    | Suc n, s, t ->
        (match decompose_n _A n s t with None -> None
          | Some a ->
            foldl (fun aa (sa, ta) ->
                    (match aa with None -> None
                      | Some ys ->
                        (match decompose_one _A sa ta with None -> None
                          | Some xs -> Some (append xs ys))))
              (Some []) a);;

let rec decompose _A s t = decompose_n _A (sz s) s t;;

let rec decompose_list _A
  ts = foldl (fun a (s, t) ->
               (match a with None -> None
                 | Some ls ->
                   (match decompose _A s t with None -> None
                     | Some ys -> Some (append ls ys))))
         (Some []) ts;;

let rec sub_case f1 (Sub funa) = f1 funa;;

let rec sub_apply
  x0 sigma = match x0, sigma with Var x, Sub sigma -> sigma x
    | Fun (f, ss), sigma -> Fun (f, map (fun t -> sub_apply t sigma) ss);;

let rec list_all2
  p xs ys = match p, xs, ys with
    p, x :: xs, y :: ys -> p x y && list_all2 p xs ys
    | p, xs, [] -> null xs
    | p, [], ys -> null ys;;

let rec eq_term _A _B
  x0 x1 = match x0, x1 with
    Fun (f, ts), Fun (g, us) -> eq _A f g && list_all2 (eq_term _A _B) ts us
    | Var x, Var y -> eq _B x y
    | Var v, Fun (f', list') -> false
    | Fun (f', list'), Var v -> false;;

let rec eq_terma _A _B = ({eq = eq_term _A _B} : ('a, 'b) term eq);;

let rec eqb _A _B (x1, y1) (x2, y2) = eqa _A x1 x2 && eqa _B y1 y2;;

let rec eqc _A _B = ({eq = eqb _A _B} : ('a * 'b) eq);;

let rec member _A
  x xa1 = match x, xa1 with x, [] -> false
    | x, y :: ys -> eqa _A x y || member _A x ys;;

let rec subst_safe _A _B
  vs x1 sigma = match vs, x1, sigma with vs, [], sigma -> Some sigma
    | vs, (Var x, t) :: ts, sigma ->
        (if member (eqc (eq_terma _B _A) (eq_terma _B _A)) (Var x, t) ts
          then subst_safe _A _B vs ts sigma
          else (if member _A x vs &&
                     not (eq_term _B _A (sub_apply (Var x) sigma) t)
                 then None
                 else (match subst_safe _A _B (x :: vs) ts sigma
                        with None -> None
                        | Some (Sub s) ->
                          Some (Sub (fun y ->
                                      (if eqa _A x y then t else s y))))));;

let rec match_all _A _B
  ts = (match decompose_list _A ts with None -> None
         | Some ls -> subst_safe _B _A [] ls (Sub (fun a -> Var a)));;

let rec nat_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Suc nat -> f2 nat
    | f1, f2, Zero_nat -> f1;;

let rec drop
  n x1 = match n, x1 with n, [] -> []
    | n, x :: xs -> (match n with Zero_nat -> x :: xs | Suc m -> drop m xs);;

let rec take
  n x1 = match n, x1 with n, [] -> []
    | n, x :: xs -> (match n with Zero_nat -> [] | Suc m -> x :: take m xs);;

type ('a, 'b) ctxt = Hole |
  More of 'a * ('a, 'b) term list * ('a, 'b) ctxt * ('a, 'b) term list;;

let rec ctxt_of_pos_term
  x0 t = match x0, t with Empty, t -> Hole
    | PCons (i, ps), Fun (f, ts) ->
        More (f, take i ts, ctxt_of_pos_term ps (nth ts i), drop (Suc i) ts);;

type nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;;

let rec shows_string x = append x;;

let rec shows_sep
  s sep x2 = match s, sep, x2 with s, sep, [] -> shows_string []
    | s, sep, [x] -> s x
    | s, sep, x :: v :: va ->
        comp (comp (s x) sep) (shows_sep s sep (v :: va));;

let rec shows_between l p r = comp l (comp p r);;

let rec shows_list_gen
  elt e l s r xs =
    (if null xs then shows_string e
      else shows_between (shows_string l) (shows_sep elt (shows_string s) xs)
             (shows_string r));;

type 'a shows =
  {shows : 'a -> char list -> char list;
    shows_list : 'a list -> char list -> char list};;
let shows _A = _A.shows;;
let shows_list _A = _A.shows_list;;

let rec shows_term _A _B
  = function Var x -> shows _B x
    | Fun (f, ts) ->
        comp (shows _A f)
          (shows_list_gen id [] ['('] [','] [')'] (map (shows_term _A _B) ts));;

let rec ctxt_apply
  x0 s = match x0, s with Hole, s -> s
    | More (f, ss1, c, ss2), s -> Fun (f, append ss1 (ctxt_apply c s :: ss2));;

type ('a, 'b) either = Left of 'a | Right of 'b;;

let rec error x = Left x;;

let rec return x = Right x;;

let rec checkB b e = (if b then return () else error e);;

let rec rewrites (_A1, _A2) (_B1, _B2)
  (s, t) c sigma (l, ra) r =
    checkB
      (eq_term _A1 _B1 s (ctxt_apply c (sub_apply l sigma)) &&
        (eq_term _A1 _B1 t (ctxt_apply c (sub_apply ra sigma)) &&
          member (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) (l, ra) r))
      (comp (shows_string ['T'; 'e'; 'r'; 'm'; ' '])
        (comp (shows_term _A2 _B2 s)
          (comp (shows_string
                  [' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'r'; 'e';
                    'w'; 'r'; 'i'; 't'; 'e'; ' '; 't'; 'o'; ' '; 't'; 'e'; 'r';
                    'm'; ' '])
            (shows_term _A2 _B2 t))));;

let rec either_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Right b -> f2 b
    | f1, f2, Left a -> f1 a;;

let rec isValue p = (match p with Left _ -> false | Right _ -> true);;

let rec getpr (_A1, _A2) (_B1, _B2)
  s t p (l, r) =
    (match match_all _A1 _B1 [(l, subt_at s p); (r, subt_at t p)]
      with None -> None
      | Some sigma ->
        (if isValue
              (rewrites (_A1, _A2) (_B1, _B2) (s, t) (ctxt_of_pos_term p s)
                sigma (l, r) [(l, r)])
          then Some (s, (ctxt_of_pos_term p s, (sigma, (l, r)))) else None));;

let rec getp (_A1, _A2) (_B1, _B2)
  s t p x3 = match s, t, p, x3 with s, t, p, [] -> None
    | s, t, p, r :: rs ->
        (match getpr (_A1, _A2) (_B1, _B2) s t p r
          with None -> getp (_A1, _A2) (_B1, _B2) s t p rs | Some a -> Some a);;

let rec find_get_p (_A1, _A2) (_B1, _B2)
  s t x2 r = match s, t, x2, r with s, t, [], r -> None
    | s, t, p :: ps, r ->
        (match getp (_A1, _A2) (_B1, _B2) s t p r
          with None -> find_get_p (_A1, _A2) (_B1, _B2) s t ps r
          | Some a -> Some a);;

let rec get (_A1, _A2) (_B1, _B2)
  s t r =
    (match find_get_p (_A1, _A2) (_B1, _B2) s t (poss_list s) r
      with None ->
        error (comp (shows_string
                      ['C'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'f';
                        'i'; 'n'; 'd'; ' '; 'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i';
                        'o'; 'n'; ' '; 'b'; 'e'; 't'; 'w'; 'e'; 'e'; 'n'; ' ';
                        't'; 'e'; 'r'; 'm'; ' '])
                (comp (shows_term _A2 _B2 s)
                  (comp (shows_string
                          [' '; 'a'; 'n'; 'd'; ' '; 't'; 'e'; 'r'; 'm'; ' '])
                    (comp (shows_term _A2 _B2 t) (shows_string ['.'])))))
      | Some a -> return a);;

let rec prod_case f1 (a, b) = f1 a b;;

let rec remdups _A
  = function [] -> []
    | x :: xs ->
        (if member _A x xs then remdups _A xs else x :: remdups _A xs);;

let rec compa _A
  es es' =
    remdups (eqc _A _A)
      (concat_map
        (fun (x, y) ->
          concat_map (fun (y', z) -> (if eqa _A y y' then [(x, z)] else []))
            es')
        es);;

type ('a, 'b, 'c, 'd, 'e, 'f) alter = Alter1 of 'a | Alter2 of 'b | Alter3 of 'c
  | Alter4 of 'd | Alter5 of 'e | Alter6 of 'f;;

let newline : char = '\010';;

let wspace : char list = [newline; '\009'; '\013'; ' '];;

let eq_char = ({eq = (fun a b -> ((a : char) = b))} : char eq);;

let rec trim
  = function c :: s -> (if member eq_char c wspace then trim s else c :: s)
    | [] -> [];;

let rec capRM2 _A _B
  uu x1 = match uu, x1 with uu, Var uv -> HHole
    | rm, Fun (f, ts) ->
        let h = HFun (f, map (capRM2 _A _B rm) ts) in
        (if any (fun r -> gcMatchI _A _B h (fst r)) (rm f) then HHole else h);;

let rec capRM _A _B
  r = (match r with None -> (fun _ -> HHole) | Some a -> capRM2 _A _B a);;

let eq_int = ({eq = Big_int.eq_big_int} : Big_int.big_int eq);;

let eq_nata = ({eq = eq_nat} : nat eq);;

let rec combine _B
  x0 s = match x0, s with
    Scg (p, q, str, wk), Scg (p', q', str', wk') ->
      Scg (p, q, remdups (eqc _B _B) (append str str'),
            remdups (eqc _B _B) (append wk wk'))
    | Null, s -> Null
    | Scg (v, va, vb, vc), Null -> Null;;

let rec list_all
  p x1 = match p, x1 with p, [] -> true
    | p, x :: xs -> p x && list_all p xs;;

let rec subsumes _A _B
  g x1 = match g, x1 with
    Scg (p, q, str, wk), Scg (p', q', str', wk') ->
      eqa _A p p' &&
        (eqa _A q q' &&
          (list_all (fun x -> member (eqc _B _B) x str') str &&
            list_all (fun x -> member (eqc _B _B) x (append str' wk')) wk))
    | g, Null -> true
    | Null, Scg (v, va, vb, vc) -> false;;

let rec filter
  p x1 = match p, x1 with p, [] -> []
    | p, x :: xs -> (if p x then x :: filter p xs else filter p xs);;

let rec subtract_list _A xs ys = filter (fun x -> not (member _A x ys)) xs;;

let rec scg_comp _B
  conn x1 g = match conn, x1, g with
    conn, Scg (p, q, str, wk), Scg (p', q', str', wk') ->
      (if not (conn q p') then Null
        else let strs =
               remdups (eqc _B _B)
                 (append (compa _B str str')
                   (append (compa _B str wk') (compa _B wk str')))
               in
             let a =
               subtract_list (eqc _B _B) (remdups (eqc _B _B) (compa _B wk wk'))
                 strs
               in
             Scg (p, q', strs, a))
    | conn, Null, g -> Null
    | conn, Scg (v, va, vb, vc), Null -> Null;;

let rec list_ex
  p x1 = match p, x1 with p, [] -> false
    | p, x :: xs -> p x || list_ex p xs;;

let rec in_situ _B
  = function Null -> true
    | Scg (p, q, str, wk) -> list_ex (fun (a, b) -> eqa _B a b) str;;

let rec sagiv _A _B
  conn g =
    (if in_situ _B g then true
      else let gg = scg_comp _B conn g g in
           (if subsumes _A _B gg g then false
             else sagiv _A _B conn (combine _B g gg)));;

let rec show _A x = shows _A x [];;

let rec args = function Var uu -> []
               | Fun (uv, ts) -> ts;;

let rec root = function Var x -> None
               | Fun (f, ts) -> Some f;;

let rec eq_shp _A
  x0 x1 = match x0, x1 with Sharp f, Sharp f' -> eqa _A f f'
    | Plain f, Plain f' -> eqa _A f f'
    | Sharp f, Plain f' -> false
    | Plain f', Sharp f -> false;;

let rec eq_shpa _A = ({eq = eq_shp _A} : 'a shp eq);;

type rat = Fract of Big_int.big_int * Big_int.big_int;;

type domain = Natural | Arctic | Int_mat of nat * nat | Arc_mat of nat |
  Rational of rat | Rat_mat of rat * nat * nat;;

type xmldoc = Xmldoc of char list * xml;;

let ord_int =
  ({less_eq = Big_int.le_big_int; less = Big_int.lt_big_int} :
    Big_int.big_int ord);;

let rec foldr
  f x1 a = match f, x1, a with f, [], a -> a
    | f, x :: xs, a -> f x (foldr f xs a);;

let ord_nat = ({less_eq = less_eq_nat; less = less_nat} : nat ord);;

let rec the (Some x) = x;;

let rec eq_list _A
  x0 x1 = match x0, x1 with [], [] -> true
    | a :: lista, a' :: list' -> eqa _A a a' && eq_list _A lista list'
    | [], a' :: list' -> false
    | a' :: list', [] -> false;;

let rec eq_scg _A _B
  x0 x1 = match x0, x1 with Null, Null -> true
    | Scg (pp1, pp2, list1, list2), Scg (pp1', pp2', list1', list2') ->
        eqa _A pp1 pp1' &&
          (eqa _A pp2 pp2' &&
            (eq_list (eqc _B _B) list1 list1' &&
              eq_list (eqc _B _B) list2 list2'))
    | Null, Scg (pp1', pp2', list1', list2') -> false
    | Scg (pp1', pp2', list1', list2'), Null -> false;;

let rec term_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Fun (f, lista) -> f2 f lista
    | f1, f2, Var v -> f1 v;;

let rec suptp _A _B
  s t = (match s with Var _ -> false
          | Fun (_, ts) ->
            member (eq_terma _A _B) t ts ||
              any (fun sa -> suptp _A _B sa t) ts);;

let rec plaina
  = function Var x -> Var x
    | Fun (f, ss) -> Fun (Plain f, map plaina ss);;

let rec sharpa
  = function Var x -> Var x
    | Fun (f, ss) -> Fun (Sharp f, map plaina ss);;

let dquote : char list = ['\034'];;

let rec eq_option _A
  x0 x1 = match x0, x1 with None, None -> true
    | Some a, Some a' -> eqa _A a a'
    | None, Some a' -> false
    | Some a', None -> false;;

let rec eq_lista _A = ({eq = eq_list _A} : ('a list) eq);;

let rec eq_xml
  (Xml (name1, attrs1, children1, text1))
    (Xml (name2, attrs2, children2, text2)) =
    eq_list eq_char name1 name2 &&
      (eq_list (eqc (eq_lista eq_char) (eq_lista eq_char)) attrs1 attrs2 &&
        (list_all2 eq_xml children1 children2 &&
          eq_option (eq_lista eq_char) text1 text2));;

type 'a spL = SimpProj of ('a * nat) list;;

let rec minus_nat
  m n = match m, n with Suc m, Suc n -> minus_nat m n
    | Zero_nat, n -> Zero_nat
    | m, Zero_nat -> m;;

let rec times_nat
  x0 n = match x0, n with Zero_nat, n -> Zero_nat
    | Suc m, n -> plus_nat n (times_nat m n);;

let rec digit2nat
  x = (if ((x : char) = '0') then return Zero_nat
        else (if ((x : char) = '1') then return one_nat
               else (if ((x : char) = '2')
                      then return (nat (Big_int.big_int_of_int 2))
                      else (if ((x : char) = '3')
                             then return (nat (Big_int.big_int_of_int 3))
                             else (if ((x : char) = '4')
                                    then return (nat (Big_int.big_int_of_int 4))
                                    else (if ((x : char) = '5')
   then return (nat (Big_int.big_int_of_int 5))
   else (if ((x : char) = '6') then return (nat (Big_int.big_int_of_int 6))
          else (if ((x : char) = '7')
                 then return (nat (Big_int.big_int_of_int 7))
                 else (if ((x : char) = '8')
                        then return (nat (Big_int.big_int_of_int 8))
                        else (if ((x : char) = '9')
                               then return (nat (Big_int.big_int_of_int 9))
                               else error (append [x]
    [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'd'; 'i'; 'g'; 'i';
      't'])))))))))));;

let rec bind p vq = (match p with Left a -> Left a | Right a -> vq a);;

let rec string2natHelp
  n x1 = match n, x1 with n, [] -> return n
    | n, d :: s ->
        bind (digit2nat d)
          (fun m ->
            string2natHelp
              (plus_nat (times_nat (nat (Big_int.big_int_of_int 10)) n) m) s);;

let rec string2nat
  s = (if null s
        then error ['c'; 'a'; 'n'; 'n'; 'o'; 't'; ' '; 'c'; 'o'; 'n'; 'v'; 'e';
                     'r'; 't'; ' '; 'e'; 'm'; 'p'; 't'; 'y'; ' '; 's'; 't'; 'r';
                     'i'; 'n'; 'g'; ' '; 'i'; 'n'; 't'; 'o'; ' '; 'n'; 'u'; 'm';
                     'b'; 'e'; 'r']
        else string2natHelp Zero_nat s);;

let rec replicate
  xa0 x = match xa0, x with Zero_nat, x -> []
    | Suc n, x -> x :: replicate n x;;

let rec shows_map
  s x1 = match s, x1 with s, [] -> id
    | s, x :: xs -> comp (s x) (shows_map s xs);;

let rec is_none = function Some x -> false
                  | None -> true;;

let rec shows_list_char cs = shows_string cs;;

let rec shows_char c = (fun a -> c :: a);;

let shows_chara =
  ({shows = shows_char; shows_list = shows_list_char} : char shows);;

let rec shows_lista _A l = shows_list _A l;;

let rec shows_attr
  av = comp (shows_lista shows_chara (fst av))
         (comp (shows_string ['='])
           (shows_string (append dquote (append (snd av) dquote))));;

let rec shows_attrs
  asa = shows_map (fun a -> comp (shows_string [' ']) (shows_attr a)) asa;;

let rec shows_XML_indent
  ind i (Xml (n, a, c, t)) =
    comp (shows_string ind)
      (comp (shows_string ['<'])
        (comp (shows_lista shows_chara n)
          (comp (shows_attrs a)
            (if null c && is_none t then shows_string ['/'; '>']
              else comp (shows_string ['>'])
                     (comp (if not (null c)
                             then comp (shows_map
 (shows_XML_indent (append ind (replicate i ' ')) i) c)
                                    (shows_string ind)
                             else id)
                       (comp (if is_none t then id else shows_string (the t))
                         (comp (shows_string ['<'; '/'])
                           (comp (shows_lista shows_chara n)
                             (shows_string ['>'])))))))));;

let rec shows_xml
  xml = shows_XML_indent [newline] (nat (Big_int.big_int_of_int 2)) xml;;

let rec shows_list_aux s xs = shows_list_gen s ['['; ']'] ['['] [','] [']'] xs;;

let rec shows_list_xml xs = shows_list_aux shows_xml xs;;

let shows_xmla =
  ({shows = shows_xml; shows_list = shows_list_xml} : xml shows);;

let rec concat = function [] -> []
                 | x :: xs -> append x (concat xs);;

let rec xml2text
  reqName (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts &&
            (null cs && (not (is_none text) && not (null (the text)))))
      then return (the text)
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'e'; 'x';
                      't'; 'r'; 'a'; 'c'; 't'; ' '; 't'; 'e'; 'x'; 't'; ' ';
                      'f'; 'o'; 'r'; ' '];
                     reqName; [' '; 'f'; 'r'; 'o'; 'm'; ' '; ' ']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml2nat reqName x = bind (xml2text reqName x) string2nat;;

let rec xml3elements
  reqName p1 p2 p3 combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts &&
            (eq_nat (size_list cs) (nat (Big_int.big_int_of_int 3)) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun a ->
               bind (p2 (nth cs one_nat))
                 (fun b ->
                   bind (p3 (nth cs (nat (Big_int.big_int_of_int 2))))
                     (comp return (combi a b))))
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec mapMonad
  uu x1 = match uu, x1 with uu, [] -> return []
    | f, xml :: rest ->
        (match f xml with Left a -> error a
          | Right a ->
            (match mapMonad f rest with Left aa -> error aa
              | Right asa -> return (a :: asa)));;

let rec xmlManyElements
  reqName p combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName && (null atts && is_none text)
      then bind (mapMonad p cs) (comp return combi)
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml2sp
  xml2name =
    xmlManyElements
      ['a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; 'F'; 'i'; 'l'; 't'; 'e'; 'r']
      (xml3elements
        ['a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; 'F'; 'i'; 'l'; 't'; 'e'; 'r';
          'E'; 'n'; 't'; 'r'; 'y']
        xml2name (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
        (xml2nat ['c'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n'; 'g'])
        (fun f _ n -> (f, minus_nat n one_nat)))
      (fun a -> SimpProj a);;

let rec eq_hterm _A () = ({eq = eq_hterma _A} : ('a, 'b) hterm eq)
and eq_hterma _A
  x0 x1 = match x0, x1 with HHole, HHole -> true
    | HFun (f, lista), HFun (f', list') ->
        eqa _A f f' && eq_list (eq_hterm _A ()) lista list'
    | HHole, HFun (f', list') -> false
    | HFun (f', list'), HHole -> false;;

let rec fun_upd _A f a b = (fun x -> (if eqa _A x a then b else f x));;

let rec abs_int
  i = (if Big_int.lt_big_int i (Big_int.big_int_of_int 0)
        then Big_int.minus_big_int i else i);;

let rec sgn_int
  i = (if Big_int.eq_big_int i (Big_int.big_int_of_int 0)
        then (Big_int.big_int_of_int 0)
        else (if Big_int.lt_big_int (Big_int.big_int_of_int 0) i
               then (Big_int.big_int_of_int 1)
               else Big_int.minus_big_int (Big_int.big_int_of_int 1)));;

let rec apsnd f (x, y) = (x, f y);;

let rec divmod_int
  k l = (if Big_int.eq_big_int k (Big_int.big_int_of_int 0)
          then ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 0))
          else (if Big_int.eq_big_int l (Big_int.big_int_of_int 0)
                 then ((Big_int.big_int_of_int 0), k)
                 else apsnd (Big_int.mult_big_int (sgn_int l))
                        (if Big_int.eq_big_int (sgn_int k) (sgn_int l)
                          then (fun k -> fun l -> Big_int.quomod_big_int
                                 (Big_int.abs_big_int k)
                                 (Big_int.abs_big_int l))
                                 k l
                          else let (r, s) =
                                 (fun k -> fun l -> Big_int.quomod_big_int
                                   (Big_int.abs_big_int k)
                                   (Big_int.abs_big_int l))
                                   k l
                                 in
                               (if Big_int.eq_big_int s
                                     (Big_int.big_int_of_int 0)
                                 then (Big_int.minus_big_int r,
(Big_int.big_int_of_int 0))
                                 else (Big_int.sub_big_int
 (Big_int.minus_big_int r) (Big_int.big_int_of_int 1),
Big_int.sub_big_int (abs_int l) s)))));;

let rec mod_int a b = snd (divmod_int a b);;

let rec gcd_int
  k l = abs_int
          (if Big_int.eq_big_int l (Big_int.big_int_of_int 0) then k
            else gcd_int l (mod_int (abs_int k) (abs_int l)));;

let rec implies
  p pa = match p, pa with p, true -> true
    | p, false -> not p
    | true, p -> p
    | false, p -> true;;

let rec shows_paren p = comp (shows_char '(') (comp p (shows_char ')'));;

let rec showsa _A _B
  p = shows_paren
        (comp (shows _A (fst p)) (comp (shows_char ',') (shows _B (snd p))));;

let rec is_var = function Var uu -> true
                 | Fun (uv, uw) -> false;;

let rec shows_nl x = shows_char newline x;;

let rec shows_rule _A _B
  (l, r) =
    comp (shows_term _A _B l)
      (comp (shows_string [' '; '-'; '>'; ' ']) (shows_term _A _B r));;

let rec shows_rules _A _B
  trs = comp (shows_list_gen (shows_rule _A _B) [] [] [newline] [] trs)
          shows_nl;;

let rec shows_p _A _B
  p = comp (shows_string
             ['d'; 'e'; 'p'; 'e'; 'n'; 'd'; 'e'; 'n'; 'c'; 'y'; ' '; 'p'; 'a';
               'i'; 'r'; 's'; ':'])
        (comp shows_nl (shows_rules _A _B p));;

let rec lookup _A
  uu x1 = match uu, x1 with uu, [] -> None
    | a, (a', b) :: xs -> (if eqa _A a a' then Some b else lookup _A a xs);;

let rec alter1a (Alter1 x) = x;;

let rec alter2a (Alter2 x) = x;;

let rec alter3a (Alter3 x) = x;;

let rec alter4a (Alter4 x) = x;;

let rec alter5a (Alter5 x) = x;;

let rec alter6a (Alter6 x) = x;;

let rec parseError
  s allowed =
    error (append
            ['e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '; 's'; 'y'; 'm'; 'b';
              'o'; 'l'; 's'; ' ']
            (append allowed
              (append [' '; 'b'; 'u'; 't'; ' '; 'f'; 'o'; 'u'; 'n'; 'd'; ' ']
                (take (nat (Big_int.big_int_of_int 30)) s))));;

let rec consume2
  x0 allowed s = match x0, allowed, s with
    [], allowed, s ->
      parseError s (concat_map (fun sa -> append sa [',']) allowed)
    | t :: asa, allowed, s ->
        (if eq_list eq_char (map snd (zip t s)) t
          then return (t, trim (drop (size_list t) s))
          else consume2 asa allowed s);;

let rec consume allowed s = consume2 allowed allowed s;;

let letters : char list
  = ['a'; 'b'; 'c'; 'd'; 'e'; 'f'; 'g'; 'h'; 'i'; 'j'; 'k'; 'l'; 'm'; 'n'; 'o';
      'p'; 'q'; 'r'; 's'; 't'; 'u'; 'v'; 'w'; 'x'; 'y'; 'z'; 'A'; 'B'; 'C'; 'D';
      'E'; 'F'; 'G'; 'H'; 'I'; 'J'; 'K'; 'L'; 'M'; 'N'; 'O'; 'P'; 'Q'; 'R'; 'S';
      'T'; 'U'; 'V'; 'W'; 'X'; 'Y'; 'Z'; '_'; '0'; '1'; '2'; '3'; '4'; '5'; '6';
      '7'; '8'; '9'; '&'; ';'; ':'];;

let rec nat2int
  = function Zero_nat -> (Big_int.big_int_of_int 0)
    | Suc n -> Big_int.add_big_int (Big_int.big_int_of_int 1) (nat2int n);;

type af_entry = Collapse of nat | AFList of nat list;;

let rec xmlChange p f x = bind (p x) (fun a -> return (f a));;

let rec xml2position
  x = xmlChange (xml2nat ['p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'])
        (fun n -> minus_nat n one_nat) x;;

let rec xmlGetName (Xml (name, uu, uv, uw)) = name;;

let rec xmlOptions
  ps x =
    (match lookup (eq_lista eq_char) (xmlGetName x) ps
      with None ->
        error (concat
                [['e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '; 'o'; 'n'; 'e';
                   ' '; 'o'; 'f'; ':'; ' '];
                  concat_map (fun p -> append (fst p) [' ']) ps; [newline];
                  ['b'; 'u'; 't'; ' '; 'f'; 'o'; 'u'; 'n'; 'd']; [newline];
                  show shows_xmla x])
      | Some p -> p x);;

let rec xml2afs
  xml2name =
    xmlManyElements
      ['a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; 'F'; 'i'; 'l'; 't'; 'e'; 'r']
      (xml3elements
        ['a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; 'F'; 'i'; 'l'; 't'; 'e'; 'r';
          'E'; 'n'; 't'; 'r'; 'y']
        xml2name (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
        (xmlOptions
          [(['c'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n'; 'g'],
             xmlChange
               (xml2nat ['c'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n'; 'g'])
               (fun i -> Collapse (minus_nat i one_nat)));
            (['n'; 'o'; 'n'; 'C'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n'; 'g'],
              xmlManyElements
                ['n'; 'o'; 'n'; 'C'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n';
                  'g']
                xml2position (fun a -> AFList a))])
        (fun f a -> (fun b -> ((f, a), b))))
      id;;

let rec string2int
  s = (match s
        with [] ->
          error ['c'; 'a'; 'n'; 'n'; 'o'; 't'; ' '; 'c'; 'o'; 'n'; 'v'; 'e';
                  'r'; 't'; ' '; 'e'; 'm'; 'p'; 't'; 'y'; ' '; 's'; 't'; 'r';
                  'i'; 'n'; 'g'; ' '; 'i'; 'n'; 't'; 'o'; ' '; 'n'; 'u'; 'm';
                  'b'; 'e'; 'r']
        | h :: t ->
          let (pos, ss) = (if ((h : char) = '-') then (false, t) else (true, s))
            in
          bind (string2nat ss)
            (fun n ->
              (if pos || eq_nat n Zero_nat then return (nat2int n)
                else return
                       (Big_int.sub_big_int (Big_int.big_int_of_int 0)
                         (nat2int n)))));;

let rec xml2int reqName x = bind (xml2text reqName x) string2int;;

let rec list2position
  = function [] -> Empty
    | n :: ns -> PCons (n, list2position ns);;

let rec xml2pos
  x = xmlManyElements
        ['p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'; 'I'; 'n'; 'T'; 'e'; 'r'; 'm']
        xml2position list2position x;;

let rec xml2elements
  reqName p1 p2 combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts &&
            (eq_nat (size_list cs) (nat (Big_int.big_int_of_int 2)) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun a -> bind (p2 (nth cs one_nat)) (comp return (combi a)))
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml2rat
  x = xmlOptions
        [(['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'],
           xmlChange (xml2int ['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'])
             (fun xa -> Fract (xa, (Big_int.big_int_of_int 1))));
          (['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'],
            xml2elements ['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l']
              (xml2int ['n'; 'u'; 'm'; 'e'; 'r'; 'a'; 't'; 'o'; 'r'])
              (xml2int ['d'; 'e'; 'n'; 'o'; 'm'; 'i'; 'n'; 'a'; 't'; 'o'; 'r'])
              (fun a b -> Fract (a, b)))]
        x;;

let rec string2bool
  s = (if eq_list eq_char s ['t'; 'r'; 'u'; 'e'] then return true
        else (if eq_list eq_char s ['f'; 'a'; 'l'; 's'; 'e'] then return false
               else error (append
                            ['c'; 'a'; 'n'; 'n'; 'o'; 't'; ' '; 'c'; 'o'; 'n';
                              'v'; 'e'; 'r'; 't'; ' ']
                            (append s
                              [' '; 'i'; 'n'; 't'; 'o'; ' '; 'B'; 'o'; 'o'; 'l';
                                'e'; 'a'; 'n']))));;

let rec xml2bool reqName x = bind (xml2text reqName x) string2bool;;

let rec xml2scg_position
  x = xml2nat ['p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'] x;;

let rec xml1element
  reqName p1 combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts && (eq_nat (size_list cs) one_nat && is_none text))
      then bind (p1 (nth cs Zero_nat)) (comp return combi)
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml1CmanyElements
  reqName p1 p2 combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts && (is_none text && not (null cs)))
      then bind (p1 (hd cs))
             (fun a -> bind (mapMonad (p2 a) (tl cs)) (comp return (combi a)))
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml1manyElements
  reqName p1 p2 = xml1CmanyElements reqName p1 (fun _ -> p2);;

let rec xml2var
  x = bind (xml2text ['v'; 'a'; 'r'] x) (comp return (fun a -> Var a));;

let rec xml2term
  xml2name x =
    xmlOptions
      [(['v'; 'a'; 'r'], xml2var);
        (['f'; 'u'; 'n'; 'a'; 'p'; 'p'],
          xml1manyElements ['f'; 'u'; 'n'; 'a'; 'p'; 'p'] xml2name
            (xml1element ['a'; 'r'; 'g'] (xml2term xml2name) id)
            (fun a b -> Fun (a, b)))]
      x;;

let rec xml2rule
  xml2name =
    xml2elements ['r'; 'u'; 'l'; 'e']
      (xml1element ['l'; 'h'; 's'] (xml2term xml2name) id)
      (xml1element ['r'; 'h'; 's'] (xml2term xml2name) id) (fun a b -> (a, b));;

let rec xml2scg
  xml2name =
    xml1manyElements
      ['s'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'G'; 'r'; 'a'; 'p';
        'h']
      (xml2rule xml2name)
      (xml3elements ['e'; 'd'; 'g'; 'e'] xml2scg_position
        (xml2bool ['s'; 't'; 'r'; 'i'; 'c'; 't']) xml2scg_position
        (fun p s q -> (s, (p, q))))
      (fun lr edges ->
        (lr, (map snd (filter fst edges),
               map snd (filter (fun x -> not (fst x)) edges))));;

let rec vec0I ze n = replicate n ze;;

let rec mat0I ze nr nc = replicate nc (vec0I ze nr);;

let rec vec1I
  ze on n i =
    append (replicate i ze)
      (on :: replicate (minus_nat (minus_nat n one_nat) i) ze);;

let rec mat1I ze on n = map (vec1I ze on n) (upt Zero_nat n);;

let rec supteqp _A _B s t = eq_term _A _B s t || suptp _A _B s t;;

let rec the_var (Var x) = x;;

let rec map_term
  fg x1 = match fg, x1 with fg, Var x -> Var x
    | fg, Fun (f, ts) -> Fun (fg f, map (map_term fg) ts);;

let rec map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr));;

let rec map_trsI fg r = map (map_rule fg) r;;

let rec shows_shp _A
  = function Sharp s -> comp (shows _A s) (shows_string ['#'])
    | Plain s -> shows _A s;;

let rec shows_list_shp _A fs = shows_list_aux (shows_shp _A) fs;;

let rec shows_shpa _A =
  ({shows = shows_shp _A; shows_list = shows_list_shp _A} : 'a shp shows);;

let rec splitterFun
  allowed x1 = match allowed, x1 with
    allowed, c :: s ->
      (if allowed c then let (l, a) = splitterFun allowed s in
                         (c :: l, a)
        else ([], c :: s))
    | uu, [] -> ([], []);;

let rec parseText
  s = let (l, r) = splitterFun (fun c -> not (eq_list eq_char [c] ['<'])) s in
      (if null (trim l) then return (None, r)
        else return (Some (rev (trim (rev (trim l)))), r));;

let rec parseAttribute
  s = bind (consume [dquote] s)
        (fun (_, s1) ->
          let (n, s2) =
            splitterFun (fun c -> not (eq_list eq_char [c] dquote)) s1 in
          bind (consume [dquote] s2) (fun (_, s3) -> return (n, s3)));;

let rec splitter allowed s = splitterFun (fun c -> member eq_char c allowed) s;;

let rec parseName
  s = let (l, r) = splitter letters s in
      (if null l then parseError s letters else return (l, trim r));;

let rec parseAttributes
  = function [] -> return ([], [])
    | c :: s ->
        (if member eq_char c ['/'; '>'] then return ([], c :: s)
          else bind (parseName (c :: s))
                 (fun (n, s1) ->
                   bind (consume [['=']] s1)
                     (fun (_, s2) ->
                       bind (parseAttribute s2)
                         (fun (a, s3) ->
                           bind (parseAttributes s3)
                             (fun (nas, s4) -> return ((n, a) :: nas, s4))))));;

let rec parseXMLList
  s = (if null s ||
            eq_list eq_char (take (nat (Big_int.big_int_of_int 2)) s) ['<'; '/']
        then return ([], s)
        else bind (consume [['<']] s)
               (fun (_, s1) ->
                 bind (parseName s1)
                   (fun (n, s2) ->
                     bind (parseAttributes s2)
                       (fun (nas, s3) ->
                         bind (consume [['/'; '>']; ['>']] s3)
                           (fun (clo, s4) ->
                             (if eq_list eq_char clo ['/'; '>']
                               then bind (parseXMLList s4)
                                      (fun (li, s5) ->
return (Xml (n, nas, [], None) :: li, s5))
                               else bind (parseText s4)
                                      (fun (t, s5) ->
bind (parseXMLList s5)
  (fun (chi, s6) ->
    bind (consume [append ['<'; '/'] (append n ['>'])] s6)
      (fun (_, s7) ->
        bind (parseXMLList s7)
          (fun (li, s8) -> return (Xml (n, nas, chi, t) :: li, s8)))))))))));;

let rec parseXML
  s = bind (consume [['<']] s)
        (fun (_, s1) ->
          bind (parseName s1)
            (fun (n, s2) ->
              bind (parseAttributes s2)
                (fun (nas, s3) ->
                  bind (consume [['/'; '>']; ['>']] s3)
                    (fun (clo, s4) ->
                      (if eq_list eq_char clo ['/'; '>']
                        then return (Xml (n, nas, [], None), trim s4)
                        else bind (parseText s4)
                               (fun (t, s5) ->
                                 bind (parseXMLList s5)
                                   (fun (chi, s6) ->
                                     bind (consume
    [append ['<'; '/'] (append n ['>'])] s6)
                                       (fun (_, s7) ->
 return (Xml (n, nas, chi, t), s7)))))))));;

let rec scanUpto
  enda x1 = match enda, x1 with
    enda, [] ->
      error (append
              ['d'; 'i'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'i'; 'n'; 'd'; ' ';
                'r'; 'e'; 'q'; 'u'; 'i'; 'r'; 'e'; 'd'; ' '; 't'; 'o'; 'k'; 'e';
                'n'; ' ']
              enda)
    | enda, c :: s ->
        (if eq_list eq_char (map snd (zip enda (c :: s))) enda
          then return
                 (take (size_list enda) (c :: s),
                   trim (drop (size_list enda) (c :: s)))
          else bind (scanUpto enda s)
                 (fun (res, rest) -> return (c :: res, rest)));;

let rec xml2poly
  xml2coeff x =
    xml1element ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l']
      (xmlOptions
        [(['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't'],
           xml1element ['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't']
             xml2coeff (fun a -> PNum a));
          (['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'],
            xmlChange (xml2nat ['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'])
              (fun n -> PVar (minus_nat n one_nat)));
          (['s'; 'u'; 'm'],
            xmlManyElements ['s'; 'u'; 'm'] (xml2poly xml2coeff)
              (fun a -> PSum a));
          (['p'; 'r'; 'o'; 'd'; 'u'; 'c'; 't'],
            xmlManyElements ['p'; 'r'; 'o'; 'd'; 'u'; 'c'; 't']
              (xml2poly xml2coeff) (fun a -> PMult a))])
      id x;;

let rec xml_term _A _B
  = function Var x -> Xml (['v'; 'a'; 'r'], [], [], Some (show _B x))
    | Fun (f, ts) ->
        Xml (['f'; 'u'; 'n'; 'a'; 'p'; 'p'], [],
              Xml (['n'; 'a'; 'm'; 'e'], [], [], Some (show _A f)) ::
                map (fun t ->
                      Xml (['a'; 'r'; 'g'], [], [xml_term _A _B t], None))
                  ts,
              None);;

let rec isLeft = function Left uu -> true
                 | Right uv -> false;;

let rec distinct _A
  = function [] -> true
    | x :: xs -> not (member _A x xs) && distinct _A xs;;

let rec get_list (_A1, _A2) (_B1, _B2)
  x0 u r = match x0, u, r with
    [], u, r ->
      error (shows_string
              ['L'; 'o'; 'o'; 'p'; ' '; 'm'; 'u'; 's'; 't'; ' '; 'b'; 'e'; ' ';
                'n'; 'o'; 'n'; '-'; 'e'; 'm'; 'p'; 't'; 'y'])
    | [s], u, r -> bind (get (_A1, _A2) (_B1, _B2) s u r) (fun y -> return [y])
    | s :: t :: ts, u, r ->
        bind (get (_A1, _A2) (_B1, _B2) s t r)
          (fun d ->
            bind (get_list (_A1, _A2) (_B1, _B2) (t :: ts) u r)
              (fun y -> return (d :: y)));;

let rec get_loop (_A1, _A2) (_B1, _B2)
  x0 uu uv uw = match x0, uu, uv, uw with
    [], uu, uv, uw ->
      error (shows_string
              ['L'; 'o'; 'o'; 'p'; ' '; 'm'; 'u'; 's'; 't'; ' '; 'b'; 'e'; ' ';
                'n'; 'o'; 'n'; '-'; 'e'; 'm'; 'p'; 't'; 'y'])
    | t :: ts, d, tau, r ->
        get_list (_A1, _A2) (_B1, _B2) (t :: ts)
          (ctxt_apply d (sub_apply t tau)) r;;

let rec wf_vec n x = eq_nat (size_list x) n;;

let rec wf_mat nr nc m = eq_nat (size_list m) nc && forall (wf_vec nr) m;;

type 'a preorder = {ord_preorder : 'a ord};;

type 'a bot = {preorder_bot : 'a preorder; bot : 'a};;
let bot _A = _A.bot;;

let rec maxa less_eq a b = (if less_eq a b then b else a);;

let rec max _A = maxa (less_eq _A);;

let rec mina less_eq a b = (if less_eq a b then a else b);;

let rec min _A = mina (less_eq _A);;

let rec insert_scg _A _B
  g gs =
    (if eq_scg _A _B g Null then None
      else (if list_ex (fun g' -> subsumes _A _B g' g) gs then None
             else Some (g :: filter (fun g' -> not (subsumes _A _B g g'))
                               gs)));;

let rec mk_trancl _A _B
  conn x1 s = match conn, x1, s with conn, [], s -> s
    | conn, g :: gs, s ->
        (match insert_scg _A _B g s with None -> mk_trancl _A _B conn gs s
          | Some s' ->
            mk_trancl _A _B conn
              (append (map (scg_comp _B conn g) s')
                (append (map (fun f -> scg_comp _B conn f g) s') gs))
              s');;

let rec check_SCT _A _B
  conn gs =
    let a = mk_trancl _A _B conn gs [] in
    list_all (sagiv _A _B conn) a;;

let rec div_int a b = fst (divmod_int a b);;

let rec normalize
  (Fract (a, b)) =
    (if Big_int.lt_big_int (Big_int.big_int_of_int 0) b
      then let g = gcd_int a b in
           (div_int a g, div_int b g)
      else (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
             then ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1))
             else let g = Big_int.minus_big_int (gcd_int a b) in
                  (div_int a g, div_int b g)));;

let rec divmod_nat
  m n = (if eq_nat n Zero_nat || less_nat m n then (Zero_nat, m)
          else let (q, a) = divmod_nat (minus_nat m n) n in
               (Suc q, a));;

let rec mod_nat m n = snd (divmod_nat m n);;

let rec div_nat m n = fst (divmod_nat m n);;

let rec digit2string
  n = (if eq_nat n Zero_nat then ['0']
        else (if eq_nat n one_nat then ['1']
               else (if eq_nat n (nat (Big_int.big_int_of_int 2)) then ['2']
                      else (if eq_nat n (nat (Big_int.big_int_of_int 3))
                             then ['3']
                             else (if eq_nat n (nat (Big_int.big_int_of_int 4))
                                    then ['4']
                                    else (if eq_nat n
       (nat (Big_int.big_int_of_int 5))
   then ['5']
   else (if eq_nat n (nat (Big_int.big_int_of_int 6)) then ['6']
          else (if eq_nat n (nat (Big_int.big_int_of_int 7)) then ['7']
                 else (if eq_nat n (nat (Big_int.big_int_of_int 8)) then ['8']
                        else (if eq_nat n (nat (Big_int.big_int_of_int 9))
                               then ['9']
                               else ['n'; 'o'; 't'; ' '; 'a'; ' '; 'd'; 'i';
                                      'g'; 'i'; 't']))))))))));;

let rec shows_nat
  n = (if less_nat n (nat (Big_int.big_int_of_int 10))
        then shows_string (digit2string n)
        else comp (shows_nat (div_nat n (nat (Big_int.big_int_of_int 10))))
               (shows_string
                 (digit2string
                   (mod_nat n (nat (Big_int.big_int_of_int 10))))));;

let rec shows_int
  i = (if Big_int.lt_big_int i (Big_int.big_int_of_int 0)
        then comp (shows_char '-')
               (shows_nat
                 (nat (Big_int.sub_big_int (Big_int.big_int_of_int 0) i)))
        else shows_nat (nat i));;

let rec show_rat
  x = let (den, num) = normalize x in
      (if Big_int.eq_big_int num (Big_int.big_int_of_int 1)
        then shows_int den []
        else comp (comp (shows_int den) (shows_char '/')) (shows_int num) []);;

let rec shows_list_int is = shows_list_aux shows_int is;;

let shows_inta =
  ({shows = shows_int; shows_list = shows_list_int} : Big_int.big_int shows);;

let rec shows_list_nat ns = shows_list_aux shows_nat ns;;

let shows_nata =
  ({shows = shows_nat; shows_list = shows_list_nat} : nat shows);;

let rec shows_rat x = shows_string (show_rat x);;

let rec shows_list_rat rs = shows_list_aux shows_rat rs;;

let shows_rata =
  ({shows = shows_rat; shows_list = shows_list_rat} : rat shows);;

let rec defs_list _A
  = function [] -> []
    | (Fun (f, uu), uv) :: ts ->
        let fs = defs_list _A ts in
        (if member _A f fs then fs else f :: fs)
    | (Var uw, ux) :: ts -> defs_list _A ts;;

let rec shows_trs _A _B
  r = comp (shows_string
             ['r'; 'e'; 'w'; 'r'; 'i'; 't'; 'e'; ' '; 's'; 'y'; 's'; 't'; 'e';
               'm'; ':'])
        (comp shows_nl (shows_rules _A _B r));;

let rec shows_dpp _A _B p r = comp (shows_p _A _B p) (shows_trs _A _B r);;

let rec rev_append
  x0 ys = match x0, ys with [], ys -> ys
    | x :: xs, ys -> rev_append xs (x :: ys);;

let rec flat_map_aux
  f x1 ys = match f, x1, ys with f, [], ys -> rev ys
    | f, x :: xs, ys -> flat_map_aux f xs (rev_append (f x) ys);;

let rec flat_map f xs = flat_map_aux f xs [];;

let rec max_list
  = function [] -> Zero_nat
    | x :: xs -> max ord_nat x (max_list xs);;

let rec fitLength
  uu x1 uv = match uu, x1, uv with uu, Zero_nat, uv -> []
    | a, Suc v, [] -> replicate (Suc v) a
    | a, Suc n, b :: asa -> b :: fitLength a n asa;;

let rec sum_lpoly
  uu x1 ys = match uu, x1, ys with uu, [], ys -> ys
    | uv, v :: va, [] -> v :: va
    | pl, x :: xs, y :: ys -> pl x y :: sum_lpoly pl xs ys;;

let rec xml2label
  xml2name =
    xmlOptions
      [(['n'; 'u'; 'm'; 'b'; 'e'; 'r'; 'L'; 'a'; 'b'; 'e'; 'l'],
         xmlManyElements ['n'; 'u'; 'm'; 'b'; 'e'; 'r'; 'L'; 'a'; 'b'; 'e'; 'l']
           (xml2nat ['n'; 'u'; 'm'; 'b'; 'e'; 'r']) (fun a -> Left a));
        (['s'; 'y'; 'm'; 'b'; 'o'; 'l'; 'L'; 'a'; 'b'; 'e'; 'l'],
          xmlManyElements
            ['s'; 'y'; 'm'; 'b'; 'o'; 'l'; 'L'; 'a'; 'b'; 'e'; 'l'] xml2name
            (fun a -> Right a))];;

let rec xml2rules
  xml2name = xmlManyElements ['r'; 'u'; 'l'; 'e'; 's'] (xml2rule xml2name) id;;

let rec xml_rules _A _B
  rls = Xml (['r'; 'u'; 'l'; 'e'; 's'], [],
              map (fun (l, r) ->
                    Xml (['r'; 'u'; 'l'; 'e'], [],
                          [Xml (['l'; 'h'; 's'], [], [xml_term _A _B l], None);
                            Xml (['r'; 'h'; 's'], [], [xml_term _A _B r],
                                  None)],
                          None))
                rls,
              None);;

let rec partition
  p x1 = match p, x1 with p, [] -> ([], [])
    | p, x :: xs ->
        let (yes, no) = partition p xs in
        (if p x then (x :: yes, no) else (yes, x :: no));;

let rec removeAll _A
  x xa1 = match x, xa1 with x, [] -> []
    | x, y :: xs ->
        (if eqa _A x y then removeAll _A x xs else y :: removeAll _A x xs);;

let rec rewrites_to (_A1, _A2) (_B1, _B2)
  x0 t r = match x0, t, r with
    [(s, (c, (sigma, rule)))], t, r ->
      rewrites (_A1, _A2) (_B1, _B2) (s, t) c sigma rule r
    | (s, (c, (sigma, rule))) :: (t, (d, (tau, rule'))) :: xs, u, r ->
        bind (rewrites (_A1, _A2) (_B1, _B2) (s, t) c sigma rule r)
          (fun _ ->
            rewrites_to (_A1, _A2) (_B1, _B2) ((t, (d, (tau, rule'))) :: xs) u
              r);;

let rec check_loop_d (_A1, _A2) (_B1, _B2)
  x0 uu uv uw = match x0, uu, uv, uw with
    [], uu, uv, uw ->
      error (shows_string
              ['T'; 'h'; 'e'; 'r'; 'e'; ' '; 'm'; 'u'; 's'; 't'; ' '; 'b'; 'e';
                ' '; 'a'; 't'; ' '; 'l'; 'e'; 'a'; 's'; 't'; ' '; 'o'; 'n'; 'e';
                ' '; 'r'; 'e'; 'w'; 'r'; 'i'; 't'; 'e'; ' '; 's'; 't'; 'e';
                'p'])
    | v :: va, d, tau, r ->
        rewrites_to (_A1, _A2) (_B1, _B2) (v :: va)
          (ctxt_apply d (sub_apply (fst (hd (v :: va))) tau)) r;;

let rec check_loop (_A1, _A2) (_B1, _B2)
  (ts, (c, sigma)) r =
    bind (get_loop (_A1, _A2) (_B1, _B2) ts c sigma r)
      (fun ls -> check_loop_d (_A1, _A2) (_B1, _B2) ls c sigma r);;

let rec catch p eq = (match p with Left a -> eq a | Right a -> Right a);;

let rec checkLoop (_A1, _A2) (_B1, _B2)
  r (ts, (c, sub)) =
    catch (match subst_safe _B1 _A1 [] sub (Sub (fun a -> Var a))
            with None ->
              error (shows_string
                      ['s'; 'u'; 'b'; 's'; 't'; 'i'; 't'; 'u'; 't'; 'i'; 'o';
                        'n'; ' '; 'n'; 'o'; 't'; ' '; 'u'; 'n'; 'i'; 'q'; 'u';
                        'e'])
            | Some sigma -> check_loop (_A1, _A2) (_B1, _B2) (ts, (c, sigma)) r)
      (fun x ->
        error (comp (shows_string
                      ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i'; 'n'; ' '; 'c'; 'e';
                        'r'; 't'; 'i'; 'f'; 'i'; 'c'; 'a'; 't'; 'i'; 'o'; 'n';
                        ' '; 'o'; 'f'; ' '; 'l'; 'o'; 'o'; 'p'])
                (comp shows_nl x)));;

let rec vec_geI ge v w = forall (fun (a, b) -> ge a b) (zip v w);;

let rec mat_geI ge m1 m2 = forall (fun (a, b) -> vec_geI ge a b) (zip m1 m2);;

let rec sub_vec x = take x;;

let rec sub_mat nr nc m = map (sub_vec nr) (take nc m);;

let rec vec_pre_gtI gt v w = any (fun (a, b) -> gt a b) (zip v w);;

let rec mat_pre_gtI
  gt m1 m2 = any (fun (a, b) -> vec_pre_gtI gt a b) (zip m1 m2);;

let rec mat_gtI
  ge gt sd m1 m2 =
    mat_geI ge m1 m2 && mat_pre_gtI gt (sub_mat sd sd m1) (sub_mat sd sd m2);;

let rec lpo_strict _A (_B1, _B2)
  pr n =
    (fun (s, t) ->
      checkB (fst (lpo _B1 pr n s t))
        (comp (shows_string
                ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o'; 'r';
                  'i'; 'e'; 'n'; 't'; ' '])
          (comp (shows_term _A _B2 s)
            (comp (shows_string [' '; '>'; 'L'; 'P'; 'O'; ' '])
              (comp (shows_term _A _B2 t) shows_nl)))));;

let rec eq_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
      then Big_int.eq_big_int c (Big_int.big_int_of_int 0) ||
             Big_int.eq_big_int d (Big_int.big_int_of_int 0)
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0)
             then Big_int.eq_big_int a (Big_int.big_int_of_int 0) ||
                    Big_int.eq_big_int b (Big_int.big_int_of_int 0)
             else Big_int.eq_big_int (Big_int.mult_big_int a d)
                    (Big_int.mult_big_int b c)));;

let eq_rata = ({eq = eq_rat} : rat eq);;

let rec shows_list_list _A ls = shows_list_aux (shows_lista _A) ls;;

let rec shows_listb _A =
  ({shows = shows_lista _A; shows_list = shows_list_list _A} :
    ('a list) shows);;

let rec ctxt_case
  f1 f2 x2 = match f1, f2, x2 with
    f1, f2, More (f, list1, ctxt, list2) -> f2 f list1 ctxt list2
    | f1, f2, Hole -> f1;;

let rec shows_list_term _A _B is = shows_list_aux (shows_term _A _B) is;;

let rec shows_terma _A _B =
  ({shows = shows_term _A _B; shows_list = shows_list_term _A _B} :
    ('a, 'b) term shows);;

let rec list_inter _A
  x0 bs = match x0, bs with [], bs -> []
    | a :: asa, bs ->
        (if member _A a bs then a :: list_inter _A asa bs
          else list_inter _A asa bs);;

let rec lpo_nstrict _A (_B1, _B2)
  pr n =
    (fun (s, t) ->
      checkB (snd (lpo _B1 pr n s t))
        (comp (shows_string
                ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o'; 'r';
                  'i'; 'e'; 'n'; 't'; ' '])
          (comp (shows_term _A _B2 s)
            (comp (shows_string [' '; '>'; '='; 'L'; 'P'; 'O'; ' '])
              (comp (shows_term _A _B2 t) shows_nl)))));;

let rec less_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
      then Big_int.lt_big_int (Big_int.big_int_of_int 0)
             (Big_int.mult_big_int (sgn_int c) (sgn_int d))
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0)
             then Big_int.lt_big_int
                    (Big_int.mult_big_int (sgn_int a) (sgn_int b))
                    (Big_int.big_int_of_int 0)
             else Big_int.lt_big_int
                    (Big_int.mult_big_int (Big_int.mult_big_int a (abs_int d))
                      (sgn_int b))
                    (Big_int.mult_big_int (Big_int.mult_big_int c (abs_int b))
                      (sgn_int d))));;

let rec less_eq_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
      then Big_int.le_big_int (Big_int.big_int_of_int 0)
             (Big_int.mult_big_int (sgn_int c) (sgn_int d))
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0)
             then Big_int.le_big_int
                    (Big_int.mult_big_int (sgn_int a) (sgn_int b))
                    (Big_int.big_int_of_int 0)
             else Big_int.le_big_int
                    (Big_int.mult_big_int (Big_int.mult_big_int a (abs_int d))
                      (sgn_int b))
                    (Big_int.mult_big_int (Big_int.mult_big_int c (abs_int b))
                      (sgn_int d))));;

let ord_rat = ({less_eq = less_eq_rat; less = less_rat} : rat ord);;

type arctic = MinInfty | Num_arc of Big_int.big_int;;

type 'a redpair = Nat_carrier of (('a * nat) * (nat * nat list)) list |
  Int_carrier of (('a * nat) * (Big_int.big_int * Big_int.big_int list)) list |
  Arctic_carrier of (('a * nat) * (arctic * arctic list)) list |
  Int_mat_carrier of
    nat * nat *
      (('a * nat) *
        ((Big_int.big_int list) list * ((Big_int.big_int list) list) list)) list
  | Arc_mat_carrier of
      nat * (('a * nat) * ((arctic list) list * ((arctic list) list) list)) list
  | Lpo of (('a * nat) * nat) list * (('a * nat) * af_entry) list |
  Rat_carrier of rat * (('a * nat) * (rat * rat list)) list |
  Rat_mat_carrier of
    rat * nat * nat *
      (('a * nat) * ((rat list) list * ((rat list) list) list)) list;;

let rec list_union _A
  x0 ys = match x0, ys with [], ys -> ys
    | x :: xs, ys ->
        let zs = list_union _A xs ys in
        (if member _A x zs then zs else x :: zs);;

let rec domain_case
  f1 f2 f3 f4 f5 f6 x6 = match f1, f2, f3, f4, f5, f6, x6 with
    f1, f2, f3, f4, f5, f6, Rat_mat (rat, nat1, nat2) -> f6 rat nat1 nat2
    | f1, f2, f3, f4, f5, f6, Rational rat -> f5 rat
    | f1, f2, f3, f4, f5, f6, Arc_mat nat -> f4 nat
    | f1, f2, f3, f4, f5, f6, Int_mat (nat1, nat2) -> f3 nat1 nat2
    | f1, f2, f3, f4, f5, f6, Arctic -> f2
    | f1, f2, f3, f4, f5, f6, Natural -> f1;;

type ('a, 'b) lab = Lab of ('a, 'b) lab * 'b |
  FunLab of ('a, 'b) lab * ('a, 'b) lab list | UnLab of 'a;;

let rec id_labeling uu cs = Left cs;;

let rec remove_comments_help
  uu x1 = match uu, x1 with
    false, c :: cs ->
      (if ((c : char) = '<') &&
            eq_list eq_char (take (nat (Big_int.big_int_of_int 3)) cs)
              ['!'; '-'; '-']
        then remove_comments_help true (tl cs)
        else c :: remove_comments_help false cs)
    | true, c :: cs ->
        (if ((c : char) = '-') &&
              eq_list eq_char (take (nat (Big_int.big_int_of_int 2)) cs)
                ['-'; '>']
          then remove_comments_help false
                 (drop (nat (Big_int.big_int_of_int 2)) cs)
          else remove_comments_help true cs)
    | uu, [] -> [];;

let rec remove_comments x = remove_comments_help false x;;

let rec stripXMLHeader
  s = (if eq_list eq_char (take (nat (Big_int.big_int_of_int 2)) (trim s))
            ['<'; '?']
        then bind (scanUpto ['?'; '>'] s)
               (fun (header1, s1) ->
                 bind (stripXMLHeader s1)
                   (fun (header2, s2) ->
                     return (append header1 header2, trim s2)))
        else return ([], trim s));;

let rec debug i t x = x;;

let rec parseXMLDoc
  s = debug ['0']
        ['p'; 'a'; 'r'; 's'; 'i'; 'n'; 'g'; ' '; 't'; 'x'; 't'; ' '; 't'; 'o';
          ' '; 'x'; 'm'; 'l']
        (bind (stripXMLHeader (remove_comments s))
          (fun (header, rest) ->
            bind (parseXML rest)
              (fun (xml, enda) ->
                (if null enda then return (Xmldoc (header, xml))
                  else error (append
                               ['t'; 'h'; 'e'; 'r'; 'e'; ' '; 'i'; 's'; ' ';
                                 't'; 'e'; 'x'; 't'; ' '; 'a'; 'f'; 't'; 'e';
                                 'r'; ' '; 'r'; 'o'; 'o'; 't'; ' '; 'e'; 'l';
                                 'e'; 'm'; 'e'; 'n'; 't'; ':'; ' ']
                               (newline :: enda))))));;

let rec xml4elements
  reqName p1 p2 p3 p4 combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts &&
            (eq_nat (size_list cs) (nat (Big_int.big_int_of_int 4)) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun a ->
               bind (p2 (nth cs one_nat))
                 (fun b ->
                   bind (p3 (nth cs (nat (Big_int.big_int_of_int 2))))
                     (fun c ->
                       bind (p4 (nth cs (nat (Big_int.big_int_of_int 3))))
                         (comp return (combi a b c)))))
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml0elements
  reqName combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName && (null atts && (null cs && is_none text))
      then return combi
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml2context
  xml2name x =
    xmlOptions
      [(['b'; 'o'; 'x'], xml0elements ['b'; 'o'; 'x'] Hole);
        (['f'; 'u'; 'n'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'],
          xml4elements ['f'; 'u'; 'n'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't']
            xml2name
            (xmlManyElements ['b'; 'e'; 'f'; 'o'; 'r'; 'e'] (xml2term xml2name)
              id)
            (xml2context xml2name)
            (xmlManyElements ['a'; 'f'; 't'; 'e'; 'r'] (xml2term xml2name) id)
            (fun a b c d -> More (a, b, c, d)))]
      x;;

type arithFun = Arg of nat | Const of nat | Sum of arithFun list |
  Max of arithFun list | Min of arithFun list | Prod of arithFun list;;

type ('a, 'b) sl_inter =
  SL_Inter of nat * ((('a, 'b) lab * nat) * arithFun) list;;

type ('a, 'b, 'c) dPProof = PisEmpty |
  RedPairProc of
    ('a, 'b) lab redpair *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) dPProof
  | RedPairUrProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | DepGraphProc of
      (('a, 'b, 'c) dPProof option *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) list
  | MonoRedPairProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | MonoRedPairUrProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | UnlabProc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SemlabProc of
      ('a, 'b) sl_inter *
        (('a, 'b) lab -> nat list -> ('b, (('a, 'b) lab list)) either) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | RootlabProc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SpscProc of
      ('a, 'b) lab spL *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SizeChangeSubTerm of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((nat * nat) list * (nat * nat) list)) list
  | SizeChangeRedPair of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
          ((nat * nat) list * (nat * nat) list)) list;;

let rec xml2arith_fun
  xml = xml1element
          ['a'; 'r'; 'i'; 't'; 'h'; 'F'; 'u'; 'n'; 'c'; 't'; 'i'; 'o'; 'n']
          (xmlOptions
            [(['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'],
               xmlChange (xml2nat ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'])
                 (fun a -> Const a));
              (['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'],
                xmlChange (xml2nat ['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'])
                  (fun n -> Arg (minus_nat n one_nat)));
              (['s'; 'u'; 'm'],
                xmlManyElements ['s'; 'u'; 'm'] xml2arith_fun (fun a -> Sum a));
              (['p'; 'r'; 'o'; 'd'; 'u'; 'c'; 't'],
                xmlManyElements ['p'; 'r'; 'o'; 'd'; 'u'; 'c'; 't']
                  xml2arith_fun (fun a -> Prod a));
              (['m'; 'i'; 'n'],
                xmlManyElements ['m'; 'i'; 'n'] xml2arith_fun (fun a -> Min a));
              (['m'; 'a'; 'x'],
                xmlManyElements ['m'; 'a'; 'x'] xml2arith_fun
                  (fun a -> Max a))])
          id xml;;

let rec xml2sl_inter
  xml2name =
    xml1element ['m'; 'o'; 'd'; 'e'; 'l']
      (xmlOptions
        [(['f'; 'i'; 'n'; 'i'; 't'; 'e'; 'M'; 'o'; 'd'; 'e'; 'l'],
           xml1manyElements
             ['f'; 'i'; 'n'; 'i'; 't'; 'e'; 'M'; 'o'; 'd'; 'e'; 'l']
             (xmlChange
               (xml2nat ['c'; 'a'; 'r'; 'r'; 'i'; 'e'; 'r'; 'S'; 'i'; 'z'; 'e'])
               (fun n -> minus_nat n one_nat))
             (xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't']
               xml2name (xml2nat ['a'; 'r'; 'i'; 't'; 'y']) xml2arith_fun
               (fun f a -> (fun b -> ((f, a), b))))
             (fun n xs -> Some (SL_Inter (n, xs))));
          (['r'; 'o'; 'o'; 't'; 'L'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'],
            xml0elements
              ['r'; 'o'; 'o'; 't'; 'L'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g']
              None)])
      id;;

let rec xml5elements
  reqName p1 p2 p3 p4 p5 combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts &&
            (eq_nat (size_list cs) (nat (Big_int.big_int_of_int 5)) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun a ->
               bind (p2 (nth cs one_nat))
                 (fun b ->
                   bind (p3 (nth cs (nat (Big_int.big_int_of_int 2))))
                     (fun c ->
                       bind (p4 (nth cs (nat (Big_int.big_int_of_int 3))))
                         (fun d ->
                           bind (p5 (nth cs (nat (Big_int.big_int_of_int 4))))
                             (comp return (combi a b c d))))))
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml2precedence
  xml2name =
    xmlManyElements
      ['s'; 't'; 'a'; 't'; 'u'; 's'; 'P'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n';
        'c'; 'e']
      (xml4elements
        ['s'; 't'; 'a'; 't'; 'u'; 's'; 'P'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n';
          'c'; 'e'; 'E'; 'n'; 't'; 'r'; 'y']
        xml2name (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
        (xml2nat ['p'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n'; 'c'; 'e'])
        (xml0elements ['l'; 'e'; 'x'] ()) (fun f a p _ -> ((f, a), p)))
      id;;

let rec xml1to2elements
  reqName p1 p2 combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts &&
            (less_eq_nat one_nat (size_list cs) &&
              (less_eq_nat (size_list cs) (nat (Big_int.big_int_of_int 2)) &&
                is_none text)))
      then bind (p1 (nth cs Zero_nat))
             (fun a ->
               (if eq_nat (size_list cs) (nat (Big_int.big_int_of_int 2))
                 then bind (p2 (nth cs one_nat))
                        (fun b -> return (combi a (Some b)))
                 else return (combi a None)))
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml2path_order
  xml2name =
    xml1to2elements ['p'; 'a'; 't'; 'h'; 'O'; 'r'; 'd'; 'e'; 'r']
      (xml2precedence xml2name) (xml2afs xml2name)
      (fun pr a ->
        (match a with None -> Lpo (pr, []) | Some aa -> Lpo (pr, aa)));;

let rat_name : char list
  = ['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's'; ' '; '('; 'w'; 'i'; 't'; 'h';
      ' '; 'n'; 'e'; 'g'; 'a'; 't'; 'i'; 'v'; 'e'; ' '; 'c'; 'o'; 'n'; 's'; 't';
      'a'; 'n'; 't'; 's'; ')'];;

let rec fract_norm
  a b = (if Big_int.eq_big_int a (Big_int.big_int_of_int 0) ||
              Big_int.eq_big_int b (Big_int.big_int_of_int 0)
          then Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1))
          else let c = gcd_int a b in
               (if Big_int.lt_big_int (Big_int.big_int_of_int 0) b
                 then Fract (div_int a c, div_int b c)
                 else Fract (Big_int.minus_big_int (div_int a c),
                              Big_int.minus_big_int (div_int b c))));;

let rec minus_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
      then Fract (Big_int.minus_big_int c, d)
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0) then Fract (a, b)
             else fract_norm
                    (Big_int.sub_big_int (Big_int.mult_big_int a d)
                      (Big_int.mult_big_int c b))
                    (Big_int.mult_big_int b d)));;

let rec rat_gt
  delta x y =
    less_eq_rat
      (if less_rat
            (Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1)))
            delta
        then min ord_rat delta
               (Fract ((Big_int.big_int_of_int 1), (Big_int.big_int_of_int 1)))
        else Fract ((Big_int.big_int_of_int 1), (Big_int.big_int_of_int 1)))
      (minus_rat x y);;

let rec ge_rat x y = less_eq_rat y x;;

let rec max0_rat
  x = max ord_rat
        (Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1))) x;;

let rec times_rat
  (Fract (a, b)) (Fract (c, d)) =
    fract_norm (Big_int.mult_big_int a c) (Big_int.mult_big_int b d);;

let rec plus_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0) then Fract (c, d)
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0) then Fract (a, b)
             else fract_norm
                    (Big_int.add_big_int (Big_int.mult_big_int a d)
                      (Big_int.mult_big_int c b))
                    (Big_int.mult_big_int b d)));;

let rec rat_components
  delta =
    ((fun _ -> true),
      (Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1)),
        (Fract ((Big_int.big_int_of_int 1), (Big_int.big_int_of_int 1)),
          (plus_rat,
            (times_rat,
              (max0_rat,
                (ge_rat,
                  (rat_gt delta, (true, (rat_name, (fun _ -> true)))))))))));;

let rec pos_arctic
  = function MinInfty -> false
    | Num_arc n -> Big_int.le_big_int (Big_int.big_int_of_int 0) n;;

let rec mat_arc_posI ap m = ap (nth (nth m Zero_nat) Zero_nat);;

let arctic_name : char list
  = ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 's'; 'e'; 'm'; 'i'; 'r'; 'i'; 'n';
      'g'];;

let rec gt_arctic
  uu x1 = match uu, x1 with uu, MinInfty -> true
    | MinInfty, Num_arc uv -> false
    | Num_arc x, Num_arc y -> Big_int.lt_big_int y x;;

let rec ge_arctic
  x xa1 = match x, xa1 with x, MinInfty -> true
    | MinInfty, Num_arc uu -> false
    | Num_arc x, Num_arc y -> Big_int.le_big_int y x;;

let rec times_arctic
  x0 y = match x0, y with MinInfty, y -> MinInfty
    | Num_arc v, MinInfty -> MinInfty
    | Num_arc x, Num_arc y -> Num_arc (Big_int.add_big_int x y);;

let rec transpose
  nr x1 = match nr, x1 with nr, [] -> replicate nr []
    | nr, v :: m -> map (fun (a, b) -> a :: b) (zip v (transpose nr m));;

let rec scalar_prodI
  ze pl ti v w = foldr (fun (x, y) -> pl (ti x y)) (zip v w) ze;;

let rec matT_vec_multI
  ze pl ti m v = map (fun w -> scalar_prodI ze pl ti w v) m;;

let rec mat_multI
  ze pl ti nr m1 m2 = map (matT_vec_multI ze pl ti (transpose nr m1)) m2;;

let rec plus_arctic
  x0 y = match x0, y with MinInfty, y -> y
    | Num_arc v, MinInfty -> Num_arc v
    | Num_arc x, Num_arc y -> Num_arc (max ord_int x y);;

let rec vec_plusI pl v w = map (fun xy -> pl (fst xy) (snd xy)) (zip v w);;

let rec mat_plusI
  pl m1 m2 = map (fun uv -> vec_plusI pl (fst uv) (snd uv)) (zip m1 m2);;

let one_arctic : arctic = Num_arc (Big_int.big_int_of_int 0);;

let zero_arctic : arctic = MinInfty;;

let rec mat_arctic_components
  n = let np = Suc (minus_nat n one_nat) in
      (wf_mat np np,
        (mat0I zero_arctic np np,
          (mat1I zero_arctic one_arctic np,
            (mat_plusI plus_arctic,
              (mat_multI zero_arctic plus_arctic times_arctic np,
                (id, (mat_geI ge_arctic,
                       (mat_geI gt_arctic,
                         (false,
                           (concat
                              [show shows_nata np; ['x']; show shows_nata np;
                                [' '; 'm'; 'a'; 't'; 'r'; 'i'; 'c'; 'e'; 's';
                                  ' '; 'o'; 'v'; 'e'; 'r'; ' '];
                                arctic_name],
                             mat_arc_posI pos_arctic))))))))));;

let rec get_comp_description
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = de;;

let rec get_comp_mono
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = mo;;

let rec get_comp_gt
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = gt;;

let rec get_comp_ge
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = ge;;

let rec get_comp_max0
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = mz;;

let rec vec_max0I x = map x;;

let rec mat_max0I mz = map (vec_max0I mz);;

let rec get_comp_times
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = ti;;

let rec get_comp_plus
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = pl;;

let rec get_comp_1
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = on;;

let rec get_comp_0
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = ze;;

let rec mat_components
  n sd c =
    let sdp = Suc (minus_nat sd one_nat) in
    let np = max ord_nat n sdp in
    (wf_mat np np,
      (mat0I (get_comp_0 c) np np,
        (mat1I (get_comp_0 c) (get_comp_1 c) np,
          (mat_plusI (get_comp_plus c),
            (mat_multI (get_comp_0 c) (get_comp_plus c) (get_comp_times c) np,
              (mat_max0I (get_comp_max0 c),
                (mat_geI (get_comp_ge c),
                  (mat_gtI (get_comp_ge c) (get_comp_gt c) sdp,
                    (get_comp_mono c,
                      (concat
                         [show shows_nata np; ['x']; show shows_nata np;
                           [' '; 'm'; 'a'; 't'; 'r'; 'i'; 'c'; 'e'; 's'; ' ';
                             'w'; 'i'; 't'; 'h'; ' '; 's'; 't'; 'r'; 'i'; 'c';
                             't'; ' '; 'd'; 'i'; 'm'; 'e'; 'n'; 's'; 'i'; 'o';
                             'n'; ' '];
                           show shows_nata sdp; [' '; 'o'; 'v'; 'e'; 'r'; ' '];
                           get_comp_description c],
                        (fun _ -> true)))))))))));;

let rec xml2vec_coeff
  xml2coeff =
    xmlManyElements ['v'; 'e'; 'c'; 't'; 'o'; 'r']
      (xml1element ['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't']
        xml2coeff id)
      id;;

let rec xml2mat_coeff
  ze xml2coeff =
    xmlOptions
      [(['m'; 'a'; 't'; 'r'; 'i'; 'x'],
         xmlManyElements ['m'; 'a'; 't'; 'r'; 'i'; 'x']
           (xml2vec_coeff xml2coeff) id);
        (['v'; 'e'; 'c'; 't'; 'o'; 'r'],
          xmlChange (xml2vec_coeff xml2coeff)
            (fun v ->
              v :: replicate (minus_nat (size_list v) one_nat)
                     (replicate (size_list v) ze)))];;

let arctic_components :
  (arctic -> bool) *
    (arctic *
      (arctic *
        ((arctic -> arctic -> arctic) *
          ((arctic -> arctic -> arctic) *
            ((arctic -> arctic) *
              ((arctic -> arctic -> bool) *
                ((arctic -> arctic -> bool) *
                  (bool * (char list * (arctic -> bool))))))))))
  = ((fun _ -> true),
      (zero_arctic,
        (one_arctic,
          (plus_arctic,
            (times_arctic,
              (id, (ge_arctic,
                     (gt_arctic, (false, (arctic_name, pos_arctic))))))))));;

let rec eq_arctic
  x0 x1 = match x0, x1 with MinInfty, MinInfty -> true
    | Num_arc int, Num_arc int' -> Big_int.eq_big_int int int'
    | MinInfty, Num_arc int' -> false
    | Num_arc int', MinInfty -> false;;

let eq_arctica = ({eq = eq_arctic} : arctic eq);;

let rec xml2arctic_coeff
  x = xmlOptions
        [(['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'],
           xmlChange (xml2int ['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'])
             (fun a -> Num_arc a));
          (['m'; 'i'; 'n'; 'u'; 's'; 'I'; 'n'; 'f'; 'i'; 'n'; 'i'; 't'; 'y'],
            xml0elements
              ['m'; 'i'; 'n'; 'u'; 's'; 'I'; 'n'; 'f'; 'i'; 'n'; 'i'; 't'; 'y']
              MinInfty)]
        x;;

let int_name : char list
  = ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 'n';
      'e'; 'g'; 'a'; 't'; 'i'; 'v'; 'e'; ' '; 'c'; 'o'; 'n'; 's'; 't'; 'a'; 'n';
      't'; 's'];;

let rec ge_int x y = Big_int.le_big_int y x;;

let rec max0_int x = max ord_int (Big_int.big_int_of_int 0) x;;

let int_components :
  (Big_int.big_int -> bool) *
    (Big_int.big_int *
      (Big_int.big_int *
        ((Big_int.big_int -> Big_int.big_int -> Big_int.big_int) *
          ((Big_int.big_int -> Big_int.big_int -> Big_int.big_int) *
            ((Big_int.big_int -> Big_int.big_int) *
              ((Big_int.big_int -> Big_int.big_int -> bool) *
                ((Big_int.big_int -> Big_int.big_int -> bool) *
                  (bool * (char list * (Big_int.big_int -> bool))))))))))
  = ((fun _ -> true),
      ((Big_int.big_int_of_int 0),
        ((Big_int.big_int_of_int 1),
          (Big_int.add_big_int,
            (Big_int.mult_big_int,
              (max0_int,
                (ge_int,
                  ((fun x y -> Big_int.lt_big_int y x),
                    (true, (int_name, (fun _ -> true)))))))))));;

let rec poly_to_lpoly _A
  uu uv uw ux x4 = match uu, uv, uw, ux, x4 with
    uu, uv, uw, ux, PNum i -> return (i, [])
    | ze, on, pl, ti, PVar x -> return (ze, append (replicate x ze) [on])
    | ze, on, pl, ti, PSum [p] -> poly_to_lpoly _A ze on pl ti p
    | ze, on, pl, ti, PSum (p :: q :: ps) ->
        bind (poly_to_lpoly _A ze on pl ti p)
          (fun (cp, ncp) ->
            bind (poly_to_lpoly _A ze on pl ti (PSum (q :: ps)))
              (fun (cq, ncq) -> return (pl cp cq, sum_lpoly pl ncp ncq)))
    | ze, on, pl, ti, PMult [p] -> poly_to_lpoly _A ze on pl ti p
    | ze, on, pl, ti, PMult (p :: q :: ps) ->
        bind (poly_to_lpoly _A ze on pl ti p)
          (fun (cp, ncp) ->
            bind (poly_to_lpoly _A ze on pl ti (PMult (q :: ps)))
              (fun (cq, ncq) ->
                (if forall (eqa _A ze) ncp
                  then return (ti cp cq, map (ti cp) ncq)
                  else (if forall (eqa _A ze) ncq
                         then return (ti cp cq, map (fun x -> ti x cq) ncp)
                         else error ['n'; 'o'; 'n'; '-'; 'l'; 'i'; 'n'; 'e';
                                      'a'; 'r'; ' '; 'p'; 'o'; 'l'; 'y'; 'n';
                                      'o'; 'm'; 'i'; 'a'; 'l'; 's'; ' '; 'n';
                                      'o'; 't'; ' '; 's'; 'u'; 'p'; 'p'; 'o';
                                      'r'; 't'; 'e'; 'd']))));;

let rec xml2int_coeff x = xml2int ['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'] x;;

let rec divide_rat
  (Fract (a, b)) (Fract (c, d)) =
    fract_norm (Big_int.mult_big_int a d) (Big_int.mult_big_int b c);;

let default_delta : rat
  = divide_rat (Fract ((Big_int.big_int_of_int 1), (Big_int.big_int_of_int 1)))
      (Fract
        ((Big_int.big_int_of_string "1000000000000000000000000000000000000"),
          (Big_int.big_int_of_int 1)));;

let rec xml2interpretation_type
  x = xml1element ['t'; 'y'; 'p'; 'e']
        (xmlOptions
          [(['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'],
             xml2elements ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l']
               (xml1element ['d'; 'o'; 'm'; 'a'; 'i'; 'n']
                 (xmlOptions
                   [(['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'],
                      xml0elements ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's']
                        Natural);
                     (['a'; 'r'; 'c'; 't'; 'i'; 'c'],
                       xml0elements ['a'; 'r'; 'c'; 't'; 'i'; 'c'] Arctic);
                     (['a'; 'r'; 'c'; 't'; 'i'; 'c'; 'B'; 'e'; 'l'; 'o'; 'w';
                        'Z'; 'e'; 'r'; 'o'],
                       xml0elements
                         ['a'; 'r'; 'c'; 't'; 'i'; 'c'; 'B'; 'e'; 'l'; 'o'; 'w';
                           'Z'; 'e'; 'r'; 'o']
                         Arctic);
                     (['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's'],
                       xml0elements
                         ['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's']
                         (Rational default_delta));
                     (['m'; 'a'; 't'; 'r'; 'i'; 'c'; 'e'; 's'],
                       xml3elements ['m'; 'a'; 't'; 'r'; 'i'; 'c'; 'e'; 's']
                         (xml2nat ['d'; 'i'; 'm'; 'e'; 'n'; 's'; 'i'; 'o'; 'n'])
                         (xml2nat
                           ['s'; 't'; 'r'; 'i'; 'c'; 't'; 'D'; 'i'; 'm'; 'e';
                             'n'; 's'; 'i'; 'o'; 'n'])
                         (xml1element ['d'; 'o'; 'm'; 'a'; 'i'; 'n']
                           (xmlOptions
                             [(['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's'],
                                xml0elements
                                  ['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's']
                                  Zero_nat);
                               (['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'],
                                 xml0elements
                                   ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's']
                                   one_nat);
                               (['a'; 'r'; 'c'; 't'; 'i'; 'c'],
                                 xml0elements ['a'; 'r'; 'c'; 't'; 'i'; 'c']
                                   (nat (Big_int.big_int_of_int 2)));
                               (['a'; 'r'; 'c'; 't'; 'i'; 'c'; 'B'; 'e'; 'l';
                                  'o'; 'w'; 'Z'; 'e'; 'r'; 'o'],
                                 xml0elements
                                   ['a'; 'r'; 'c'; 't'; 'i'; 'c'; 'B'; 'e'; 'l';
                                     'o'; 'w'; 'Z'; 'e'; 'r'; 'o']
                                   (nat (Big_int.big_int_of_int 2)))])
                           id)
                         (fun di sd doa ->
                           (if eq_nat doa one_nat then Int_mat (di, sd)
                             else (if eq_nat doa Zero_nat
                                    then Rat_mat (default_delta, di, sd)
                                    else Arc_mat di))))])
                 id)
               (xml2nat ['d'; 'e'; 'g'; 'r'; 'e'; 'e']) (fun typea _ -> typea));
            (['m'; 'a'; 't'; 'r'; 'i'; 'x'; 'I'; 'n'; 't'; 'e'; 'r'; 'p'; 'r';
               'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n'],
              xml3elements
                ['m'; 'a'; 't'; 'r'; 'i'; 'x'; 'I'; 'n'; 't'; 'e'; 'r'; 'p';
                  'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n']
                (xml1element ['d'; 'o'; 'm'; 'a'; 'i'; 'n']
                  (xmlOptions
                    [(['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's'],
                       xml0elements
                         ['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's']
                         Zero_nat);
                      (['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'],
                        xml0elements ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's']
                          one_nat);
                      (['a'; 'r'; 'c'; 't'; 'i'; 'c'],
                        xml0elements ['a'; 'r'; 'c'; 't'; 'i'; 'c']
                          (nat (Big_int.big_int_of_int 2)));
                      (['a'; 'r'; 'c'; 't'; 'i'; 'c'; 'B'; 'e'; 'l'; 'o'; 'w';
                         'Z'; 'e'; 'r'; 'o'],
                        xml0elements
                          ['a'; 'r'; 'c'; 't'; 'i'; 'c'; 'B'; 'e'; 'l'; 'o';
                            'w'; 'Z'; 'e'; 'r'; 'o']
                          (nat (Big_int.big_int_of_int 2)))])
                  id)
                (xml2nat ['d'; 'i'; 'm'; 'e'; 'n'; 's'; 'i'; 'o'; 'n'])
                (xml2nat
                  ['s'; 't'; 'r'; 'i'; 'c'; 't'; 'D'; 'i'; 'm'; 'e'; 'n'; 's';
                    'i'; 'o'; 'n'])
                (fun doa di sd ->
                  (if eq_nat doa one_nat then Int_mat (di, sd)
                    else (if eq_nat doa Zero_nat
                           then Rat_mat (default_delta, di, sd)
                           else Arc_mat di))))])
        id x;;

let rec xml2interpretation
  xml2name =
    xml1CmanyElements
      ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n']
      xml2interpretation_type
      (fun a ->
        (match a
          with Natural ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (xml2poly xml2int_coeff x)
                  (poly_to_lpoly eq_int (get_comp_0 int_components)
                    (get_comp_1 int_components) (get_comp_plus int_components)
                    (get_comp_times int_components)))
              (fun f aa poly ->
                Alter1
                  ((f, aa),
                    (fst poly,
                      fitLength (get_comp_0 int_components) aa (snd poly))))
          | Arctic ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (xml2poly xml2arctic_coeff x)
                  (poly_to_lpoly eq_arctica (get_comp_0 arctic_components)
                    (get_comp_1 arctic_components)
                    (get_comp_plus arctic_components)
                    (get_comp_times arctic_components)))
              (fun f aa poly ->
                Alter3
                  ((f, aa),
                    (fst poly,
                      fitLength (get_comp_0 arctic_components) aa (snd poly))))
          | Int_mat (n, sd) ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (xml2poly
                       (xml2mat_coeff (Big_int.big_int_of_int 0) xml2int_coeff)
                       x)
                  (poly_to_lpoly (eq_lista (eq_lista eq_int))
                    (get_comp_0 (mat_components n sd int_components))
                    (get_comp_1 (mat_components n sd int_components))
                    (get_comp_plus (mat_components n sd int_components))
                    (get_comp_times (mat_components n sd int_components))))
              (fun f aa poly ->
                Alter4
                  ((f, aa),
                    (fst poly,
                      fitLength
                        (get_comp_0 (mat_components n sd int_components)) aa
                        (snd poly))))
          | Arc_mat n ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (xml2poly (xml2mat_coeff MinInfty xml2arctic_coeff) x)
                  (poly_to_lpoly (eq_lista (eq_lista eq_arctica))
                    (get_comp_0 (mat_arctic_components n))
                    (get_comp_1 (mat_arctic_components n))
                    (get_comp_plus (mat_arctic_components n))
                    (get_comp_times (mat_arctic_components n))))
              (fun f aa poly ->
                Alter6
                  ((f, aa),
                    (fst poly,
                      fitLength (get_comp_0 (mat_arctic_components n)) aa
                        (snd poly))))
          | Rational d ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (xml2poly xml2rat x)
                  (poly_to_lpoly eq_rata (get_comp_0 (rat_components d))
                    (get_comp_1 (rat_components d))
                    (get_comp_plus (rat_components d))
                    (get_comp_times (rat_components d))))
              (fun f aa poly ->
                Alter2
                  ((f, aa),
                    (fst poly,
                      fitLength (get_comp_0 (rat_components d)) aa (snd poly))))
          | Rat_mat (d, n, sd) ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (xml2nat ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (xml2poly
                       (xml2mat_coeff
                         (Fract
                           ((Big_int.big_int_of_int 0),
                             (Big_int.big_int_of_int 1)))
                         xml2rat)
                       x)
                  (poly_to_lpoly (eq_lista (eq_lista eq_rata))
                    (get_comp_0 (mat_components n sd (rat_components d)))
                    (get_comp_1 (mat_components n sd (rat_components d)))
                    (get_comp_plus (mat_components n sd (rat_components d)))
                    (get_comp_times (mat_components n sd (rat_components d)))))
              (fun f aa poly ->
                Alter5
                  ((f, aa),
                    (fst poly,
                      fitLength
                        (get_comp_0 (mat_components n sd (rat_components d))) aa
                        (snd poly))))))
      (fun typea pi ->
        (match typea with Natural -> Int_carrier (map alter1a pi)
          | Arctic -> Arctic_carrier (map alter3a pi)
          | Int_mat (n, sd) -> Int_mat_carrier (n, sd, map alter4a pi)
          | Arc_mat n -> Arc_mat_carrier (n, map alter6a pi)
          | Rational d -> Rat_carrier (d, map alter2a pi)
          | Rat_mat (d, n, sd) -> Rat_mat_carrier (d, n, sd, map alter5a pi)));;

let rec xml2redpair
  xml2name =
    xml1element ['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r']
      (xmlOptions
        [(['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o';
            'n'],
           xml2interpretation xml2name);
          (['p'; 'a'; 't'; 'h'; 'O'; 'r'; 'd'; 'e'; 'r'],
            xml2path_order xml2name)])
      id;;

let rec xml2ndChoice
  reqName p1 choiceName p2 combi (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name reqName &&
          (null atts &&
            (less_eq_nat (nat (Big_int.big_int_of_int 2)) (size_list cs) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun a ->
               bind (xml2bool choiceName (nth cs one_nat))
                 (fun choice ->
                   (if choice
                     then bind (p2 (nth cs (minus_nat (size_list cs) one_nat)))
                            (fun b -> return (combi a (Some b)))
                     else return (combi a None))))
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml2dpProof
  ld xml2name x =
    xmlOptions
      [(['p'; 'I'; 's'; 'E'; 'm'; 'p'; 't'; 'y'],
         xml0elements ['p'; 'I'; 's'; 'E'; 'm'; 'p'; 't'; 'y'] PisEmpty);
        (['d'; 'e'; 'p'; 'G'; 'r'; 'a'; 'p'; 'h'; 'P'; 'r'; 'o'; 'c'],
          xmlManyElements
            ['d'; 'e'; 'p'; 'G'; 'r'; 'a'; 'p'; 'h'; 'P'; 'r'; 'o'; 'c']
            (xml2ndChoice ['c'; 'o'; 'm'; 'p'; 'o'; 'n'; 'e'; 'n'; 't']
              (xml1element ['d'; 'p'; 's'] (xml2rules (xml2name ld)) id)
              ['r'; 'e'; 'a'; 'l'; 'S'; 'c'; 'c'] (xml2dpProof ld xml2name)
              (fun dps prfOpt -> (prfOpt, dps)))
            (fun a -> DepGraphProc a));
        (['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o'; 'c'],
          xml3elements ['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o'; 'c']
            (xml2redpair (xml2name ld))
            (xml1element ['d'; 'p'; 's'] (xml2rules (xml2name ld)) id)
            (xml2dpProof ld xml2name) (fun a b c -> RedPairProc (a, b, c)));
        (['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'U'; 'r'; 'P'; 'r'; 'o'; 'c'],
          xml4elements
            ['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'U'; 'r'; 'P'; 'r'; 'o'; 'c']
            (xml2redpair (xml2name ld))
            (xml1element ['d'; 'p'; 's'] (xml2rules (xml2name ld)) id)
            (xml1element ['u'; 's'; 'a'; 'b'; 'l'; 'e'; 'R'; 'u'; 'l'; 'e'; 's']
              (xml2rules (xml2name ld)) id)
            (xml2dpProof ld xml2name)
            (fun a b c d -> RedPairUrProc (a, b, c, d)));
        (['m'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o';
           'c'],
          xml4elements
            ['m'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r';
              'o'; 'c']
            (xml2redpair (xml2name ld))
            (xml1element ['d'; 'p'; 's'] (xml2rules (xml2name ld)) id)
            (xml1element ['t'; 'r'; 's'] (xml2rules (xml2name ld)) id)
            (xml2dpProof ld xml2name)
            (fun a b c d -> MonoRedPairProc (a, b, c, d)));
        (['m'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'U'; 'r'; 'P';
           'r'; 'o'; 'c'],
          xml5elements
            ['m'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'U'; 'r';
              'P'; 'r'; 'o'; 'c']
            (xml2redpair (xml2name ld))
            (xml1element ['d'; 'p'; 's'] (xml2rules (xml2name ld)) id)
            (xml1element ['t'; 'r'; 's'] (xml2rules (xml2name ld)) id)
            (xml1element ['u'; 's'; 'a'; 'b'; 'l'; 'e'; 'R'; 'u'; 'l'; 'e'; 's']
              (xml2rules (xml2name ld)) id)
            (xml2dpProof ld xml2name)
            (fun a b c d e -> MonoRedPairUrProc (a, b, c, d, e)));
        (['s'; 'e'; 'm'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c'],
          xml4elements ['s'; 'e'; 'm'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
            (xml2sl_inter (xml2name ld))
            (xml1element ['d'; 'p'; 's'] (xml2rules (xml2name (Suc ld))) id)
            (xml1element ['t'; 'r'; 's'] (xml2rules (xml2name (Suc ld))) id)
            (xml2dpProof (Suc ld) xml2name)
            (fun a ->
              (match a with None -> (fun aa b c -> RootlabProc (aa, b, c))
                | Some i ->
                  (fun aa b c -> SemlabProc (i, id_labeling, aa, b, c)))));
        (['u'; 'n'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c'],
          xml3elements ['u'; 'n'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
            (xml1element ['d'; 'p'; 's']
              (xml2rules (xml2name (minus_nat ld one_nat))) id)
            (xml1element ['t'; 'r'; 's']
              (xml2rules (xml2name (minus_nat ld one_nat))) id)
            (xml2dpProof (minus_nat ld one_nat) xml2name)
            (fun a b c -> UnlabProc (a, b, c)));
        (['s'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'P'; 'r'; 'o'; 'c'],
          xml3elements ['s'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'P'; 'r'; 'o'; 'c']
            (xml2sp (xml2name ld))
            (xml1element ['d'; 'p'; 's'] (xml2rules (xml2name ld)) id)
            (xml2dpProof ld xml2name) (fun a b c -> SpscProc (a, b, c)));
        (['s'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'C'; 'r'; 'i'; 't';
           'e'; 'r'; 'i'; 'o'; 'n'],
          xml1manyElements
            ['s'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'C'; 'r'; 'i';
              't'; 'e'; 'r'; 'i'; 'o'; 'n']
            (xmlOptions
              [(['s'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'C'; 'r'; 'i'; 't'; 'e';
                  'r'; 'i'; 'o'; 'n'],
                 xml0elements
                   ['s'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'C'; 'r'; 'i'; 't'; 'e';
                     'r'; 'i'; 'o'; 'n']
                   None);
                (['r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; 'P'; 'a'; 'i';
                   'r'],
                  xml2elements
                    ['r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; 'P'; 'a'; 'i';
                      'r']
                    (xml2redpair (xml2name ld))
                    (xml1element
                      ['u'; 's'; 'a'; 'b'; 'l'; 'e'; 'R'; 'u'; 'l'; 'e'; 's']
                      (xml2rules (xml2name ld)) id)
                    (fun redp ur -> Some (redp, ur)))])
            (xml2scg (xml2name ld))
            (fun a ->
              (match a with None -> (fun aa -> SizeChangeSubTerm aa)
                | Some redp_ur ->
                  (fun aa ->
                    SizeChangeRedPair (fst redp_ur, snd redp_ur, aa)))))]
      x;;

let rec foldR
  f x1 b = match f, x1, b with f, [], b -> b
    | f, x :: xs, b -> f x xs (foldR f xs b);;

let rec mat_monoI geq1 sd m = forall (any geq1) (sub_mat sd sd m);;

let rec mul_pvarsI _A
  uu uv a x3 = match uu, uv, a, x3 with uu, uv, a, [] -> []
    | ze, ti, a, (x, b) :: vas ->
        let p = ti a b in
        let res = mul_pvarsI _A ze ti a vas in
        (if eqa _A p ze then res else (x, p) :: res);;

type ('a, 'b) l_poly = LPoly of 'b * ('a * 'b) list;;

let rec mul_lpolyI _A
  ze ti a (LPoly (b, vas)) = LPoly (ti a b, mul_pvarsI _A ze ti a vas);;

let rec list_sumI
  ze pl x2 = match ze, pl, x2 with ze, pl, [] -> ze
    | ze, pl, x :: xs -> pl x (list_sumI ze pl xs);;

let rec add_varI _A _B
  uu uv x a xa4 = match uu, uv, x, a, xa4 with uu, uv, x, a, [] -> [(x, a)]
    | ze, pl, x, a, (y, b) :: vas ->
        (if eqa _B x y
          then let s = pl a b in
               (if eqa _A s ze then vas else (x, s) :: vas)
          else (y, b) :: add_varI _A _B ze pl x a vas);;

let rec sum_pvarsI _A _B
  uu uv x2 vbs = match uu, uv, x2, vbs with uu, uv, [], vbs -> vbs
    | ze, pl, (x, a) :: vas, vbs ->
        (if eqa _A a ze then sum_pvarsI _A _B ze pl vas vbs
          else sum_pvarsI _A _B ze pl vas (add_varI _A _B ze pl x a vbs));;

let rec sum_lpolyI _A _B
  ze pl (LPoly (a, vas)) (LPoly (b, vbs)) =
    LPoly (pl a b, sum_pvarsI _A _B ze pl vas vbs);;

let rec l_poly_case f1 (LPoly (a, lista)) = f1 a lista;;

let rec pleftI _A _C
  ze on pl ti mz pi x6 = match ze, on, pl, ti, mz, pi, x6 with
    ze, on, pl, ti, mz, pi, Var x -> LPoly (ze, [(x, on)])
    | ze, on, pl, ti, mz, pi, Fun (f, ts) ->
        let (c, asa) = pi f (size_list ts) in
        (match
          sum_lpolyI _A _C ze pl (LPoly (c, []))
            (list_sumI (LPoly (ze, [])) (sum_lpolyI _A _C ze pl)
              (map (fun at ->
                     mul_lpolyI _A ze ti (fst at)
                       (pleftI _A _C ze on pl ti mz pi (snd at)))
                (zip asa ts)))
          with LPoly (d, []) -> LPoly (mz d, [])
          | LPoly (d, ab :: lista) -> LPoly (d, ab :: lista));;

type ('a, 'b, 'c) tRSProof =
  DPTrans of
    ((('a, 'b) lab shp, 'c) term * (('a, 'b) lab shp, 'c) term) list *
      (('a, 'b) lab shp, 'b, 'c) dPProof
  | RuleRemoval of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Unlab of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Semlab of
      ('a, 'b) sl_inter *
        (('a, 'b) lab -> nat list -> ('b, (('a, 'b) lab list)) either) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Rootlab of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Fcc of
      (('a, 'b) lab, 'c) ctxt list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | StringReversal of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | RisEmpty;;

let rec lpoS x = shows_string ['>'; '_'; 'L'; 'P'; 'O'; '+'; 'A'; 'F'] x;;

let rec xml2trsInput
  xml2name =
    xml1element ['i'; 'n'; 'p'; 'u'; 't']
      (xml1element ['t'; 'r'; 's'; 'I'; 'n'; 'p'; 'u'; 't']
        (xml1element ['t'; 'r'; 's'] (xml2rules xml2name) id) id)
      id;;

let rec xml2flat_contexts
  xml2name =
    xmlManyElements ['f'; 'l'; 'a'; 't'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 's']
      (xml2context xml2name) id;;

let rec xml2trsProof
  xml2name xml2sname xml2lsname x =
    xmlOptions
      [(['r'; 'I'; 's'; 'E'; 'm'; 'p'; 't'; 'y'],
         xml0elements ['r'; 'I'; 's'; 'E'; 'm'; 'p'; 't'; 'y'] RisEmpty);
        (['d'; 'p'; 'T'; 'r'; 'a'; 'n'; 's'],
          xml3elements ['d'; 'p'; 'T'; 'r'; 'a'; 'n'; 's']
            (xml1element ['d'; 'p'; 's'] (xml2rules xml2sname) id)
            (xml2bool
              ['m'; 'a'; 'r'; 'k'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l'; 's'])
            (xml2dpProof Zero_nat xml2lsname)
            (fun dps _ -> (fun a -> DPTrans (dps, a))));
        (['r'; 'u'; 'l'; 'e'; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l'],
          xml3elements ['r'; 'u'; 'l'; 'e'; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l']
            (xml2redpair xml2name)
            (xml1element ['t'; 'r'; 's'] (xml2rules xml2name) id)
            (xml2trsProof xml2name xml2sname xml2lsname)
            (fun a b c -> RuleRemoval (a, b, c)));
        (['u'; 'n'; 'l'; 'a'; 'b'],
          xml2elements ['u'; 'n'; 'l'; 'a'; 'b']
            (xml1element ['t'; 'r'; 's'] (xml2rules xml2name) id)
            (xml2trsProof xml2name xml2sname xml2lsname)
            (fun a b -> Unlab (a, b)));
        (['s'; 't'; 'r'; 'i'; 'n'; 'g'; 'R'; 'e'; 'v'; 'e'; 'r'; 's'; 'a'; 'l'],
          xml2elements
            ['s'; 't'; 'r'; 'i'; 'n'; 'g'; 'R'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
              'l']
            (xml1element ['t'; 'r'; 's'] (xml2rules xml2name) id)
            (xml2trsProof xml2name xml2sname xml2lsname)
            (fun a b -> StringReversal (a, b)));
        (['s'; 'e'; 'm'; 'l'; 'a'; 'b'],
          xml3elements ['s'; 'e'; 'm'; 'l'; 'a'; 'b'] (xml2sl_inter xml2name)
            (xml1element ['t'; 'r'; 's'] (xml2rules xml2name) id)
            (xml2trsProof xml2name xml2sname xml2lsname)
            (fun a ->
              (match a with None -> (fun aa b -> Rootlab (aa, b))
                | Some i -> (fun aa b -> Semlab (i, id_labeling, aa, b)))));
        (['f'; 'l'; 'a'; 't'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 'C'; 'l'; 'o';
           's'; 'u'; 'r'; 'e'],
          xml3elements
            ['f'; 'l'; 'a'; 't'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 'C'; 'l';
              'o'; 's'; 'u'; 'r'; 'e']
            (xml2flat_contexts xml2name)
            (xml1element ['t'; 'r'; 's'] (xml2rules xml2name) id)
            (xml2trsProof xml2name xml2sname xml2lsname)
            (fun a b c -> Fcc (a, b, c)))]
      x;;

let rec checkF f g x = (if f x then return () else error (g x));;

let rec checkAllMI
  f xs =
    foldr (fun x iq ->
            bind (catch (f x) (fun xa -> error (x, xa))) (fun _ -> iq))
      xs (return ());;

let rec check_edges
  ss g c d =
    catch (checkAllMI
            (fun n ->
              catch (checkAllMI
                      (fun m ->
                        checkB (not (g (m, n)))
                          (comp (shows_string
                                  ['e'; 'd'; 'g'; 'e'; ' '; 'f'; 'r'; 'o'; 'm';
                                    ' '])
                            (comp (ss m)
                              (comp (shows_string [' '; 't'; 'o'; ' '])
                                (ss n)))))
                      d)
                (fun x -> error (snd x)))
            c)
      (fun x -> error (snd x));;

let bot_bool : bool = false;;

let rec less_bool x0 b = match x0, b with true, b -> false
                    | false, b -> b;;

let rec less_eq_bool x0 b = match x0, b with true, b -> b
                       | false, b -> true;;

let ord_bool = ({less_eq = less_eq_bool; less = less_bool} : bool ord);;

let preorder_bool = ({ord_preorder = ord_bool} : bool preorder);;

let bot_boola = ({preorder_bot = preorder_bool; bot = bot_bool} : bool bot);;

let rec bot_fun _B = (fun _ -> bot _B);;

let rec prightI _A _C
  ze on pl ti mz pi x6 = match ze, on, pl, ti, mz, pi, x6 with
    ze, on, pl, ti, mz, pi, Var x -> LPoly (ze, [(x, on)])
    | ze, on, pl, ti, mz, pi, Fun (f, ts) ->
        let (c, asa) = pi f (size_list ts) in
        let LPoly (d, a) =
          sum_lpolyI _A _C ze pl (LPoly (c, []))
            (list_sumI (LPoly (ze, [])) (sum_lpolyI _A _C ze pl)
              (map (fun at ->
                     mul_lpolyI _A ze ti (fst at)
                       (prightI _A _C ze on pl ti mz pi (snd at)))
                (zip asa ts)))
          in
        LPoly (mz d, a);;

type ('a, 'b) tRSDisproof =
  Loop of
    (('a, 'b) term list *
      (('a, 'b) ctxt * (('a, 'b) term * ('a, 'b) term) list))
  | NotWellFormed;;

let rec lpoNS x = shows_string ['>'; '='; '_'; 'L'; 'P'; 'O'; '+'; 'A'; 'F'] x;;

let rec funs_term_list'
  = function Var uu -> []
    | Fun (f, ts) -> f :: flat_map funs_term_list' ts;;

let rec funs_rule_list'
  r = append (funs_term_list' (fst r)) (funs_term_list' (snd r));;

let rec funs_trs_list _A trs = remdups _A (flat_map funs_rule_list' trs);;

let rec vars_term_list'
  = function Var x -> [x]
    | Fun (uu, ts) -> flat_map vars_term_list' ts;;

let rec vars_rule_list'
  r = append (vars_term_list' (fst r)) (vars_term_list' (snd r));;

let rec vars_trs_list _B trs = remdups _B (flat_map vars_rule_list' trs);;

let rec xml2plainName
  x = xmlOptions
        [(['n'; 'a'; 'm'; 'e'],
           xmlChange (xml2text ['n'; 'a'; 'm'; 'e']) (fun a -> UnLab a));
          (['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l'],
            xml2elements
              ['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l']
              xml2plainName (xml2label xml2plainName)
              (fun f a ->
                (match a with Left aa -> Lab (f, aa)
                  | Right aa -> FunLab (f, aa))))]
        x;;

let rec xml2sharpName
  x = (if eq_list eq_char (xmlGetName x) ['s'; 'h'; 'a'; 'r'; 'p']
        then xml1element ['s'; 'h'; 'a'; 'r'; 'p'] xml2plainName
               (fun a -> Sharp a) x
        else xmlChange xml2plainName (fun a -> Plain a) x);;

let rec length_unique _A
  = function [] -> Zero_nat
    | x :: xs ->
        (if member _A x xs then length_unique _A xs
          else Suc (length_unique _A xs));;

type certificationResult = Certified of bool | Error of char list;;

let rec vars_term_list _B t = remdups _B (vars_term_list' t);;

let rec enum_vectors
  c x1 = match c, x1 with c, [] -> [[]]
    | c, x :: xs ->
        let a = enum_vectors c xs in
        concat_map (fun vec -> map (fun ca -> (x, ca) :: vec) c) a;;

let rec assign_from_vector _A vec x = the (lookup _A x vec);;

let rec eq_lab _A _B
  x0 x1 = match x0, x1 with
    FunLab (f, ls), FunLab (g, ks) ->
      eq_lab _A _B f g && list_all2 (eq_lab _A _B) ls ks
    | Lab (lab, l), Lab (lab', l') -> eq_lab _A _B lab lab' && eqa _B l l'
    | UnLab f, UnLab f' -> eqa _A f f'
    | Lab (lab, l), FunLab (lab', list') -> false
    | FunLab (lab', list'), Lab (lab, l) -> false
    | Lab (lab, l), UnLab f' -> false
    | UnLab f', Lab (lab, l) -> false
    | FunLab (lab, lista), UnLab f' -> false
    | UnLab f', FunLab (lab, lista) -> false;;

let rec eq_laba _A _B = ({eq = eq_lab _A _B} : ('a, 'b) lab eq);;

let rec shows_lab _A _B
  = function UnLab f -> shows _A f
    | Lab (f, l) ->
        comp (shows_lab _A _B f)
          (comp (shows_string ['[']) (comp (shows _B l) (shows_string [']'])))
    | FunLab (f, l) ->
        comp (shows_lab _A _B f)
          (comp (shows_string ['['])
            (comp (shows_map (shows_lab _A _B) l) (shows_string [']'])));;

let rec shows_list_lab _A _B fs = shows_list_aux (shows_lab _A _B) fs;;

let rec shows_laba _A _B =
  ({shows = shows_lab _A _B; shows_list = shows_list_lab _A _B} :
    ('a, 'b) lab shows);;

let rec label
  f x1 = match f, x1 with f, Left l -> Lab (f, l)
    | f, Right l -> FunLab (f, l);;

let rec eval_lab
  i l alpha x3 = match i, l, alpha, x3 with
    i, l, alpha, Var x -> (alpha x, Var x)
    | i, l, alpha, Fun (f, ts) ->
        let clts = map (eval_lab i l alpha) ts in
        let cs = map fst clts in
        let c = i f cs in
        let lts = map snd clts in
        let la = l f cs in
        (c, Fun (label f la, lts));;

let rec check_sl_rule_ass (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  mc i la lR alpha (l, r) =
    let cl_ll = eval_lab i la alpha l in
    let cr_lr = eval_lab i la alpha r in
    bind (checkB (implies mc (eqa _C1 (fst cl_ll) (fst cr_lr)))
           (comp (shows_string ['r'; 'u'; 'l'; 'e'; ' '])
             (comp (shows_rule (shows_laba _A2 _B2) _D2 (l, r))
               (comp (shows_string
                       [' '; 'v'; 'i'; 'o'; 'l'; 'a'; 't'; 'e'; 's'; ' '; 't';
                         'h'; 'e'; ' '; 'm'; 'o'; 'd'; 'e'; 'l'; ' '; 'c'; 'o';
                         'n'; 'd'; 'i'; 't'; 'i'; 'o'; 'n'; ','; ' '; '['; 'l';
                         'h'; 's'; ']'; ' '; '='; ' '])
                 (comp (shows _C2 (fst cl_ll))
                   (comp (shows_string
                           [','; ' '; '['; 'r'; 'h'; 's'; ']'; ' '; '='; ' '])
                     (shows _C2 (fst cr_lr))))))))
      (fun _ ->
        checkB
          (member
            (eqc (eq_terma (eq_laba _A1 _B1) _D1)
              (eq_terma (eq_laba _A1 _B1) _D1))
            (snd cl_ll, snd cr_lr) lR)
          (comp (shows_string
                  ['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; ' '; 'r'; 'u'; 'l'; 'e';
                    ' '])
            (comp (shows_rule (shows_laba _A2 _B2) _D2 (snd cl_ll, snd cr_lr))
              (shows_string [' '; 'm'; 'i'; 's'; 's'; 'i'; 'n'; 'g']))));;

let rec check_sl_rule (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  i la c mc lR (l, r) =
    catch (checkAllMI
            (fun alpha ->
              check_sl_rule_ass (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2) mc i
                la lR alpha (l, r))
            (map (assign_from_vector _D1)
              (enum_vectors c
                (list_union _D1 (vars_term_list _D1 l)
                  (vars_term_list _D1 r)))))
      (fun x -> error (snd x));;

let rec check_sl_model_lab_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  i l c lR r =
    catch (checkAllMI
            (check_sl_rule (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2) i l c
              true lR)
            r)
      (fun x -> error (snd x));;

let rec rl_fun uu cs = Right cs;;

let rec rl_inter_list _A fs d g uu = (if member _A g fs then g else d);;

let rec check_sl_lab_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  i l c lP p =
    catch (checkAllMI
            (check_sl_rule (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2) i l c
              false lP)
            p)
      (fun x -> error (snd x));;

let rec get_lab = function Lab (f, l) -> Left l
                  | FunLab (f, l) -> Right l;;

let rec unlab
  = function Lab (f, l) -> f
    | FunLab (f, l) -> f
    | UnLab v -> UnLab v;;

let rec check_sl_only_lab_and_decr (_A1, _A2) (_B1, _B2) (_C1, _C2)
  gr r =
    (fun xs ->
      catch (checkAllMI
              (fun (ll, lr) ->
                let l = map_term unlab ll in
                let ra = map_term unlab lr in
                checkB
                  (if eq_term (eq_laba _B1 _A1) _C1 l ra
                    then not (is_var ll) &&
                           (not (is_var lr) &&
                             gr (get_lab (the (root ll)),
                                  get_lab (the (root lr))))
                           || eq_term (eq_laba _B1 _A1) _C1 ll lr &&
                                member
                                  (eqc (eq_terma (eq_laba _B1 _A1) _C1)
                                    (eq_terma (eq_laba _B1 _A1) _C1))
                                  (l, ra) r
                    else member
                           (eqc (eq_terma (eq_laba _B1 _A1) _C1)
                             (eq_terma (eq_laba _B1 _A1) _C1))
                           (l, ra) r)
                  (comp (shows_string
                          ['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; ' '; 'r'; 'u';
                            'l'; 'e'; ' '])
                    (comp (shows_rule (shows_laba _B2 _A2) _C2 (l, ra))
                      (shows_string
                        [' '; 'i'; 's'; ' '; 'n'; 'e'; 'i'; 't'; 'h'; 'e'; 'r';
                          ' '; 'd'; 'e'; 'c'; 'r'; 'e'; 'a'; 's'; 'i'; 'n'; 'g';
                          ' '; 'n'; 'o'; 'r'; ' '; 'a'; ' '; 'l'; 'a'; 'b'; 'e';
                          'l'; 'e'; 'd'; ' '; 'v'; 'e'; 'r'; 's'; 'i'; 'o'; 'n';
                          ' '; 'o'; 'f'; ' '; 'o'; 'r'; 'i'; 'g'; 'i'; 'n'; 'a';
                          'l'; ' '; 'r'; 'u'; 'l'; 'e']))))
              xs)
        (fun x -> error (snd x)));;

let rec check_sem_lab_tt (_A1, _A2) (_B1, _B2) (_C1, _C2)
  check_lab check_model_lab gr i l r lR =
    catch (bind (check_model_lab lR r)
            (fun _ ->
              check_sl_only_lab_and_decr (_B1, _B2) (_A1, _A2) (_C1, _C2) gr r
                lR))
      (fun x ->
        error (comp (shows_string
                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't';
                        'h'; ' '; 'l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; ' '; 'T';
                        'R'; 'S'; ':'])
                (comp shows_nl x)));;

let rec rl_tt (_A1, _A2) (_B1, _B2)
  r lR =
    let fs = funs_trs_list (eq_laba _A1 _B1) r in
    bind (checkB (not (null fs))
           (comp (shows_string
                   ['r'; 'o'; 'o'; 't'; '-'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                     'g'; ' '; 'r'; 'e'; 'q'; 'u'; 'i'; 'r'; 'e'; 's'; ' '; 'a';
                     't'; ' '; 'l'; 'e'; 'a'; 's'; 't'; ' '; 'o'; 'n'; 'e'; ' ';
                     'f'; 'u'; 'n'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 's'; 'y'; 'm';
                     'b'; 'o'; 'l'; ' '; 'i'; 'n'; ' '; 't'; 'h'; 'e'; ' '; 's';
                     'i'; 'g'; 'n'; 'a'; 't'; 'u'; 'r'; 'e'])
             shows_nl))
      (fun _ ->
        let d = hd fs in
        check_sem_lab_tt (_A1, _A2) (_B1, _B2)
          (eq_lista eq_char, shows_listb shows_chara)
          (check_sl_lab_trs (_A1, _A2) (_B1, _B2)
            (eq_laba _A1 _B1, shows_laba _A2 _B2)
            (eq_lista eq_char, shows_listb shows_chara)
            (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun fs)
          (check_sl_model_lab_trs (_A1, _A2) (_B1, _B2)
            (eq_laba _A1 _B1, shows_laba _A2 _B2)
            (eq_lista eq_char, shows_listb shows_chara)
            (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun fs)
          (bot_fun bot_boola) (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun r
          lR);;

let rec funas_term_list'
  = function Var uu -> []
    | Fun (f, ts) -> (f, size_list ts) :: flat_map funas_term_list' ts;;

let rec funas_rule_list'
  r = append (funas_term_list' (fst r)) (funas_term_list' (snd r));;

let rec funas_trs_list _A
  trs = remdups (eqc _A eq_nata) (flat_map funas_rule_list' trs);;

let rec supteq_list
  = function Var x -> [Var x]
    | Fun (f, ts) -> Fun (f, ts) :: concat_map supteq_list ts;;

let rec dP_list _A _B
  r = let d = defs_list _A r in
      concat_map
        (fun lr ->
          let l = fst lr in
          let s = sharpa l in
          map (fun u -> (s, sharpa u))
            (filter
              (fun u ->
                not (suptp _A _B l u) &&
                  (not (is_var u) && member _A (the (root u)) d))
              (supteq_list (snd lr))))
        r;;

let rec iD_list r = map (fun lr -> (plaina (fst lr), plaina (snd lr))) r;;

let rec indent
  p s = flat_map
          (fun c -> (if ((c : char) = newline) then [newline; ' '] else [c]))
          (p s);;

let rec check_rule_reflecting (_A1, _A2)
  fcs rs rule =
    checkB
      (any (fun (l, r) ->
             any (fun c ->
                   eq_term _A1 (eq_lista eq_char) (fst rule) (ctxt_apply c l) &&
                     eq_term _A1 (eq_lista eq_char) (snd rule) (ctxt_apply c r))
               (Hole :: fcs))
        rs)
      (comp (shows_string ['t'; 'h'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
        (comp (shows_rule _A2 (shows_listb shows_chara) rule)
          (comp (shows_string
                  [' '; 'i'; 's'; ' '; 'n'; 'e'; 'i'; 't'; 'h'; 'e'; 'r'; ' ';
                    'c'; 'o'; 'n'; 't'; 'a'; 'i'; 'n'; 'e'; 'd'; ' '; 'i'; 'n';
                    ' '; 't'; 'h'; 'e'; ' '; 'o'; 'r'; 'i'; 'g'; 'i'; 'n'; 'a';
                    'l'; ' '; 's'; 'e'; 't'; ' '; 'o'; 'f'; ' '; 'r'; 'u'; 'l';
                    'e'; 's'; ' '; 'n'; 'o'; 'r'; ' '; 'o'; 'b'; 't'; 'a'; 'i';
                    'n'; 'e'; 'd'; ' '; 'b'; 'y'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                    'i'; 'n'; 'g'; ' '; 'a'; ' '; 'f'; 'l'; 'a'; 't'; ' '; 'c';
                    'o'; 'n'; 't'; 'e'; 'x'; 't'])
            shows_nl)));;

let rec check_rule_preserving (_A1, _A2)
  fcs rs' rule =
    checkB
      (member
         (eqc (eq_terma _A1 (eq_lista eq_char))
           (eq_terma _A1 (eq_lista eq_char)))
         rule rs'
        || forall
             (fun c ->
               member
                 (eqc (eq_terma _A1 (eq_lista eq_char))
                   (eq_terma _A1 (eq_lista eq_char)))
                 (ctxt_apply c (fst rule), ctxt_apply c (snd rule)) rs')
             fcs)
      (comp (shows_string ['t'; 'h'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
        (comp (shows_rule _A2 (shows_listb shows_chara) rule)
          (comp (shows_string
                  [' '; 'i'; 's'; ' '; 'n'; 'e'; 'i'; 't'; 'h'; 'e'; 'r'; ' ';
                    'c'; 'o'; 'n'; 't'; 'a'; 'i'; 'n'; 'e'; 'd'; ' '; 'i'; 'n';
                    ' '; 't'; 'h'; 'e'; ' '; 'r'; 'e'; 's'; 'u'; 'l'; 't'; 'i';
                    'n'; 'g'; ' '; 's'; 'e'; 't'; ' '; 'o'; 'f'; ' '; 'r'; 'u';
                    'l'; 'e'; 's'; ' '; 'n'; 'o'; 'r'; ' '; 'c'; 'l'; 'o'; 's';
                    'e'; 'd'; ' '; 'u'; 'n'; 'd'; 'e'; 'r'; ' '; 'a'; 'l'; 'l';
                    ' '; 'f'; 'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x';
                    't'; 's'])
            shows_nl)));;

let rec hole_at _A
  n i f x3 = match n, i, f, x3 with
    n, i, f, More (g, ss1, Hole, ss2) ->
      eqa _A g f &&
        (eq_nat (size_list ss1) i &&
          eq_nat (size_list ss2) (minus_nat (minus_nat n i) one_nat))
    | n, i, f, Hole -> false
    | n, i, f, More (v, va, More (vd, ve, vf, vg), vc) -> false;;

let rec check_flat_ctxt_complete (_A1, _A2)
  fcs fa =
    checkB
      (forall (fun i -> any (hole_at _A1 (snd fa) i (fst fa)) fcs)
        (upt Zero_nat (snd fa)))
      (comp (shows_string
              ['t'; 'h'; 'e'; ' '; 'l'; 'i'; 's'; 't'; ' '; 'o'; 'f'; ' '; 'f';
                'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 's'; ' ';
                'i'; 's'; ' '; 'i'; 'n'; 'c'; 'o'; 'm'; 'p'; 'l'; 'e'; 't';
                'e'])
        shows_nl);;

let rec shows_ctxt _A _B
  = function Hole -> shows_string ['['; ']']
    | More (f, ss1, d, ss2) ->
        comp (shows _A f)
          (comp (shows_string ['('])
            (comp (shows_list_gen (shows_term _A _B) [] [] [','] [','] ss1)
              (comp (shows_ctxt _A _B d)
                (shows_list_gen (shows_term _A _B) [')'] [','] [','] [')']
                  ss2))));;

let rec is_flat_ctxt_list _A _B
  vs fas x2 = match vs, fas, x2 with
    vs, fas, More (f, ss1, Hole, ss2) ->
      member (eqc _B eq_nata) (f, Suc (size_list (append ss1 ss2))) fas &&
        (forall is_var (append ss1 ss2) &&
          (distinct (eq_terma _B _A) (append ss1 ss2) &&
            null (list_inter _A (map the_var (append ss1 ss2)) vs)))
    | vs, fas, Hole -> false
    | vs, fas, More (v, va, More (vd, ve, vf, vg), vc) -> false;;

let rec check_is_flat_ctxt (_A1, _A2)
  vs fas c =
    checkB (is_flat_ctxt_list (eq_lista eq_char) _A1 vs fas c)
      (comp (shows_ctxt _A2 (shows_listb shows_chara) c)
        (comp (shows_string
                [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'f'; 'l';
                  'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'])
          shows_nl));;

let rec check_flat_ctxt (_A1, _A2)
  vs c =
    (match c
      with Hole ->
        error (comp (shows_ctxt _A2 (shows_listb shows_chara) c)
                (comp (shows_string
                        [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'f';
                          'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x';
                          't'])
                  shows_nl))
      | More (_, ss1, Hole, ss2) ->
        bind (bind (checkB
                     (distinct (eq_terma _A1 (eq_lista eq_char))
                       (append ss1 ss2))
                     (comp (shows_ctxt _A2 (shows_listb shows_chara) c)
                       (comp (shows_string
                               [' '; 'c'; 'o'; 'n'; 't'; 'a'; 'i'; 'n'; 's';
                                 ' '; 'd'; 'u'; 'p'; 'l'; 'i'; 'c'; 'a'; 't';
                                 'e'; ' '; 'v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l';
                                 'e'; 's'])
                         shows_nl)))
               (fun _ ->
                 checkB (forall is_var (append ss1 ss2))
                   (comp (shows_ctxt _A2 (shows_listb shows_chara) c)
                     (comp (shows_string
                             [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'l';
                               'a'; 't'; ','; ' '; 'i'; '.'; 'e'; '.'; ','; ' ';
                               'h'; 'a'; 's'; ' '; 'd'; 'e'; 'p'; 't'; 'h'; ' ';
                               'g'; 'r'; 'e'; 'a'; 't'; 'e'; 'r'; ' '; 't'; 'h';
                               'a'; 'n'; ' '; 'o'; 'n'; 'e'])
                       shows_nl))))
          (fun _ ->
            checkB
              (forall (fun t -> not (member (eq_lista eq_char) (the_var t) vs))
                (append ss1 ss2))
              (comp (shows_ctxt _A2 (shows_listb shows_chara) c)
                (comp (shows_string
                        [' '; 'h'; 'a'; 's'; ' '; 't'; 'o'; ' '; 'c'; 'o'; 'n';
                          't'; 'a'; 'i'; 'n'; ' '; 'o'; 'n'; 'l'; 'y'; ' '; 'f';
                          'r'; 'e'; 's'; 'h'; ' '; 'v'; 'a'; 'r'; 'i'; 'a'; 'b';
                          'l'; 'e'; 's'])
                  shows_nl)))
      | More (_, ss1, More (_, _, _, _), ss2) ->
        error (comp (shows_ctxt _A2 (shows_listb shows_chara) c)
                (comp (shows_string
                        [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'f';
                          'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x';
                          't'])
                  shows_nl)));;

let rec fcc_tt (_A1, _A2)
  r fcs lR =
    let vs = vars_trs_list (eq_lista eq_char) r in
    let fas = funas_trs_list _A1 r in
    bind (bind (bind (bind (bind (checkB (not (null fcs))
                                   (comp (shows_string
   ['a'; 't'; ' '; 'l'; 'e'; 'a'; 's'; 't'; ' '; 'o'; 'n'; 'e'; ' '; 'f'; 'l';
     'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; ' '; 'i'; 's'; ' '; 'r';
     'e'; 'q'; 'u'; 'i'; 'r'; 'e'; 'd'; ' '; 'f'; 'o'; 'r'; ' '; 't'; 'h'; 'e';
     ' '; 'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'u'; 'n'; 'd'; 'e'; 'r'; ' ';
     'f'; 'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 's'])
                                     shows_nl))
                             (fun _ ->
                               catch (checkAllMI (check_flat_ctxt (_A1, _A2) vs)
                                       fcs)
                                 (fun x -> error (snd x))))
                       (fun _ ->
                         catch (checkAllMI
                                 (check_is_flat_ctxt (_A1, _A2) vs fas) fcs)
                           (fun x -> error (snd x))))
                 (fun _ ->
                   catch (checkAllMI (check_flat_ctxt_complete (_A1, _A2) fcs)
                           fas)
                     (fun x -> error (snd x))))
           (fun _ ->
             catch (checkAllMI (check_rule_preserving (_A1, _A2) fcs lR) r)
               (fun x -> error (snd x))))
      (fun _ ->
        catch (checkAllMI (check_rule_reflecting (_A1, _A2) fcs r) lR)
          (fun x -> error (snd x)));;

let rec pi_full uu uv uw = true;;

let terminationGoal : xml
  = Xml (['g'; 'o'; 'a'; 'l'], [],
          [Xml (['t'; 'e'; 'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n'], [], [],
                 None)],
          None);;

let rec xml2substitution
  xml2name =
    xmlManyElements ['s'; 'u'; 'b'; 's'; 't'; 'i'; 't'; 'u'; 't'; 'i'; 'o'; 'n']
      (xml2elements ['s'; 'u'; 'b'; 's'; 't'; 'E'; 'n'; 't'; 'r'; 'y'] xml2var
        (xml2term xml2name) (fun a b -> (a, b)))
      id;;

let rec xmlMany2Elements
  reqName p1 p2 p3 combi (Xml (name, atts, cs, text)) =
    let ds = rev cs in
    (if eq_list eq_char name reqName &&
          (null atts &&
            (is_none text &&
              less_eq_nat (nat (Big_int.big_int_of_int 2)) (size_list cs)))
      then bind (mapMonad p1 (rev (tl (tl ds))))
             (fun a ->
               bind (p2 (nth ds one_nat))
                 (fun b ->
                   bind (p3 (nth ds Zero_nat)) (comp return (combi a b))))
      else error (concat
                   [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r';
                      'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e';
                      ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                      'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't';
                      ' '; '('; 'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
                     reqName; [')']; [newline];
                     show shows_xmla (Xml (name, atts, cs, text))]));;

let rec xml2trsDisproof
  xml2name =
    xmlOptions
      [(['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; 'C'; 'o'; 'n'; 'd'; 'i'; 't';
          'i'; 'o'; 'n'; 'V'; 'i'; 'o'; 'l'; 'a'; 't'; 'e'; 'd'],
         xml0elements
           ['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; 'C'; 'o'; 'n'; 'd'; 'i';
             't'; 'i'; 'o'; 'n'; 'V'; 'i'; 'o'; 'l'; 'a'; 't'; 'e'; 'd']
           NotWellFormed);
        (['l'; 'o'; 'o'; 'p'],
          xmlMany2Elements ['l'; 'o'; 'o'; 'p']
            (xml3elements ['r'; 'e'; 'd'; 'e'; 'x'] (xml2term xml2name) xml2pos
              (xml2rule xml2name) (fun t _ _ -> t))
            (xml2substitution xml2name) (xml2context xml2name)
            (fun ts sigma c -> Loop (ts, (c, sigma))))];;

let nonTerminationGoal : xml
  = Xml (['g'; 'o'; 'a'; 'l'], [],
          [Xml (['n'; 'o'; 'n'; 'T'; 'e'; 'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i';
                  'o'; 'n'],
                 [], [], None)],
          None);;

type ('a, 'b, 'c) certificationProblem =
  CertificationProblem of
    bool * (('a, 'c) term * ('a, 'c) term) list *
      (('a, 'b, 'c) tRSProof, ('a, 'c) tRSDisproof) either;;

let rec xml2certProblem
  xml2uname xml2name xml2sname xml2lsname (Xml (name, egal, cs, text)) =
    (if eq_list eq_char name
          ['c'; 'e'; 'r'; 't'; 'i'; 'f'; 'i'; 'c'; 'a'; 't'; 'i'; 'o'; 'n'; 'P';
            'r'; 'o'; 'b'; 'l'; 'e'; 'm']
          && (is_none text &&
               less_eq_nat (nat (Big_int.big_int_of_int 3)) (size_list cs))
      then bind (xml2trsInput xml2uname (nth cs one_nat))
             (fun trs ->
               (if eq_xml (nth cs Zero_nat) terminationGoal
                 then bind (xml1element ['p'; 'r'; 'o'; 'o'; 'f']
                             (xml2trsProof xml2name xml2sname xml2lsname) id
                             (nth cs (nat (Big_int.big_int_of_int 2))))
                        (fun p ->
                          return (CertificationProblem (true, trs, Left p)))
                 else (if eq_xml (nth cs Zero_nat) nonTerminationGoal
                        then bind (xml1element ['p'; 'r'; 'o'; 'o'; 'f']
                                    (xml2trsDisproof xml2uname) id
                                    (nth cs (nat (Big_int.big_int_of_int 2))))
                               (fun p ->
                                 return
                                   (CertificationProblem (false, trs, Right p)))
                        else error ['u'; 'n'; 'k'; 'n'; 'o'; 'w'; 'n'; ' '; 'g';
                                     'o'; 'a'; 'l'; ' '; 't'; 'y'; 'p'; 'e';
                                     ' '; 'i'; 'n'; ' '; 'c'; 'e'; 'r'; 't';
                                     'i'; 'f'; 'i'; 'c'; 'a'; 't'; 'i'; 'o';
                                     'n'; ' '; 'p'; 'r'; 'o'; 'b'; 'l'; 'e';
                                     'm'])))
      else error ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'p'; 'a';
                   'r'; 's'; 'e'; ' '; 'r'; 'o'; 'o'; 't'; ' '; 'e'; 'l'; 'e';
                   'm'; 'e'; 'n'; 't']);;

type 'a filtered = FPair of 'a * nat;;

let rec checkSubset _A
  xs ys =
    catch (checkAllMI (checkF (fun x -> member _A x ys) id) xs)
      (fun x -> error (snd x));;

let rec check_wf_trs _A (_B1, _B2)
  r = bind (catch
             (checkAllMI
               (fun lr ->
                 checkB (not (is_var (fst lr)))
                   (comp (shows_string
                           ['t'; 'h'; 'e'; ' '; 'l'; 'e'; 'f'; 't'; '-'; 'h';
                             'a'; 'n'; 'd'; ' '; 's'; 'i'; 'd'; 'e'; ' '; 'o';
                             'f'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                     (comp (shows_rule _A _B2 lr)
                       (comp (shows_string
                               [' '; 'i'; 's'; ' '; 'a'; ' '; 'v'; 'a'; 'r';
                                 'i'; 'a'; 'b'; 'l'; 'e'])
                         shows_nl))))
               r)
             (fun x -> error (snd x)))
        (fun _ ->
          catch (catch
                  (checkAllMI
                    (fun lr ->
                      catch (checkSubset _B1 (vars_term_list _B1 (snd lr))
                              (vars_term_list _B1 (fst lr)))
                        (fun x ->
                          error (comp (shows_string
['t'; 'h'; 'e'; ' '; 'v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; ' '])
                                  (comp (shows _B2 x)
                                    (comp (shows_string
    [' '; 'd'; 'o'; 'e'; 's'; ' '; 'o'; 'n'; 'l'; 'y'; ' '; 'o'; 'c'; 'c'; 'u';
      'r'; ' '; 'i'; 'n'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'i'; 'g'; 'h'; 't'; '-';
      'h'; 'a'; 'n'; 'd'; ' '; 's'; 'i'; 'd'; 'e'; ' '; 'o'; 'f'; ' '; 'r'; 'u';
      'l'; 'e'; ' '])
                                      (shows_rule _A _B2 lr))))))
                    r)
                  (fun x -> error (snd x)))
            (fun x ->
              error (comp (shows_string
                            ['t'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'; ' '; 'i'; 's';
                              ' '; 'n'; 'o'; 't'; ' '; 'w'; 'e'; 'l'; 'l'; '-';
                              'f'; 'o'; 'r'; 'm'; 'e'; 'd'])
                      (comp shows_nl x))));;

let rec check_DP (_A1, _A2) (_B1, _B2)
  r p = catch (bind (check_wf_trs _A2 (_B1, _B2) r)
                (fun _ ->
                  catch (checkSubset
                          (eqc (eq_terma (eq_shpa _A1) _B1)
                            (eq_terma (eq_shpa _A1) _B1))
                          (dP_list _A1 _B1 r) p)
                    (fun x ->
                      error (comp (shows_string
                                    ['t'; 'h'; 'e'; ' '; 'D'; 'P'; ' '])
                              (comp (shows_rule (shows_shpa _A2) _B2 x)
                                (comp (shows_string
[' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'p'; 'p'; 'e'; 'a'; 'r';
  ' '; 'i'; 'n'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b';
  'l'; 'e'; 'm'])
                                  shows_nl))))))
          (fun x ->
            error (comp (shows_string
                          ['t'; 'h'; 'e'; ' '; 'D'; 'P'; '-'; 't'; 'r'; 'a';
                            'n'; 's'; 'f'; 'o'; 'r'; 'm'; 'a'; 't'; 'i'; 'o';
                            'n'; ' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a';
                            'p'; 'p'; 'l'; 'i'; 'e'; 'd'; ' '; 'c'; 'o'; 'r';
                            'r'; 'e'; 'c'; 't'; 'l'; 'y'; '.'])
                    (comp shows_nl x)));;

let rec ge_nat x y = less_eq_nat y x;;

let rec mat_mono sd m = mat_monoI m (Suc (minus_nat sd one_nat));;

let nat_name : char list = ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'];;

let rec create_afI _A _B
  ze x1 f n i = match ze, x1, f, n, i with ze, [], f, n, i -> true
    | ze, ((g, m), (c, coeffs)) :: rest, f, n, i ->
        (if eqa _B f g && eq_nat m n
          then less_nat i (size_list coeffs) && not (eqa _A (nth coeffs i) ze)
          else create_afI _A _B ze rest f n i);;

let rec lpo_repr x = shows_string ['L'; 'P'; 'O'; '+'; 'A'; 'F'] x;;

let rec check_sem_lab_proc (_A1, _A2) (_B1, _B2) (_C1, _C2)
  check_lab check_model_lab gr i l p r lP lR =
    bind (bind (catch (check_lab lP p)
                 (fun x ->
                   error (comp (shows_string
                                 ['n'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; ' '; 'l';
                                   'a'; 'b'; 'e'; 'l'; 'e'; 'd'; ' '; 'D'; 'P';
                                   's'; ' '; 'p'; 'r'; 'e'; 's'; 'e'; 'n'; 't';
                                   ':'])
                           (comp shows_nl x))))
           (fun _ ->
             catch (check_model_lab lR r)
               (fun x ->
                 error (comp (shows_string
                               ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                 'i'; 't'; 'h'; ' '; 'l'; 'a'; 'b'; 'e'; 'l';
                                 'e'; 'd'; ' '; 'T'; 'R'; 'S'; ':'])
                         (comp shows_nl x)))))
      (fun _ ->
        catch (check_sl_only_lab_and_decr (_B1, _B2) (_A1, _A2) (_C1, _C2) gr r
                lR)
          (fun x -> error (comp x shows_nl)));;

let rec list_diff _A
  xs x1 = match xs, x1 with xs, [] -> xs
    | xs, y :: ys -> list_diff _A (removeAll _A y xs) ys;;

let rec rl_proc (_A1, _A2) (_B1, _B2)
  p r lP lR =
    let fs =
      list_union (eq_laba _A1 _B1) (funs_trs_list (eq_laba _A1 _B1) r)
        (list_diff (eq_laba _A1 _B1) (funs_trs_list (eq_laba _A1 _B1) p)
          (defs_list (eq_laba _A1 _B1) p))
      in
    bind (checkB (not (null fs))
           (comp (shows_string
                   ['r'; 'o'; 'o'; 't'; '-'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                     'g'; ' '; 'r'; 'e'; 'q'; 'u'; 'i'; 'r'; 'e'; 's'; ' '; 'a';
                     't'; ' '; 'l'; 'e'; 'a'; 's'; 't'; ' '; 'o'; 'n'; 'e'; ' ';
                     'f'; 'u'; 'n'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 's'; 'y'; 'm';
                     'b'; 'o'; 'l'; ' '; 'i'; 'n'; ' '; 't'; 'h'; 'e'; ' '; 's';
                     'i'; 'g'; 'n'; 'a'; 't'; 'u'; 'r'; 'e'])
             shows_nl))
      (fun _ ->
        let d = hd fs in
        check_sem_lab_proc (_A1, _A2) (_B1, _B2)
          (eq_lista eq_char, shows_listb shows_chara)
          (check_sl_lab_trs (_A1, _A2) (_B1, _B2)
            (eq_laba _A1 _B1, shows_laba _A2 _B2)
            (eq_lista eq_char, shows_listb shows_chara)
            (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun fs)
          (check_sl_model_lab_trs (_A1, _A2) (_B1, _B2)
            (eq_laba _A1 _B1, shows_laba _A2 _B2)
            (eq_lista eq_char, shows_listb shows_chara)
            (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun fs)
          (bot_fun bot_boola) (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun p r
          lP lR);;

let rec funas_term_list _A t = remdups (eqc _A eq_nata) (funas_term_list' t);;

let rec xml2labSharpName
  = function Zero_nat -> xmlChange xml2sharpName (fun a -> UnLab a)
    | Suc n ->
        xml2elements
          ['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l']
          (xml2labSharpName n) (xml2label (xml2labSharpName n))
          (fun f a ->
            (match a with Left aa -> Lab (f, aa)
              | Right aa -> FunLab (f, aa)));;

let rec xmldoc2certProblem
  (Xmldoc (header, xml)) =
    debug ['0']
      ['p'; 'a'; 'r'; 's'; 'i'; 'n'; 'g'; ' '; 'x'; 'm'; 'l'; ' '; 't'; 'o';
        ' '; 'c'; 'e'; 'r'; 't'; 'i'; 'f'; 'i'; 'c'; 'a'; 't'; 'i'; 'o'; 'n';
        ' '; 'p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm']
      (xml2certProblem (xml2text ['n'; 'a'; 'm'; 'e']) xml2plainName
        xml2sharpName xml2labSharpName xml);;

let rec parseCertProblem s = bind (parseXMLDoc s) xmldoc2certProblem;;

let rec checkAllLI_help
  uu x1 uv = match uu, x1, uv with uu, [], uv -> return ()
    | f, x :: xs, n -> bind (f x n) (fun _ -> checkAllLI_help f xs (Suc n));;

let rec checkAllLI f xs = checkAllLI_help f xs Zero_nat;;

let rec show_pvars (_A1, _A2) _B
  on x1 = match on, x1 with on, [] -> []
    | on, (x, c) :: vas ->
        append (if eqa _A1 on c then [] else show _A2 c) (show _B x) ::
          show_pvars (_A1, _A2) _B on vas;;

let rec af_entry_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, AFList lista -> f2 lista
    | f1, f2, Collapse nat -> f1 nat;;

let rec shows_afs _A
  af = shows_map
         (fun (a, b) ->
           let (f, n) = a in
           (fun e ->
             comp (shows_string ['p'; 'i'; '('])
               (comp (shows _A f)
                 (comp (shows_string ['/'])
                   (comp (shows_nat n)
                     (comp (shows_string [')'; ' '; '='; ' '])
                       (comp (match e with Collapse i -> shows_nat (Suc i)
                               | AFList ids ->
                                 shows_lista shows_nata
                                   (map (fun aa -> Suc aa) ids))
                         shows_nl))))))
             b)
         af;;

let rec get_arg
  t x1 = match t, x1 with t, Zero_nat -> t
    | t, Suc n -> nth (args t) n;;

let rec int_mono x = Big_int.le_big_int (Big_int.big_int_of_int 1) x;;

let rec max0_nat x = x;;

let rec nat_mono x = not (eq_nat x Zero_nat);;

let rec rat_mono
  x = less_eq_rat
        (Fract ((Big_int.big_int_of_int 1), (Big_int.big_int_of_int 1))) x;;

let rec lookup_rest _A
  x xa1 = match x, xa1 with x, [] -> None
    | x, (y, c) :: ycs ->
        (if eqa _A x y then Some (c, ycs)
          else (match lookup_rest _A x ycs with None -> None
                 | Some (d, yccs) -> Some (d, (y, c) :: yccs)));;

let rec shows_lpoly (_A1, _A2) _B
  ze on (LPoly (c, cs)) =
    (match show_pvars (_A1, _A2) _B on cs with [] -> shows _A2 c
      | a :: lista ->
        comp (if eqa _A1 c ze then id
               else comp (shows _A2 c) (shows_string [' '; '+'; ' ']))
          (shows_list_gen (shows_lista shows_chara) [] [] [' '; '+'; ' '] []
            (a :: lista)));;

let rec wf_fun_terms_list _A
  uu x1 = match uu, x1 with uu, Var uv -> true
    | siga, Fun (f, ts) ->
        member (eqc _A eq_nata) (f, size_list ts) siga &&
          forall (wf_fun_terms_list _A siga) ts;;

let rec apply_af_entry
  uu x1 ts = match uu, x1, ts with uu, Collapse i, ts -> nth ts i
    | f, AFList is, ts -> Fun (f, map (nth ts) is);;

let rec afs_term
  pi x1 = match pi, x1 with
    pi, Fun (f, ts) ->
      let l = size_list ts in
      apply_af_entry (FPair (f, l)) (pi f l) (map (afs_term pi) ts)
    | pi, Var x -> Var x;;

let rec search_len
  t s n =
    (if eq_list eq_char (take n t) s then Some ([], drop n t)
      else (match t with [] -> None
             | h :: tt ->
               (match search_len tt s n with None -> None
                 | Some (bef, aft) -> Some (h :: bef, aft))));;

let rec reverse_trs r = map (fun lr -> (snd lr, fst lr)) r;;

let rec fresh_strings_list
  name offset used n =
    take n
      (filter
        (fun s ->
          not (member (eq_lista eq_char) s (remdups (eq_lista eq_char) used)))
        (map (fun i -> append name (show shows_nata (plus_nat i offset)))
          (upt Zero_nat
            (plus_nat n (length_unique (eq_lista eq_char) used)))));;

let rec shows_arctic
  = function Num_arc i -> shows_int i
    | MinInfty -> shows_string ['-'; 'i'; 'n'; 'f'];;

let rec shows_list_arctic asa = shows_list_aux shows_arctic asa;;

let shows_arctica =
  ({shows = shows_arctic; shows_list = shows_list_arctic} : arctic shows);;

let rec check_pvarsI _B
  ze ge vas x3 = match ze, ge, vas, x3 with
    ze, ge, vas, [] ->
      catch (checkAllMI (fun va -> checkB (ge (snd va) ze) (fst va)) vas)
        (fun x -> error (snd x))
    | ze, ge, vas, (x, b) :: vbs ->
        let (a', vas') =
          (match lookup_rest _B x vas with None -> (ze, vas)
            | Some (a, ba) -> (a, ba))
          in
        bind (checkB (ge a' b) x) (fun _ -> check_pvarsI _B ze ge vas' vbs);;

let rec check_lpoly_sI (_A1, _A2) (_B1, _B2)
  mode ze on ge gt (LPoly (a, vas)) (LPoly (b, vbs)) =
    catch (bind (checkB (gt a b)
                  (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n';
                      ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g'; ' ';
                      'c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'; ' '; 'p'; 'a';
                      'r'; 't']))
            (fun _ ->
              catch (check_pvarsI _B1 ze (if mode then ge else gt) vas vbs)
                (fun x ->
                  error (comp (shows_string
                                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                  'h'; 'e'; 'n'; ' '; 'c'; 'o'; 'm'; 'p'; 'a';
                                  'r'; 'i'; 'n'; 'g'; ' '; 'c'; 'o'; 'e'; 'f';
                                  'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't'; 's'; ' ';
                                  'o'; 'f'; ' '; 'v'; 'a'; 'r'; 'i'; 'a'; 'b';
                                  'l'; 'e'; ' '])
                          (shows _B2 x)))))
      (fun x ->
        error (comp (shows_string
                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e';
                        'n'; ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g';
                        ' '])
                (comp (shows_lpoly (_A1, _A2) _B2 ze on (LPoly (a, vas)))
                  (comp (shows_string [' '; '>'; ' '])
                    (comp (shows_lpoly (_A1, _A2) _B2 ze on (LPoly (b, vbs)))
                      (comp shows_nl x))))));;

let rec check_polo_sI (_A1, _A2) _B (_C1, _C2)
  mode ze on pl ti mz ge gt pi (s, t) =
    let left = pleftI _A1 _C1 ze on pl ti mz pi s in
    let right = prightI _A1 _C1 ze on pl ti mz pi t in
    catch (check_lpoly_sI (_A1, _A2) (_C1, _C2) mode ze on ge gt left right)
      (fun x ->
        error (comp (shows_string
                      ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'e';
                        'n'; 's'; 'u'; 'r'; 'e'; ' '])
                (comp (shows_term _B _C2 s)
                  (comp (shows_string [' '; '>'; ' '])
                    (comp (shows_term _B _C2 t) (comp shows_nl x))))));;

let rec get_comp_apos
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = ap;;

let rec rev_term
  c x1 = match c, x1 with c, Var x -> ctxt_apply c (Var x)
    | c, Fun (f, [t]) -> rev_term (More (f, [], c, [])) t;;

let rec rev_rule (l, r) = (rev_term Hole l, rev_term Hole r);;

let rec filtered_fun (FPair (f, n)) = f;;

let rec shows_filtered _A ff = shows _A (filtered_fun ff);;

let rec shows_list_filtered _A ffs = shows_list_aux (shows_filtered _A) ffs;;

let rec shows_filtereda _A =
  ({shows = shows_filtered _A; shows_list = shows_list_filtered _A} :
    'a filtered shows);;

let rec afs_check _A _B
  rel pi g lr =
    let pl = afs_term pi (fst lr) in
    let pr = afs_term pi (snd lr) in
    catch (g (pl, pr))
      (fun x ->
        error (comp (shows_string
                      ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o';
                        'r'; 'i'; 'e'; 'n'; 't'; ' '])
                (comp (shows_term _A _B (fst lr))
                  (comp (shows_string [' '])
                    (comp rel
                      (comp (shows_string [' '])
                        (comp (shows_term _A _B (snd lr))
                          (comp shows_nl
                            (comp (shows_string ['p'; 'i'; '('; ' '])
                              (comp (shows_term _A _B (fst lr))
                                (comp (shows_string [' '; ')'; ' '; '='; ' '])
                                  (comp (shows_term (shows_filtereda _A) _B pl)
                                    (comp shows_nl
                                      (comp (shows_string ['p'; 'i'; '('; ' '])
(comp (shows_term _A _B (snd lr))
  (comp (shows_string [' '; ')'; ' '; '='; ' '])
    (comp (shows_term (shows_filtereda _A) _B pr)
      (comp shows_nl x))))))))))))))))));;

let rec checkAllCombinations
  c xs =
    foldR (fun x ys res ->
            catch (checkAllMI
                    (fun y ->
                      bind (catch (c x y) (fun xa -> error ((x, y), xa)))
                        (fun _ -> res))
                    ys)
              (fun xa -> error (snd xa)))
      xs (return ());;

let rec check_graph_decomp
  ss g rcs =
    bind (catch (checkAllCombinations (check_edges ss g) (map snd rcs))
           (fun x -> error (snd x)))
      (fun _ ->
        catch (checkAllMI (fun c -> check_edges ss g c c)
                (map snd (filter (fun rc -> not (fst rc)) rcs)))
          (fun x -> error (snd x)));;

let rec check_lpoly_nsI (_A1, _A2) (_B1, _B2)
  ze on ge (LPoly (a, vas)) (LPoly (b, vbs)) =
    catch (bind (checkB (ge a b)
                  (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n';
                      ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g'; ' ';
                      'c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'; ' '; 'p'; 'a';
                      'r'; 't'; 's']))
            (fun _ ->
              catch (check_pvarsI _B1 ze ge vas vbs)
                (fun x ->
                  error (comp (shows_string
                                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                  'h'; 'e'; 'n'; ' '; 'c'; 'o'; 'm'; 'p'; 'a';
                                  'r'; 'i'; 'n'; 'g'; ' '; 'c'; 'o'; 'e'; 'f';
                                  'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't'; 's'; ' ';
                                  'o'; 'f'; ' '; 'v'; 'a'; 'r'; 'i'; 'a'; 'b';
                                  'l'; 'e'; ' '])
                          (shows _B2 x)))))
      (fun x ->
        error (comp (shows_string
                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e';
                        'n'; ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g';
                        ' '])
                (comp (shows_lpoly (_A1, _A2) _B2 ze on (LPoly (a, vas)))
                  (comp (shows_string [' '; '>'; '='; ' '])
                    (comp (shows_lpoly (_A1, _A2) _B2 ze on (LPoly (b, vbs)))
                      (comp shows_nl x))))));;

let rec check_polo_nsI (_A1, _A2) _B (_C1, _C2)
  ze on pl ti mz ge pi (s, t) =
    let left = pleftI _A1 _C1 ze on pl ti mz pi s in
    let right = prightI _A1 _C1 ze on pl ti mz pi t in
    catch (check_lpoly_nsI (_A1, _A2) (_C1, _C2) ze on ge left right)
      (fun x ->
        error (comp (shows_string
                      ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'e';
                        'n'; 's'; 'u'; 'r'; 'e'; ' '])
                (comp (shows_term _B _C2 s)
                  (comp (shows_string [' '; '>'; '='; ' '])
                    (comp (shows_term _B _C2 t) (comp shows_nl x))))));;

let rec to_lpoly_inter _B
  on i f n =
    (match lookup (eqc _B eq_nata) (f, n) i with None -> (on, replicate n on)
      | Some cas -> cas);;

let rec internal_error
  x = comp (shows_string
             ['i'; 'n'; 't'; 'e'; 'r'; 'n'; 'a'; 'l'; ' '; 'e'; 'r'; 'r'; 'o';
               'r'])
        (comp shows_nl
          (comp (shows_string
                  ['p'; 'l'; 'e'; 'a'; 's'; 'e'; ' '; 'c'; 'o'; 'n'; 't'; 'a';
                    'c'; 't'; ' '; 'o'; 'n'; 'e'; ' '; 'o'; 'f'; ' '; 't'; 'h';
                    'e'; ' '; 'C'; 'e'; 'T'; 'A'; ' '; 'm'; 'e'; 'm'; 'b'; 'e';
                    'r'; 's'])
            (comp shows_nl
              (comp (shows_string
                      ['h'; 't'; 't'; 'p'; ':'; '/'; '/'; 'c'; 'l'; '-'; 'i';
                        'n'; 'f'; 'o'; 'r'; 'm'; 'a'; 't'; 'i'; 'k'; '.'; 'u';
                        'i'; 'b'; 'k'; '.'; 'a'; 'c'; '.'; 'a'; 't'; '/'; 's';
                        'o'; 'f'; 't'; 'w'; 'a'; 'r'; 'e'; '/'; 'c'; 'e'; 't';
                        'a'])
                (comp shows_nl
                  (comp (shows_string
                          ['c'; 'e'; 't'; 'a'; '@'; 'i'; 'n'; 'f'; 'o'; 'r';
                            'm'; 'a'; 't'; 'i'; 'k'; '.'; 'u'; 'i'; 'b'; 'k';
                            '.'; 'a'; 'c'; '.'; 'a'; 't'])
                    shows_nl))))))
        x;;

let rec checkRedPairProc _A _B
  checkSS checkNSS name representation p rP r =
    let (_, nSp) = partition (fun foo -> isValue (checkSS foo)) p in
    catch (bind (bind (catch
                        (catch (checkAllMI checkNSS r) (fun x -> error (snd x)))
                        (fun x ->
                          error (comp (shows_string
['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
  'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'T'; 'R'; 'S'])
                                  (comp shows_nl x))))
                  (fun _ ->
                    catch (catch (checkAllMI checkNSS nSp)
                            (fun x -> error (snd x)))
                      (fun x ->
                        error (comp (shows_string
                                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'D';
'P'; 's'])
                                (comp shows_nl x)))))
            (fun _ ->
              catch (checkSubset (eqc (eq_terma _A _B) (eq_terma _A _B)) nSp rP)
                (fun x ->
                  error (match checkSS x with Left err -> err
                          | Right _ -> internal_error))))
      (fun x ->
        error (comp (shows_string
                      ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a';
                        'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'e';
                        'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a'; 'i';
                        'r'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r';
                        ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f';
                        'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'])
                (comp shows_nl (comp representation (comp shows_nl x)))));;

let rec shows_lpo_repr _A
  prs = comp (shows_string
               ['L'; 'P'; 'O'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' ';
                 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' '; 'p'; 'r';
                 'e'; 'c'; 'e'; 'd'; 'e'; 'n'; 'c'; 'e'])
          (comp shows_nl
            (comp (shows_map
                    (fun (a, b) ->
                      let (f, n) = a in
                      (fun pr ->
                        comp (shows_string
                               ['p'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n'; 'c';
                                 'e'; '('])
                          (comp (shows _A f)
                            (comp (shows_string ['['])
                              (comp (shows_nat n)
                                (comp (shows_string [']'; ')'; ' '; '='; ' '])
                                  (comp (shows_nat pr) shows_nl))))))
                        b)
                    prs)
              (comp shows_nl
                (comp (shows_string
                        ['p'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n'; 'c'; 'e'; '(';
                          '_'; ')'; ' '; '='; ' '; '0'])
                  shows_nl))));;

let rec shows_lpo_af _A
  repr af =
    comp (shows_lpo_repr _A repr)
      (comp shows_nl
        (comp (shows_string
                ['i'; 'n'; ' '; 'c'; 'o'; 'm'; 'b'; 'i'; 'n'; 'a'; 't'; 'i';
                  'o'; 'n'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' ';
                  'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' '; 'a'; 'r';
                  'g'; 'u'; 'm'; 'e'; 'n'; 't'; ' '; 'f'; 'i'; 'l'; 't'; 'e';
                  'r'])
          (comp shows_nl (shows_afs _A af))));;

let rec check_supt (_A1, _A2) (_B1, _B2)
  lr = catch (checkB (suptp _A1 _B1 (fst lr) (snd lr))
               (comp (shows_string
                       ['t'; 'h'; 'e'; ' '; 'r'; 'h'; 's'; ' '; 'i'; 's'; ' ';
                         'n'; 'o'; 't'; ' '; 'a'; ' '; 'p'; 'r'; 'o'; 'p'; 'e';
                         'r'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; ' '; 'o';
                         'f'; ' '; 't'; 'h'; 'e'; ' '; 'l'; 'h'; 's'])
                 shows_nl))
         (fun x ->
           error (comp (shows_string
                         ['C'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o';
                           'r'; 'i'; 'e'; 'n'; 't'; ' '; 'r'; 'u'; 'l'; 'e';
                           ' '])
                   (comp (shows_rule _A2 _B2 lr) (comp shows_nl x))));;

let rec extract_goal_input
  s = debug ['0']
        ['e'; 'x'; 't'; 'r'; 'a'; 'c'; 't'; 'i'; 'n'; 'g'; ' '; 'g'; 'o'; 'a';
          'l'; ' '; 'a'; 'n'; 'd'; ' '; 'T'; 'R'; 'S'; ' '; 'f'; 'r'; 'o'; 'm';
          ' '; 'i'; 'n'; 'p'; 'u'; 't'; ' '; 's'; 't'; 'r'; 'i'; 'n'; 'g']
        (match
          search_len s ['<'; 'g'; 'o'; 'a'; 'l'; '>']
            (size_list ['<'; 'g'; 'o'; 'a'; 'l'; '>'])
          with None -> None
          | Some (_, mgoal) ->
            (match
              search_len mgoal ['<'; '/'; 'g'; 'o'; 'a'; 'l'; '>']
                (size_list ['<'; '/'; 'g'; 'o'; 'a'; 'l'; '>'])
              with None -> None
              | Some (goal, s') ->
                (match
                  search_len s' ['<'; 'i'; 'n'; 'p'; 'u'; 't'; '>']
                    (size_list ['<'; 'i'; 'n'; 'p'; 'u'; 't'; '>'])
                  with None -> None
                  | Some (_, minput) ->
                    (match
                      search_len minput ['<'; '/'; 'i'; 'n'; 'p'; 'u'; 't'; '>']
                        (size_list ['<'; '/'; 'i'; 'n'; 'p'; 'u'; 't'; '>'])
                      with None -> None
                      | Some (input, _) ->
                        Some (concat
                               [['<'; 'p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; '>'];
                                 ['<'; 'g'; 'o'; 'a'; 'l'; '>']; goal;
                                 ['<'; '/'; 'g'; 'o'; 'a'; 'l'; '>'];
                                 ['<'; 'i'; 'n'; 'p'; 'u'; 't'; '>']; input;
                                 ['<'; '/'; 'i'; 'n'; 'p'; 'u'; 't'; '>'];
                                 ['<'; '/'; 'p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm';
                                   '>']])))));;

let rec check_not_wf_trs _A (_B1, _B2)
  r = checkB (not (isValue (check_wf_trs _A (_B1, _B2) r)))
        (comp (shows_string
                ['T'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'; ' '; 'i'; 's'; ' '; 'w';
                  'e'; 'l'; 'l'; ' '; 'f'; 'o'; 'r'; 'm'; 'e'; 'd'])
          shows_nl);;

let rec checkTRSDisproof (_A1, _A2) (_B1, _B2)
  i r x2 = match i, r, x2 with
    i, r, Loop l ->
      debug (i []) ['L'; 'o'; 'o'; 'p']
        (catch (checkLoop (_A1, _A2) (_B1, _B2) r l)
          (fun x ->
            error (comp i
                    (comp (shows_string
                            [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i'; 'n';
                              ' '; 'l'; 'o'; 'o'; 'p'; ' '; 'p'; 'r'; 'o'; 'o';
                              'f'])
                      (comp shows_nl x)))))
    | i, r, NotWellFormed ->
        debug (i [])
          ['n'; 'o'; 't'; ' '; 'w'; 'e'; 'l'; 'l'; '-'; 'f'; 'o'; 'r'; 'm'; 'e';
            'd']
          (catch (check_not_wf_trs _A2 (_B1, _B2) r)
            (fun x ->
              error (comp i
                      (comp (shows_string
                              [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i'; 'n';
                                ' '; 'n'; 'o'; 't'; '-'; 'w'; 'e'; 'l'; 'l';
                                '-'; 'f'; 'o'; 'r'; 'm'; 'e'; 'd'; ' '; 'p';
                                'r'; 'o'; 'o'; 'f'])
                        (comp shows_nl x)))));;

let rec init_lab f = UnLab f;;

let rec unary_term
  = function Var uu -> true
    | Fun (uv, [t]) -> unary_term t
    | Fun (v, []) -> false
    | Fun (v, vb :: vd :: ve) -> false;;

let rec check_unary_signature
  r = catch (catch
              (checkAllMI
                (checkF (fun (l, ra) -> unary_term l && unary_term ra) id) r)
              (fun x -> error (snd x)))
        (fun _ ->
          error (shows_string
                  ['t'; 'h'; 'e'; ' '; 't'; 'r'; 's'; ' '; 'd'; 'o'; 'e'; 's';
                    ' '; 'n'; 'o'; 't'; ' '; 'h'; 'a'; 'v'; 'e'; ' '; 'a'; ' ';
                    'u'; 'n'; 'a'; 'r'; 'y'; ' '; 's'; 'i'; 'g'; 'n'; 'a'; 't';
                    'u'; 'r'; 'e']));;

let rec check_string_rev (_A1, _A2) (_B1, _B2)
  r rR =
    bind (check_unary_signature r)
      (fun _ ->
        catch (checkSubset (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                (map rev_rule r) rR)
          (fun x ->
            error (comp (shows_string
                          ['t'; 'h'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                    (comp (shows_rule _A2 _B2 x)
                      (comp (shows_string
                              [' '; 'i'; 's'; ' '; 'm'; 'i'; 's'; 's'; 'i'; 'n';
                                'g'])
                        shows_nl)))));;

let rec shows_arithFun
  = function Arg i -> comp (shows_string ['x'; '_']) (shows_nat (Suc i))
    | Const n -> shows_nat n
    | Sum fs -> shows_list_gen shows_arithFun [] [] [' '; '+'; ' '] [] fs
    | Prod fs -> shows_list_gen shows_arithFun [] [] [' '; '*'; ' '] [] fs
    | Min fs ->
        shows_list_gen shows_arithFun [] ['m'; 'i'; 'n'; '('] [','] [')'] fs
    | Max fs ->
        shows_list_gen shows_arithFun [] ['m'; 'a'; 'x'; '('] [','] [')'] fs;;

let rec shows_sl_inter _A _B
  (SL_Inter (n, fnas)) =
    comp (shows_string
           ['c'; 'a'; 'r'; 'r'; 'i'; 'e'; 'r'; ' '; '{'; '0'; ','; '.'; '.';
             '.'; ','])
      (comp (shows_nat n)
        (comp (shows_string ['}'])
          (comp shows_nl
            (shows_map
              (fun (a, b) ->
                let (f, na) = a in
                (fun aa ->
                  comp (shows_string ['['])
                    (comp (shows_lab _A _B f)
                      (comp (shows_string ['/'])
                        (comp (shows_nat na)
                          (comp (shows_string [']'; ' '; '='; ' '])
                            (comp (shows_arithFun aa) shows_nl))))))
                  b)
              fnas))));;

let rec take_default
  def x1 uu = match def, x1, uu with def, [], uu -> def
    | uv, x :: xs, Zero_nat -> x
    | def, uw :: xs, Suc i -> take_default def xs i;;

let rec eval_arithFun_unbound
  nats x1 = match nats, x1 with nats, Arg i -> take_default Zero_nat nats i
    | nats, Const n -> n
    | nats, Sum [] -> Zero_nat
    | nats, Sum (f :: fs) ->
        plus_nat (eval_arithFun_unbound nats f)
          (eval_arithFun_unbound nats (Sum fs))
    | nats, Prod [] -> one_nat
    | nats, Prod (f :: fs) ->
        times_nat (eval_arithFun_unbound nats f)
          (eval_arithFun_unbound nats (Prod fs))
    | nats, Max [f] -> eval_arithFun_unbound nats f
    | nats, Max (f :: v :: va) ->
        max ord_nat (eval_arithFun_unbound nats f)
          (eval_arithFun_unbound nats (Max (v :: va)))
    | nats, Min [f] -> eval_arithFun_unbound nats f
    | nats, Min (f :: v :: va) ->
        min ord_nat (eval_arithFun_unbound nats f)
          (eval_arithFun_unbound nats (Min (v :: va)));;

let rec eval_arithFun c f nats = mod_nat (eval_arithFun_unbound nats f) c;;

let rec sl_inter_to_inter _A _B
  (SL_Inter (c, ls)) fl cs =
    (match lookup (eqc (eq_laba _A _B) eq_nata) (fl, size_list cs) ls
      with None -> Zero_nat | Some f -> eval_arithFun (Suc c) f cs);;

let rec get_largest_element (SL_Inter (n, uu)) = n;;

let rec create_sem_lab_tt (_A1, _A2) (_B1, _B2) (_C1, _C2)
  sli l r lr =
    let c = get_largest_element sli in
    let ca = upt Zero_nat (Suc c) in
    let i = sl_inter_to_inter _A1 _B1 sli in
    catch (check_sem_lab_tt (_A1, _A2) (_B1, _B2) (_C1, _C2)
            (check_sl_lab_trs (_A1, _A2) (_B1, _B2) (eq_nata, shows_nata)
              (_C1, _C2) i l ca)
            (check_sl_model_lab_trs (_A1, _A2) (_B1, _B2) (eq_nata, shows_nata)
              (_C1, _C2) i l ca)
            (fun _ -> false) i l r lr)
      (fun x -> error (comp (shows_sl_inter _A2 _B2 sli) (comp shows_nl x)));;

let rec unlab_of_trsI _A _B _C
  r = remdups (eqc (eq_terma (eq_laba _A _B) _C) (eq_terma (eq_laba _A _B) _C))
        (map_trsI unlab
          (filter
            (fun (l, ra) ->
              eq_term (eq_laba _A _B) _C l ra ||
                not (eq_term (eq_laba _A _B) _C (map_term unlab l)
                      (map_term unlab ra)))
            r));;

let rec unlab_trs_tt (_A1, _A2) (_B1, _B2) (_C1, _C2)
  r rR =
    let u = unlab_of_trsI _A1 _B1 _C1 r in
    catch (bind (checkSubset
                  (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                    (eq_terma (eq_laba _A1 _B1) _C1))
                  rR u)
            (fun _ ->
              checkSubset
                (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                  (eq_terma (eq_laba _A1 _B1) _C1))
                u rR))
      (fun _ ->
        error (comp (shows_rules (shows_laba _A2 _B2) _C2 rR)
                (comp shows_nl
                  (comp (shows_string
                          [' '; ' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o'; 't';
                            ' '; 'c'; 'o'; 'r'; 'r'; 'e'; 's'; 'p'; 'o'; 'n';
                            'd'; ' '; 't'; 'o'; ' '; 'u'; 'n'; 'l'; 'a'; 'b';
                            'e'; 'l'; 'e'; 'd'; ' '; 'T'; 'R'; 'S'; ' '; ' '])
                    (comp shows_nl
                      (comp (shows_rules (shows_laba _A2 _B2) _C2 u)
                        shows_nl))))));;

let rec check_mono_afs_help _A
  ((f, a), e) =
    comp (shows_string ['e'; 'n'; 't'; 'r'; 'y'; ' '; 'f'; 'o'; 'r'; ' '])
      (comp (shows _A f)
        (comp (shows_string
                [' '; 'n'; 'o'; 't'; ' '; 'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n';
                  'i'; 'c'])
          shows_nl));;

let rec mono_af_entry
  n x1 = match n, x1 with n, Collapse i -> less_eq_nat n one_nat
    | n, AFList ids -> forall (fun i -> member eq_nata i ids) (upt Zero_nat n);;

let rec check_mono_afs _A
  pi = catch (catch
               (checkAllMI
                 (checkF (fun x -> mono_af_entry (snd (fst x)) (snd x)) id) pi)
               (fun x -> error (snd x)))
         (fun x -> error (check_mono_afs_help _A x));;

let rec checkRuleRemovalProc _A _B
  checkMono checkSS checkNSS name representation r rR =
    let (_, nSp) = partition (fun foo -> isValue (checkSS foo)) r in
    catch (bind (bind (bind (catch checkMono
                              (fun x ->
                                error (comp
(shows_string
  ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 'm'; 'o';
    'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; 'i'; 't'; 'y'; ' '; 'o'; 'f'; ' '; 's';
    't'; 'r'; 'i'; 'c'; 't'; ' '; 'o'; 'r'; 'd'; 'e'; 'r'])
(comp shows_nl x))))
                        (fun _ ->
                          catch (catch (checkAllMI checkNSS r)
                                  (fun x -> error (snd x)))
                            (fun x ->
                              error (comp (shows_string
    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r';
      'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'r'; 'u'; 'l'; 'e'; 's'])
                                      (comp shows_nl x)))))
                  (fun _ ->
                    catch (checkSubset (eqc (eq_terma _A _B) (eq_terma _A _B))
                            nSp rR)
                      (fun x ->
                        error (match checkSS x with Left err -> err
                                | Right _ -> internal_error))))
            (fun _ ->
              catch (checkSubset (eqc (eq_terma _A _B) (eq_terma _A _B)) rR r)
                (fun x ->
                  error (match checkSS x with Left err -> err
                          | Right _ -> internal_error))))
      (fun x ->
        error (comp (shows_string
                      ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a';
                        'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'u';
                        'l'; 'e'; ' '; 'r'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l'; ' ';
                        'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'; ' '; 'w';
                        'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o'; 'l';
                        'l'; 'o'; 'w'; 'i'; 'n'; 'g'])
                (comp shows_nl (comp representation (comp shows_nl x)))));;

let rec check_wf_afs _A
  pi = catch (checkAllMI
               (fun (a, b) ->
                 let (f, n) = a in
                 (fun e ->
                   catch (match e with Collapse i -> checkB (less_nat i n) i
                           | AFList ids ->
                             catch (checkAllMI
                                     (checkF (fun i -> less_nat i n) id) ids)
                               (fun x -> error (snd x)))
                     (fun x ->
                       error (comp (shows_string ['i'; 'n'; 'd'; 'e'; 'x'; ' '])
                               (comp (shows_nat x)
                                 (comp (shows_string
 [' '; 't'; 'o'; 'o'; ' '; 'h'; 'i'; 'g'; 'h'; ' '; 'f'; 'o'; 'r'; ' '; 'A';
   'F'; '-'; 'e'; 'n'; 't'; 'r'; 'y'; ' '; 'o'; 'f'])
                                   (comp (showsa _A shows_nata (f, n))
                                     shows_nl))))))
                   b)
               pi)
         (fun x -> error (snd x));;

let rec default_af_entry n = AFList (upt Zero_nat n);;

let rec afs_list_to_afs _A
  pi f n =
    (match lookup (eqc _A eq_nata) (f, n) pi with None -> default_af_entry n
      | Some entry -> entry);;

let rec lpo_repr_to_pr _A
  prs (FPair (f, a)) ignored =
    (match lookup (eqc _A eq_nata) (f, a) prs with None -> Zero_nat
      | Some x -> x);;

let rec get_max_arity _A
  r = max_list
        (map snd
          (concat_map
            (fun (l, ra) ->
              append (funas_term_list _A l) (funas_term_list _A ra))
            r));;

let nat_components :
  (nat -> bool) *
    (nat *
      (nat *
        ((nat -> nat -> nat) *
          ((nat -> nat -> nat) *
            ((nat -> nat) *
              ((nat -> nat -> bool) *
                ((nat -> nat -> bool) *
                  (bool * (char list * (nat -> bool))))))))))
  = ((fun _ -> true),
      (Zero_nat,
        (one_nat,
          (plus_nat,
            (times_nat,
              (max0_nat,
                (ge_nat,
                  ((fun x y -> less_nat y x),
                    (true, (nat_name, (fun _ -> true)))))))))));;

let rec create_lpoly_repr (_A1, _A2) (_B1, _B2)
  c i = let pi = to_lpoly_inter _B1 (get_comp_1 c) i in
        comp (shows_string
               ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'i'; 'n';
                 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'r'; 'a'; 't'; 'i'; 'o';
                 'n'; ' '; 'o'; 'v'; 'e'; 'r'; ' '])
          (comp (shows_string (get_comp_description c))
            (comp shows_nl
              (shows_sep
                (fun (f, n) ->
                  let t =
                    Fun (f, map (fun a -> Var a)
                              (fresh_strings_list ['x'; '_'] one_nat [] n))
                    in
                  comp (shows_string ['P'; 'o'; 'l'; '('])
                    (comp (shows_term _B2 (shows_listb shows_chara) t)
                      (comp (shows_string [')'; ' '; '='; ' '])
                        (shows_lpoly (_A1, _A2) (shows_listb shows_chara)
                          (get_comp_0 c) (get_comp_1 c)
                          (pleftI _A1 (eq_lista eq_char) (get_comp_0 c)
                            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
                            (get_comp_max0 c) pi t)))))
                shows_nl (remdups (eqc _B1 eq_nata) (map fst i)))));;

let rec check_poly_monoI _A _B
  ze ge cm =
    (fun xs ->
      catch (checkAllMI
              (fun (a, b) ->
                let (f, n) = a in
                (fun (c, cs) ->
                  catch (bind (bind (checkB (ge c ze)
                                      (comp
(shows_string
  ['c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'; ' '; 'p'; 'a'; 'r'; 't'; ' '])
(comp (shows _A c)
  (comp (shows_string
          [' '; 'm'; 'u'; 's'; 't'; ' '; 'b'; 'e'; ' '; 'a'; 't'; ' '; 'l'; 'e';
            'a'; 's'; 't'; ' '])
    (comp (shows _A ze) shows_nl)))))
                                (fun _ ->
                                  checkB (less_eq_nat n (size_list cs))
                                    (comp (shows_string
    ['t'; 'h'; 'e'; ' '; 'l'; 'a'; 's'; 't'; ' '; 'a'; 'r'; 'g'; 'u'; 'm'; 'e';
      'n'; 't'; ' '; 'i'; 's'; ' '; 'i'; 'g'; 'n'; 'o'; 'r'; 'e'; 'd'])
                                      shows_nl)))
                          (fun _ ->
                            catch (checkAllMI
                                    (fun d ->
                                      checkB (cm d)
(comp (shows_string
        ['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't'; ' '])
  (comp (shows _A d)
    (comp (shows_string
            [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w';
              'e'; 'd'])
      shows_nl))))
                                    cs)
                              (fun x -> error (snd x))))
                    (fun x ->
                      error (comp (shows_string
                                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
                                      'w'; 'i'; 't'; 'h'; ' '; 'm'; 'o'; 'n';
                                      'o'; 't'; 'o'; 'n'; 'i'; 'c'; 'i'; 't';
                                      'y'; ' '; 'd'; 'u'; 'e'; ' '; 't'; 'o';
                                      ' '; 'i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r';
                                      'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n'; ' ';
                                      'o'; 'f'; ' '])
                              (comp (shows _B f)
                                (comp (shows_string ['/'])
                                  (comp (shows_nat n) (comp shows_nl x)))))))
                  b)
              xs)
        (fun x -> error (snd x)));;

let rec get_comp_wf_elem
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = wfe;;

let rec check_lpoly_coeffs _A _B
  wfe apos ch i =
    catch (checkAllMI
            (fun (a, b) ->
              let (f, n) = a in
              (fun (c, cs) ->
                catch (bind (bind (checkB (wfe c)
                                    (comp (shows_string
    ['c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'; ' '; 'p'; 'a'; 'r'; 't'; ' '])
                                      (comp (shows _A c)
(comp (shows_string
        [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'w'; 'e'; 'l'; 'l'; '-'; 'f';
          'o'; 'r'; 'm'; 'e'; 'd'])
  shows_nl))))
                              (fun _ ->
                                checkB (apos c || any apos (take n cs))
                                  (comp (shows_string
  ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'i'; 'n'; 'd'; ' ';
    'p'; 'o'; 's'; 'i'; 't'; 'i'; 'v'; 'e'; ' '; 'e'; 'n'; 't'; 'r'; 'y'; ' ';
    'w'; 'h'; 'i'; 'c'; 'h'; ' '; 'i'; 's'; ' '; 'r'; 'e'; 'q'; 'u'; 'i'; 'r';
    'e'; 'd'; ' '; 'f'; 'o'; 'r'; ' '; 'a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'i';
    'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n'; 's'])
                                    shows_nl)))
                        (fun _ ->
                          catch (checkAllMI
                                  (fun aa ->
                                    checkB (ch aa && wfe aa)
                                      (comp
(shows_string ['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't'; ' '])
(comp (shows _A aa)
  (comp (shows_string
          [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w'; 'e';
            'd'])
    shows_nl))))
                                  cs)
                            (fun x -> error (snd x))))
                  (fun x ->
                    error (comp (shows_string
                                  ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                    'i'; 't'; 'h'; ' '; 'i'; 'n'; 't'; 'e'; 'r';
                                    'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n';
                                    ' '; 'o'; 'f'; ' '])
                            (comp (shows _B f)
                              (comp (shows_string ['/'])
                                (comp (shows_nat n) (comp shows_nl x)))))))
                b)
            i)
      (fun x -> error (snd x));;

let rec create_rule_removal_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c cm i r rR =
    let pi = to_lpoly_inter _B1 (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        checkRuleRemovalProc _B1 _C1
          (check_poly_monoI _A2 _B2 (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) r rR);;

let rec create_rule_removal_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_rule_removal_proc_comp (eq_int, shows_inta) (_A1, _A2) (_B1, _B2)
        int_components int_mono i
    | Nat_carrier i ->
        create_rule_removal_proc_comp (eq_nata, shows_nata) (_A1, _A2)
          (_B1, _B2) nat_components nat_mono i
    | Rat_carrier (delta, i) ->
        create_rule_removal_proc_comp (eq_rata, shows_rata) (_A1, _A2)
          (_B1, _B2) (rat_components delta) rat_mono i
    | Arctic_carrier i ->
        (fun _ _ ->
          error (comp (shows_string
                        ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                          'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                          'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                  shows_nl))
    | Arc_mat_carrier (n, i) ->
        (fun _ _ ->
          error (comp (shows_string
                        ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                          'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                          'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                  shows_nl))
    | Int_mat_carrier (n, sd, i) ->
        create_rule_removal_proc_comp
          (eq_lista (eq_lista eq_int), shows_listb (shows_listb shows_inta))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components)
          (mat_mono sd int_mono) i
    | Rat_mat_carrier (delta, n, sd, i) ->
        create_rule_removal_proc_comp
          (eq_lista (eq_lista eq_rata), shows_listb (shows_listb shows_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components delta))
          (mat_mono sd rat_mono) i
    | Lpo (pri, pi) ->
        (fun r rR ->
          let m = get_max_arity _A1 r in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              checkRuleRemovalProc _A1 _B1 (check_mono_afs _A2 pi)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (shows_filtereda _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (shows_filtereda _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) r rR));;

let rec computeRuleMap _A
  = function [] -> Some (fun _ -> [])
    | (Fun (f, ts), r) :: rules ->
        (match computeRuleMap _A rules with None -> None
          | Some rm -> Some (fun_upd _A rm f ((Fun (f, ts), r) :: rm f)))
    | (Var uu, uv) :: rules -> None;;

let rec afs_list_to_af _A
  pi f n =
    (fun i ->
      (match lookup (eqc _A eq_nata) (f, n) pi with None -> true
        | Some (Collapse a) -> eq_nat i a
        | Some (AFList a) -> member eq_nata i a));;

let rec matchCapRMBelow _A _B
  l rm (Fun (f, ts)) = gcMatchI _A _B (HFun (f, map (capRM _A _B rm) ts)) l;;

let rec check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2)
  uu uv uw x3 = match uu, uv, uw, x3 with uu, uv, uw, Var x -> return ()
    | rm, ur, pi, Fun (f, ts) ->
        let n = size_list ts in
        bind (checkAllLI
               (fun t i ->
                 (if pi f n i
                   then check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm ur
                          pi t
                   else return ()))
               ts)
          (fun _ ->
            catch (checkAllMI
                    (fun lr ->
                      checkB
                        (member (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) lr
                           ur
                          || not (matchCapRMBelow _A1 _B1 (fst lr) (Some rm)
                                   (Fun (f, ts))))
                        (comp (shows_string
                                ['d'; 'u'; 'e'; ' '; 't'; 'o'; ' '; 't'; 'h';
                                  'e'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm';
                                  ' '])
                          (comp (shows_term _A2 _B2 (Fun (f, ts)))
                            (comp (shows_string
                                    [' '; 'o'; 'f'; ' '; 's'; 'o'; 'm'; 'e';
                                      ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' ';
                                      'r'; 'h'; 's'; ','; ' '; 'r'; 'u'; 'l';
                                      'e'; ' '])
                              (comp (shows_rule _A2 _B2 lr)
                                (shows_string
                                  [' '; 's'; 'h'; 'o'; 'u'; 'l'; 'd'; ' '; 'b';
                                    'e'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e';
                                    '.']))))))
                    (rm f))
              (fun x -> error (snd x)));;

let rec check_sct_ur_af_proc (_A1, _A2) (_B1, _B2)
  checkSS checkNSS representation p gs rm rrm ur pi =
    (match rm
      with None ->
        error (comp (shows_string
                      ['T'; 'R'; 'S'; ' '; 'h'; 'a'; 's'; ' '; 'v'; 'a'; 'r';
                        'i'; 'a'; 'b'; 'l'; 'e'; 's'; ' '; 'a'; 's'; ' '; 'l';
                        'e'; 'f'; 't'; '-'; 'h'; 'a'; 'n'; 'd'; ' '; 's'; 'i';
                        'd'; 'e'; 's'])
                shows_nl)
      | Some rma ->
        catch (let ggs =
                 filter
                   (fun g ->
                     member (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) (fst g)
                       p)
                   gs
                 in
               bind (bind (bind (bind (catch
(catch
  (checkAllMI
    (fun lr ->
      check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rma ur pi (snd lr))
    ur)
  (fun x -> error (snd x)))
(fun x ->
  error (comp (shows_string
                ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'c';
                  'h'; 'e'; 'c'; 'k'; 'i'; 'n'; 'g'; ' '; 'c'; 'l'; 'o'; 's';
                  'u'; 'r'; 'e'; ' '; 'p'; 'r'; 'o'; 'p'; 'e'; 'r'; 't'; 'i';
                  'e'; 's'; ' '; 'o'; 'f'; ' '; 'r'; 'h'; 's'; ' '; 'o'; 'f';
                  ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e';
                  's'])
          (comp shows_nl x))))
                                  (fun _ ->
                                    catch (checkSubset
    (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) p (map fst ggs))
                                      (fun x ->
error (comp (shows_string
              ['t'; 'h'; 'e'; 'r'; 'e'; ' '; 'i'; 's'; ' '; 'n'; 'o'; ' '; 's';
                'i'; 'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g'; 'e'; ' '; 'g'; 'r';
                'a'; 'p'; 'h'; ' '; 'f'; 'o'; 'r'; ' '; 'D'; 'P'; ' '])
        (shows_rule _A2 _B2 x)))))
                            (fun _ ->
                              catch (checkAllMI
                                      (fun (a, b) ->
let (s, t) = a in
(fun (stri, nstri) ->
  catch (bind (bind (bind (checkB
                            (not (is_var s) &&
                              (not (is_var t) && null (rma (the (root t)))))
                            (comp (shows_string
                                    ['r'; 'o'; 'o'; 't'; 's'; ' '; 'o'; 'f';
                                      ' '; 'D'; 'P'; ' '; 'm'; 'u'; 's'; 't';
                                      ' '; 'b'; 'e'; ' '; 'c'; 'o'; 'n'; 's';
                                      't'; 'r'; 'u'; 'c'; 't'; 'o'; 'r'; 's'])
                              shows_nl))
                      (fun _ ->
                        catch (checkAllMI
                                (fun i ->
                                  checkB (less_eq_nat i (size_list (args s)))
                                    (comp (shows_string
    ['l'; 'e'; 'f'; 't'; '-'; 'i'; 'n'; 'd'; 'e'; 'x'; ' '; 't'; 'o'; ' '; 'l';
      'a'; 'r'; 'g'; 'e'])
                                      (comp (shows_nat i) shows_nl)))
                                (remdups eq_nata (map fst (append stri nstri))))
                          (fun x -> error (snd x))))
                (fun _ ->
                  catch (checkAllMI
                          (fun j ->
                            checkB
                              (less_eq_nat j (size_list (args t)) &&
                                isValue
                                  (check_ur_closed_term_rm_af (_A1, _A2)
                                    (_B1, _B2) rma ur pi (get_arg t j)))
                              (comp (shows_string
                                      ['r'; 'i'; 'g'; 'h'; 't'; '-'; 'i'; 'n';
'd'; 'e'; 'x'; ' '; 't'; 'o'; ' '; 'l'; 'a'; 'r'; 'g'; 'e'; ' '; 'o'; 'r'; ' ';
'a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; ' '; 'v'; 'i'; 'o'; 'l'; 'a'; 't'; 'e';
's'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; '-'; 'r'; 'u'; 'l'; 'e'; 's'; ' '; 'c';
'o'; 'n'; 'd'; 'i'; 't'; 'i'; 'o'; 'n'])
                                (shows_nat j)))
                          (remdups eq_nata (map snd (append stri nstri))))
                    (fun x -> error (snd x))))
          (fun _ ->
            let _ = args s in
            let _ = args t in
            bind (catch
                   (checkAllMI
                     (fun (i, j) ->
                       checkB (isValue (checkSS (get_arg s i, get_arg t j)))
                         (comp (shows_string
                                 ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                   'i'; 't'; 'h'; ' '; 'e'; 'd'; 'g'; 'e'; ' '])
                           (comp (shows_nat i)
                             (comp (shows_string [' '; '-'; 'S'; '-'; '>'; ' '])
                               (shows_nat j)))))
                     stri)
                   (fun x -> error (snd x)))
              (fun _ ->
                catch (checkAllMI
                        (fun (i, j) ->
                          checkB (isValue (checkNSS (get_arg s i, get_arg t j)))
                            (comp (shows_string
                                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
                                      'w'; 'i'; 't'; 'h'; ' '; 'e'; 'd'; 'g';
                                      'e'; ' '])
                              (comp (shows_nat i)
                                (comp (shows_string
[' '; '-'; 'N'; 'S'; '-'; '>'; ' '])
                                  (shows_nat j)))))
                        nstri)
                  (fun x -> error (snd x)))))
    (fun x ->
      error (comp (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; 's'; ' '; 'w'; 'i'; 't';
                      'h'; ' '; 'D'; 'P'; ' '])
              (comp (shows_rule _A2 _B2 (s, t)) (comp shows_nl x)))))
  b)
                                      ggs)
                                (fun x -> error (snd x))))
                      (fun _ ->
                        catch (catch (checkAllMI checkNSS ur)
                                (fun x -> error (snd x)))
                          (fun x ->
                            error (comp (shows_string
  ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r';
    'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' ';
    'r'; 'u'; 'l'; 'e'; 's'])
                                    (comp shows_nl x)))))
                 (fun _ ->
                   checkB
                     (check_SCT
                       (eqc (eq_terma _A1 _B1)
                         (eqc (eq_terma _A1 _B1)
                           (eqc (eq_hterm _A1 ()) (eq_hterm _A1 ()))))
                       eq_nata
                       (fun (_, (t, (_, ct))) (u, (_, (cu, _))) ->
                         gcMatchI _A1 _B1 ct u && gcMatchI _A1 _B1 cu t)
                       (map (fun (a, b) ->
                              let (s, t) = a in
                              (fun (stri, nstri) ->
                                let e =
                                  (s, (t,
(capRM _A1 _B1 rrm s, capRM _A1 _B1 rm t)))
                                  in
                                Scg (e, e, stri, nstri))
                                b)
                         ggs))
                     (comp (shows_string
                             ['s'; 'i'; 'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g';
                               'e'; ' '; 'a'; 'n'; 'a'; 'l'; 'y'; 's'; 'i'; 's';
                               ' '; 'f'; 'a'; 'i'; 'l'; 'e'; 'd'])
                       shows_nl)))
          (fun x ->
            error (comp (shows_string
                          ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' ';
                            'a'; 'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' ';
                            's'; 'i'; 'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g';
                            'e'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                            'r'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e';
                            ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g';
                            ' '])
                    (comp shows_nl
                      (comp representation
                        (comp shows_nl
                          (comp (shows_string
                                  ['f'; 'o'; 'r'; ' '; 't'; 'h'; 'e'; ' '; 'f';
                                    'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                                    'r'; 'e'; 'a'; 's'; 'o'; 'n'])
                            (comp shows_nl x))))))));;

let rec create_sct_ur_af_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c i p rm rrm ur graphs =
    let pi = to_lpoly_inter _B1 (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        check_sct_ur_af_proc (_B1, _B2) (_C1, _C2)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c))
              (comp (shows_string [')'])
                (comp shows_nl
                  (comp (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i)
                    shows_nl)))))
          p graphs rm rrm ur (create_afI _A1 _B1 (get_comp_0 c) i));;

let rec create_sct_ur_af_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_sct_ur_af_proc_comp (eq_int, shows_inta) (_A1, _A2) (_B1, _B2)
        int_components i
    | Nat_carrier i ->
        create_sct_ur_af_proc_comp (eq_nata, shows_nata) (_A1, _A2) (_B1, _B2)
          nat_components i
    | Rat_carrier (delta, i) ->
        create_sct_ur_af_proc_comp (eq_rata, shows_rata) (_A1, _A2) (_B1, _B2)
          (rat_components delta) i
    | Arctic_carrier i ->
        create_sct_ur_af_proc_comp (eq_arctica, shows_arctica) (_A1, _A2)
          (_B1, _B2) arctic_components i
    | Int_mat_carrier (n, sd, i) ->
        create_sct_ur_af_proc_comp
          (eq_lista (eq_lista eq_int), shows_listb (shows_listb shows_inta))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components) i
    | Rat_mat_carrier (delta, n, sd, i) ->
        create_sct_ur_af_proc_comp
          (eq_lista (eq_lista eq_rata), shows_listb (shows_listb shows_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components delta)) i
    | Arc_mat_carrier (n, i) ->
        create_sct_ur_af_proc_comp
          (eq_lista (eq_lista eq_arctica),
            shows_listb (shows_listb shows_arctica))
          (_A1, _A2) (_B1, _B2) (mat_arctic_components n) i
    | Lpo (pri, pi) ->
        (fun p rm rrm ur graphs ->
          let m = max ord_nat (get_max_arity _A1 ur) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              check_sct_ur_af_proc (_A1, _A2) (_B1, _B2)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (shows_filtereda _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (shows_filtereda _A2) (_B1, _B2) pr m))
                (shows_lpo_af _A2 pri pi) p graphs rm rrm ur
                (afs_list_to_af _A1 pi)));;

let rec check_supteq (_A1, _A2) (_B1, _B2)
  lr = catch (checkB (supteqp _A1 _B1 (fst lr) (snd lr))
               (comp (shows_string
                       ['t'; 'h'; 'e'; ' '; 'r'; 'h'; 's'; ' '; 'i'; 's'; ' ';
                         'n'; 'o'; 't'; ' '; 'a'; ' '; 's'; 'u'; 'b'; 't'; 'e';
                         'r'; 'm'; ' '; 'o'; 'f'; ' '; 't'; 'h'; 'e'; ' '; 'l';
                         'h'; 's'])
                 shows_nl))
         (fun x ->
           error (comp (shows_string
                         ['C'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o';
                           'r'; 'i'; 'e'; 'n'; 't'; ' '; 'r'; 'u'; 'l'; 'e';
                           ' '])
                   (comp (shows_rule _A2 _B2 lr) (comp shows_nl x))));;

let rec check_sct_subterm_proc (_A1, _A2) (_B1, _B2)
  p gs rm rrm =
    (match rm
      with None ->
        error (comp (shows_string
                      ['T'; 'R'; 'S'; ' '; 'h'; 'a'; 's'; ' '; 'v'; 'a'; 'r';
                        'i'; 'a'; 'b'; 'l'; 'e'; 's'; ' '; 'a'; 's'; ' '; 'l';
                        'e'; 'f'; 't'; '-'; 'h'; 'a'; 'n'; 'd'; ' '; 's'; 'i';
                        'd'; 'e'; 's'])
                shows_nl)
      | Some irm ->
        catch (bind (bind (catch
                            (checkSubset
                              (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) p
                              (map fst gs))
                            (fun x ->
                              error (comp (shows_string
    ['t'; 'h'; 'e'; 'r'; 'e'; ' '; 'i'; 's'; ' '; 'n'; 'o'; ' '; 's'; 'i'; 'z';
      'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g'; 'e'; ' '; 'g'; 'r'; 'a'; 'p'; 'h'; ' ';
      'f'; 'o'; 'r'; ' '; 'D'; 'P'; ' '])
                                      (comp (shows_rule _A2 _B2 x) shows_nl))))
                      (fun _ ->
                        catch (checkAllMI
                                (fun (a, b) ->
                                  let (s, t) = a in
                                  (fun (stri, nstri) ->
                                    catch (bind
    (checkB (not (is_var s) && (not (is_var t) && null (irm (the (root t)))))
      (comp (shows_string
              ['r'; 'o'; 'o'; 't'; 's'; ' '; 'o'; 'f'; ' '; 'D'; 'P'; ' '; 'm';
                'u'; 's'; 't'; ' '; 'b'; 'e'; ' '; 'c'; 'o'; 'n'; 's'; 't'; 'r';
                'u'; 'c'; 't'; 'o'; 'r'; 's'])
        shows_nl))
    (fun _ ->
      let n = size_list (args s) in
      let m = size_list (args t) in
      bind (catch
             (checkAllMI
               (fun (i, j) ->
                 checkB
                   (less_eq_nat i n &&
                     (less_eq_nat j m &&
                       isValue
                         (check_supt (_A1, _A2) (_B1, _B2)
                           (get_arg s i, get_arg t j))))
                   (comp (shows_string
                           ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i';
                             't'; 'h'; ' '; 'e'; 'd'; 'g'; 'e'; ' '])
                     (comp (shows_nat i)
                       (comp (shows_string [' '; '|'; '>'; ' '])
                         (comp (shows_nat j) shows_nl)))))
               stri)
             (fun x -> error (snd x)))
        (fun _ ->
          catch (checkAllMI
                  (fun (i, j) ->
                    checkB
                      (less_eq_nat i n &&
                        (less_eq_nat j m &&
                          isValue
                            (check_supteq (_A1, _A2) (_B1, _B2)
                              (get_arg s i, get_arg t j))))
                      (comp (shows_string
                              ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i';
                                't'; 'h'; ' '; 'e'; 'd'; 'g'; 'e'; ' '])
                        (comp (shows_nat i)
                          (comp (shows_string [' '; '|'; '>'; '='; ' '])
                            (comp (shows_nat j) shows_nl)))))
                  nstri)
            (fun x -> error (snd x)))))
                                      (fun x ->
error (comp (shows_string
              ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; 's'; ' '; 'w'; 'i'; 't'; 'h';
                ' '; 'D'; 'P'; ' '])
        (comp (shows_rule _A2 _B2 (s, t)) (comp shows_nl x)))))
                                    b)
                                gs)
                          (fun x -> error (snd x))))
                (fun _ ->
                  checkB
                    (check_SCT
                      (eqc (eq_terma _A1 _B1)
                        (eqc (eq_terma _A1 _B1)
                          (eqc (eq_hterm _A1 ()) (eq_hterm _A1 ()))))
                      eq_nata
                      (fun (_, (t, (_, ct))) (u, (_, (cu, _))) ->
                        gcMatchI _A1 _B1 ct u && gcMatchI _A1 _B1 cu t)
                      (map (fun (a, b) ->
                             let (s, t) = a in
                             (fun (stri, nstri) ->
                               let e =
                                 (s, (t, (capRM _A1 _B1 rrm s,
   capRM _A1 _B1 rm t)))
                                 in
                               Scg (e, e, stri, nstri))
                               b)
                        gs))
                    (comp (shows_string
                            ['s'; 'i'; 'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g';
                              'e'; ' '; 'a'; 'n'; 'a'; 'l'; 'y'; 's'; 'i'; 's';
                              ' '; 'f'; 'a'; 'i'; 'l'; 'e'; 'd'])
                      shows_nl)))
          (fun x ->
            error (comp (shows_string
                          ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' ';
                            'a'; 'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' ';
                            's'; 'i'; 'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g';
                            'e'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                            'r'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 's'; 'u'; 'b';
                            't'; 'e'; 'r'; 'm'; '-'; 'r'; 'e'; 'l'; 'a'; 't';
                            'i'; 'o'; 'n'; ' '; 'f'; 'o'; 'r'; ' '; 't'; 'h';
                            'e'; ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n';
                            'g'; ' '; 'r'; 'e'; 'a'; 's'; 'o'; 'n'])
                    (comp shows_nl x))));;

let rec create_sp_representation _A
  (SimpProj sp) =
    shows_map
      (fun (f, n) ->
        comp (shows_string ['p'; 'i'; '('])
          (comp (shows _A f)
            (comp (shows_string [')'; ' '; '='; ' '])
              (comp (shows_nat n) shows_nl))))
      sp;;

let rec lift2term
  p x1 = match p, x1 with p, Var x -> Var x
    | p, Fun (f, ss) -> nth ss (p f);;

let rec check_supteq_proj (_A1, _A2) (_B1, _B2)
  p lr =
    catch (checkB
            (supteqp _A1 _B1 (lift2term p (fst lr)) (lift2term p (snd lr)))
            (comp (shows_string
                    ['t'; 'h'; 'e'; ' '; 'p'; 'r'; 'o'; 'j'; 'e'; 'c'; 't'; 'e';
                      'd'; ' '; 'r'; 'h'; 's'; ' '; 'i'; 's'; ' '; 'n'; 'o';
                      't'; ' '; 'a'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm';
                      ' '; 'o'; 'f'; ' '; 't'; 'h'; 'e'; ' '; 'p'; 'r'; 'o';
                      'j'; 'e'; 'c'; 't'; 'e'; 'd'; ' '; 'l'; 'h'; 's'])
              shows_nl))
      (fun x ->
        error (comp (shows_string
                      ['C'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o';
                        'r'; 'i'; 'e'; 'n'; 't'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                (comp (showsa (shows_terma _A2 _B2) (shows_terma _A2 _B2) lr)
                  (comp shows_nl
                    (comp (shows_string ['i'; '.'; 'e'; '.'; ','; ' '])
                      (comp (shows_term _A2 _B2 (lift2term p (fst lr)))
                        (comp (shows_string [' '; '|'; '>'; '='; ' '])
                          (comp (shows_term _A2 _B2 (lift2term p (snd lr)))
                            (comp shows_nl x)))))))));;

let rec check_projection _A _B
  ds p x2 = match ds, p, x2 with
    ds, p, (Fun (f, ss), Fun (g, ts)) ->
      let ls = size_list ss in
      let lt = size_list ts in
      bind (bind (bind (bind (checkB (less_nat Zero_nat ls)
                               (comp (shows_string
                                       ['l'; 'h'; 's'; ' '; 'i'; 's'; ' '; 'a';
 ' '; 'c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'])
                                 shows_nl))
                         (fun _ ->
                           checkB (less_nat (p f) ls)
                             (comp (shows_nat (p f))
                               (comp (shows_string
                                       [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' ';
 'a'; 'n'; ' '; 'a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; ' '; 'p'; 'o'; 's'; 'i';
 't'; 'i'; 'o'; 'n'; ' '; 'o'; 'f'; ' '])
                                 (comp (shows_term _A _B (Fun (f, ss)))
                                   shows_nl)))))
                   (fun _ ->
                     checkB (less_nat Zero_nat lt)
                       (comp (shows_string
                               ['r'; 'h'; 's'; ' '; 'i'; 's'; ' '; 'a'; ' ';
                                 'c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'])
                         shows_nl)))
             (fun _ ->
               checkB (null (ds g))
                 (comp (shows _A g)
                   (comp (shows_string
                           [' '; 'i'; 's'; ' '; 'd'; 'e'; 'f'; 'i'; 'n'; 'e';
                             'd'; ' '; 'i'; 'n'; ' '; 'R'])
                     shows_nl))))
        (fun _ ->
          checkB (less_nat (p g) lt)
            (comp (shows_nat (p g))
              (comp (shows_string
                      [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'n'; ' ';
                        'a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; ' '; 'p'; 'o';
                        's'; 'i'; 't'; 'i'; 'o'; 'n'; ' '; 'o'; 'f'; ' '])
                (comp (shows_term _A _B (Fun (g, ts))) shows_nl))))
    | ds, p, (Var v, t) ->
        error (comp (shows_string
                      ['c'; 'a'; 'n'; 'n'; 'o'; 't'; ' '; 'p'; 'r'; 'o'; 'j';
                        'e'; 'c'; 't'; ' '; 'a'; ' '; 'v'; 'a'; 'r'; 'i'; 'a';
                        'b'; 'l'; 'e'])
                shows_nl)
    | ds, p, (s, Var v) ->
        error (comp (shows_string
                      ['c'; 'a'; 'n'; 'n'; 'o'; 't'; ' '; 'p'; 'r'; 'o'; 'j';
                        'e'; 'c'; 't'; ' '; 'a'; ' '; 'v'; 'a'; 'r'; 'i'; 'a';
                        'b'; 'l'; 'e'])
                shows_nl);;

let rec check_supt_proj (_A1, _A2) (_B1, _B2)
  p lr =
    catch (checkB (suptp _A1 _B1 (lift2term p (fst lr)) (lift2term p (snd lr)))
            (comp (shows_string
                    ['t'; 'h'; 'e'; ' '; 'p'; 'r'; 'o'; 'j'; 'e'; 'c'; 't'; 'e';
                      'd'; ' '; 'r'; 'h'; 's'; ' '; 'i'; 's'; ' '; 'n'; 'o';
                      't'; ' '; 'a'; ' '; 'p'; 'r'; 'o'; 'p'; 'e'; 'r'; ' ';
                      's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; ' '; 'o'; 'f'; ' ';
                      't'; 'h'; 'e'; ' '; 'p'; 'r'; 'o'; 'j'; 'e'; 'c'; 't';
                      'e'; 'd'; ' '; 'l'; 'h'; 's'])
              shows_nl))
      (fun x ->
        error (comp (shows_string
                      ['C'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o';
                        'r'; 'i'; 'e'; 'n'; 't'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                (comp (shows_rule _A2 _B2 lr)
                  (comp shows_nl
                    (comp (shows_string ['i'; '.'; 'e'; '.'; ','; ' '])
                      (comp (shows_term _A2 _B2 (lift2term p (fst lr)))
                        (comp (shows_string [' '; '|'; '>'; ' '])
                          (comp (shows_term _A2 _B2 (lift2term p (snd lr)))
                            (comp shows_nl x)))))))));;

let rec check_spsc_proc (_A1, _A2) (_B1, _B2)
  proj rep p rP r x5 = match proj, rep, p, rP, r, x5 with
    proj, rep, p, rP, r, Some rm ->
      let (_, nSp) =
        partition
          (fun ra -> isValue (check_supt_proj (_A1, _A2) (_B1, _B2) proj ra)) p
        in
      catch (bind (bind (catch
                          (catch
                            (checkAllMI (check_projection _A2 _B2 rm proj) p)
                            (fun x -> error (snd x)))
                          (fun x ->
                            error (comp (shows_string
  ['t'; 'h'; 'e'; ' '; 's'; 'u'; 'p'; 'p'; 'l'; 'i'; 'e'; 'd'; ' '; 'p'; 'r';
    'o'; 'j'; 'e'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 'i'; 's'; ' '; 'm'; 'a'; 'l';
    'f'; 'o'; 'r'; 'm'; 'e'; 'd'])
                                    (comp shows_nl x))))
                    (fun _ ->
                      catch (catch
                              (checkAllMI
                                (check_supteq_proj (_A1, _A2) (_B1, _B2) proj)
                                nSp)
                              (fun x -> error (snd x)))
                        (fun x ->
                          error (comp (shows_string
['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
  'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'D'; 'P'; 's'])
                                  (comp shows_nl x)))))
              (fun _ ->
                catch (checkSubset (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                        nSp rP)
                  (fun x ->
                    error (match check_supt_proj (_A1, _A2) (_B1, _B2) proj x
                            with Left err -> err | Right _ -> internal_error))))
        (fun x ->
          error (comp (shows_string
                        ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a';
                          'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 's'; 'u';
                          'b'; 't'; 'e'; 'r'; 'm'; ' '; 'c'; 'r'; 'i'; 't'; 'e';
                          'r'; 'i'; 'o'; 'n'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's';
                          's'; 'o'; 'r'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h';
                          'e'; ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g';
                          ' '; 's'; 'i'; 'm'; 'p'; 'l'; 'e'; ' '; 'p'; 'r'; 'o';
                          'j'; 'e'; 'c'; 't'; 'i'; 'o'; 'n'])
                  (comp shows_nl (comp rep (comp shows_nl x)))))
    | proj, rep, p, rP, r, None -> return ();;

let rec create_sp _A
  (SimpProj sp) f = (match lookup _A f sp with None -> Zero_nat | Some n -> n);;

let rec create_spsc_proc (_A1, _A2) (_B1, _B2)
  sp p rP r rm =
    let pi = create_sp _A1 sp in
    check_spsc_proc (_A1, _A2) (_B1, _B2) pi (create_sp_representation _A2 sp) p
      rP r rm;;

let rec create_sem_lab_proc (_A1, _A2) (_B1, _B2) (_C1, _C2)
  sli l r1 r2 r3 r4 =
    let c = get_largest_element sli in
    let ca = upt Zero_nat (Suc c) in
    let i = sl_inter_to_inter _A1 _B1 sli in
    catch (check_sem_lab_proc (_A1, _A2) (_B1, _B2) (_C1, _C2)
            (check_sl_lab_trs (_A1, _A2) (_B1, _B2) (eq_nata, shows_nata)
              (_C1, _C2) i l ca)
            (check_sl_model_lab_trs (_A1, _A2) (_B1, _B2) (eq_nata, shows_nata)
              (_C1, _C2) i l ca)
            (fun _ -> false) i l r1 r2 r3 r4)
      (fun x -> error (comp (shows_sl_inter _A2 _B2 sli) (comp shows_nl x)));;

let rec check_unlab_proc (_A1, _A2) (_B1, _B2) (_C1, _C2)
  p r ulP ulR =
    let ulr = unlab_of_trsI _A1 _B1 _C1 r in
    catch (bind (bind (checkSubset
                        (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                          (eq_terma (eq_laba _A1 _B1) _C1))
                        (map_trsI unlab p) ulP)
                  (fun _ ->
                    checkSubset
                      (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                        (eq_terma (eq_laba _A1 _B1) _C1))
                      ulr ulR))
            (fun _ ->
              checkSubset
                (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                  (eq_terma (eq_laba _A1 _B1) _C1))
                ulR ulr))
      (fun x ->
        error (comp (shows_string
                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't';
                        'h'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                (comp (shows_rule (shows_laba _A2 _B2) _C2 x) shows_nl)));;

let rec matchCapBelowI _A _B
  l r (Fun (f, ts)) = gcMatchI _A _B (HFun (f, map (capI _A _B r) ts)) l;;

let rec check_ur_closed_term_af (_A1, _A2) (_B1, _B2)
  uu uv uw x3 = match uu, uv, uw, x3 with uu, uv, uw, Var x -> return ()
    | r, ur, pi, Fun (f, ts) ->
        let n = size_list ts in
        bind (checkAllLI
               (fun t i ->
                 (if pi f n i
                   then check_ur_closed_term_af (_A1, _A2) (_B1, _B2) r ur pi t
                   else return ()))
               ts)
          (fun _ ->
            catch (checkAllMI
                    (fun lr ->
                      checkB
                        (member (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) lr
                           ur
                          || not (matchCapBelowI _A1 _B1 (fst lr) r
                                   (Fun (f, ts))))
                        (comp (shows_string
                                ['d'; 'u'; 'e'; ' '; 't'; 'o'; ' '; 't'; 'h';
                                  'e'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm';
                                  ' '])
                          (comp (shows_term _A2 _B2 (Fun (f, ts)))
                            (comp (shows_string
                                    [' '; 'o'; 'f'; ' '; 's'; 'o'; 'm'; 'e';
                                      ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' ';
                                      'r'; 'h'; 's'; ','; ' '; 'r'; 'u'; 'l';
                                      'e'; ' '])
                              (comp (shows_rule _A2 _B2 lr)
                                (shows_string
                                  [' '; 's'; 'h'; 'o'; 'u'; 'l'; 'd'; ' '; 'b';
                                    'e'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e';
                                    '.']))))))
                    r)
              (fun x -> error (snd x)));;

let rec check_ur_P_closed_af (_A1, _A2) (_B1, _B2)
  r ur p pi =
    bind (catch
           (catch
             (checkAllMI
               (fun lr ->
                 check_ur_closed_term_af (_A1, _A2) (_B1, _B2) r ur pi (snd lr))
               ur)
             (fun x -> error (snd x)))
           (fun x ->
             error (comp (shows_string
                           ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e'; 'n';
                             ' '; 'c'; 'h'; 'e'; 'c'; 'k'; 'i'; 'n'; 'g'; ' ';
                             'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p'; 'r';
                             'o'; 'p'; 'e'; 'r'; 't'; 'i'; 'e'; 's'; ' '; 'o';
                             'f'; ' '; 'r'; 'h'; 's'; ' '; 'o'; 'f'; ' '; 'u';
                             's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e';
                             's'])
                     (comp shows_nl x))))
      (fun _ ->
        catch (catch
                (checkAllMI
                  (fun st ->
                    check_ur_closed_term_af (_A1, _A2) (_B1, _B2) r ur pi
                      (snd st))
                  p)
                (fun x -> error (snd x)))
          (fun x ->
            error (comp (shows_string
                          ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e'; 'n';
                            ' '; 'c'; 'h'; 'e'; 'c'; 'k'; 'i'; 'n'; 'g'; ' ';
                            'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p'; 'r';
                            'o'; 'p'; 'e'; 'r'; 't'; 'i'; 'e'; 's'; ' '; 'o';
                            'f'; ' '; 'r'; 'h'; 's'; ' '; 'o'; 'f'; ' '; 'D';
                            'P'; 's'])
                    (comp shows_nl x))));;

let rec check_ur_P_closed_rm_af (_A1, _A2) (_B1, _B2)
  r rmo ur p pi =
    (match rmo with None -> check_ur_P_closed_af (_A1, _A2) (_B1, _B2) r ur p pi
      | Some rm ->
        bind (catch
               (catch
                 (checkAllMI
                   (fun lr ->
                     check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm ur pi
                       (snd lr))
                   ur)
                 (fun x -> error (snd x)))
               (fun x ->
                 error (comp (shows_string
                               ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e';
                                 'n'; ' '; 'c'; 'h'; 'e'; 'c'; 'k'; 'i'; 'n';
                                 'g'; ' '; 'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e';
                                 ' '; 'p'; 'r'; 'o'; 'p'; 'e'; 'r'; 't'; 'i';
                                 'e'; 's'; ' '; 'o'; 'f'; ' '; 'r'; 'h'; 's';
                                 ' '; 'o'; 'f'; ' '; 'u'; 's'; 'a'; 'b'; 'l';
                                 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'])
                         (comp shows_nl x))))
          (fun _ ->
            catch (catch
                    (checkAllMI
                      (fun st ->
                        check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm ur
                          pi (snd st))
                      p)
                    (fun x -> error (snd x)))
              (fun x ->
                error (comp (shows_string
                              ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e'; 'n';
                                ' '; 'c'; 'h'; 'e'; 'c'; 'k'; 'i'; 'n'; 'g';
                                ' '; 'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' ';
                                'p'; 'r'; 'o'; 'p'; 'e'; 'r'; 't'; 'i'; 'e';
                                's'; ' '; 'o'; 'f'; ' '; 'r'; 'h'; 's'; ' ';
                                'o'; 'f'; ' '; 'D'; 'P'; 's'])
                        (comp shows_nl x)))));;

let rec checkMonoRedPairUrProc (_A1, _A2) (_B1, _B2)
  checkMono checkSS checkNSS name representation p rP r rm rR ur =
    let us = concat_map (comp (funas_term_list _A1) snd) (append p ur) in
    let filt = filter (fun lr -> wf_fun_terms_list _A1 us (fst lr)) in
    let (_, nSp) = partition (fun foo -> isValue (checkSS foo)) p in
    let (_, nSr) = partition (fun foo -> isValue (checkSS foo)) ur in
    catch (bind (bind (bind (bind (bind (bind
  (catch checkMono
    (fun x ->
      error (comp (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't'; 'h';
                      ' '; 'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; 'i';
                      't'; 'y'; ' '; 'o'; 'f'; ' '; 's'; 't'; 'r'; 'i'; 'c';
                      't'; ' '; 'o'; 'r'; 'd'; 'e'; 'r'])
              (comp shows_nl x))))
  (fun _ ->
    catch (check_ur_P_closed_rm_af (_A1, _A2) (_B1, _B2) r rm ur p pi_full)
      (fun x ->
        error (comp (shows_string
                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't';
                        'h'; ' '; 'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p';
                        'r'; 'o'; 'p'; 'e'; 'r'; 't'; 'i'; 'e'; 's'; ' '; 'o';
                        'f'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u';
                        'l'; 'e'; 's'])
                (comp shows_nl x)))))
                                    (fun _ ->
                                      catch
(catch (checkAllMI checkNSS nSr) (fun x -> error (snd x)))
(fun x ->
  error (comp (shows_string
                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n';
                  ' '; 'o'; 'r'; 'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'u';
                  's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'])
          (comp shows_nl x)))))
                              (fun _ ->
                                catch (catch (checkAllMI checkNSS nSp)
(fun x -> error (snd x)))
                                  (fun x ->
                                    error (comp
    (shows_string
      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o';
        'r'; 'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'D'; 'P'; 's'])
    (comp shows_nl x)))))
                        (fun _ ->
                          catch (checkSubset
                                  (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) rR
                                  r)
                            (fun _ ->
                              error (comp (shows_string
    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ':'; ' '; 'a'; 'd'; 'd'; 'i'; 't'; 'i';
      'o'; 'n'; 'a'; 'l'; ' '; 'r'; 'u'; 'l'; 'e'; '('; 's'; ')'; ' '; 't'; 'o';
      ' '; 'T'; 'R'; 'S'; ' '; 'a'; 'd'; 'd'; 'e'; 'd'])
                                      shows_nl))))
                  (fun _ ->
                    catch (checkSubset
                            (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                            (filt nSr) rR)
                      (fun x ->
                        error (match checkSS x with Left err -> err
                                | Right _ -> internal_error))))
            (fun _ ->
              catch (checkSubset (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                      (filt nSp) rP)
                (fun x ->
                  error (match checkSS x with Left err -> err
                          | Right _ -> internal_error))))
      (fun x ->
        error (comp (shows_string
                      ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a';
                        'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; '('; 'm';
                        'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ')'; ' '; 'r';
                        'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a';
                        'i'; 'r'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                        'r'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' ';
                        'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'])
                (comp shows_nl
                  (comp representation
                    (comp shows_nl
                      (comp (shows_string
                              ['a'; 'n'; 'd'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o';
                                'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' '; 'u';
                                's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l';
                                'e'; 's'])
                        (comp (shows_rules _A2 _B2 ur)
                          (comp shows_nl x))))))));;

let rec create_mono_redpair_ur_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c cm i p rP r rm rR ur =
    let pi = to_lpoly_inter _B1 (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        checkMonoRedPairUrProc (_B1, _B2) (_C1, _C2)
          (check_poly_monoI _A2 _B2 (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) p rP r rm rR ur);;

let rec create_mono_redpair_ur_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_mono_redpair_ur_proc_comp (eq_int, shows_inta) (_A1, _A2)
        (_B1, _B2) int_components int_mono i
    | Nat_carrier i ->
        create_mono_redpair_ur_proc_comp (eq_nata, shows_nata) (_A1, _A2)
          (_B1, _B2) nat_components nat_mono i
    | Rat_carrier (delta, i) ->
        create_mono_redpair_ur_proc_comp (eq_rata, shows_rata) (_A1, _A2)
          (_B1, _B2) (rat_components delta) rat_mono i
    | Arctic_carrier i ->
        (fun _ _ _ _ _ _ ->
          error (comp (shows_string
                        ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                          'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                          'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                  shows_nl))
    | Arc_mat_carrier (n, i) ->
        (fun _ _ _ _ _ _ ->
          error (comp (shows_string
                        ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                          'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                          'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                  shows_nl))
    | Int_mat_carrier (n, sd, i) ->
        create_mono_redpair_ur_proc_comp
          (eq_lista (eq_lista eq_int), shows_listb (shows_listb shows_inta))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components)
          (mat_mono sd int_mono) i
    | Rat_mat_carrier (d, n, sd, i) ->
        create_mono_redpair_ur_proc_comp
          (eq_lista (eq_lista eq_rata), shows_listb (shows_listb shows_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components d))
          (mat_mono sd rat_mono) i
    | Lpo (pri, pi) ->
        (fun p rP r rR rm ur ->
          let m = max ord_nat (get_max_arity _A1 r) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              checkMonoRedPairUrProc (_A1, _A2) (_B1, _B2)
                (check_mono_afs _A2 pi)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (shows_filtereda _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (shows_filtereda _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) p rP r rR rm ur));;

let rec checkMonoRedPairProc _A _B
  checkMono checkSS checkNSS name representation p rP r rR =
    let (_, nSp) = partition (fun foo -> isValue (checkSS foo)) p in
    let (_, nSr) = partition (fun foo -> isValue (checkSS foo)) r in
    catch (bind (bind (bind (bind (bind (catch checkMono
  (fun x ->
    error (comp (shows_string
                  ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't'; 'h';
                    ' '; 'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; 'i'; 't';
                    'y'; ' '; 'o'; 'f'; ' '; 's'; 't'; 'r'; 'i'; 'c'; 't'; ' ';
                    'o'; 'r'; 'd'; 'e'; 'r'])
            (comp shows_nl x))))
                                    (fun _ ->
                                      catch
(catch (checkAllMI checkNSS nSr) (fun x -> error (snd x)))
(fun x ->
  error (comp (shows_string
                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n';
                  ' '; 'o'; 'r'; 'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'T';
                  'R'; 'S'])
          (comp shows_nl x)))))
                              (fun _ ->
                                catch (catch (checkAllMI checkNSS nSp)
(fun x -> error (snd x)))
                                  (fun x ->
                                    error (comp
    (shows_string
      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o';
        'r'; 'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'D'; 'P'; 's'])
    (comp shows_nl x)))))
                        (fun _ ->
                          catch (checkSubset
                                  (eqc (eq_terma _A _B) (eq_terma _A _B)) rR r)
                            (fun _ ->
                              error (shows_string
                                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ':';
' '; 'a'; 'd'; 'd'; 'i'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; ' ']))))
                  (fun _ ->
                    catch (checkSubset (eqc (eq_terma _A _B) (eq_terma _A _B))
                            nSr rR)
                      (fun x ->
                        error (match checkSS x with Left err -> err
                                | Right _ -> internal_error))))
            (fun _ ->
              catch (checkSubset (eqc (eq_terma _A _B) (eq_terma _A _B)) nSp rP)
                (fun x ->
                  error (match checkSS x with Left err -> err
                          | Right _ -> internal_error))))
      (fun x ->
        error (comp (shows_string
                      ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a';
                        'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'e';
                        'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a'; 'i';
                        'r'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r';
                        ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f';
                        'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'])
                (comp shows_nl (comp representation (comp shows_nl x)))));;

let rec create_mono_redpair_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c cm i p rP r rR =
    let pi = to_lpoly_inter _B1 (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        checkMonoRedPairProc _B1 _C1
          (check_poly_monoI _A2 _B2 (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) p rP r rR);;

let rec create_mono_redpair_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_mono_redpair_proc_comp (eq_int, shows_inta) (_A1, _A2) (_B1, _B2)
        int_components int_mono i
    | Nat_carrier i ->
        create_mono_redpair_proc_comp (eq_nata, shows_nata) (_A1, _A2)
          (_B1, _B2) nat_components nat_mono i
    | Rat_carrier (delta, i) ->
        create_mono_redpair_proc_comp (eq_rata, shows_rata) (_A1, _A2)
          (_B1, _B2) (rat_components delta) rat_mono i
    | Arctic_carrier i ->
        (fun _ _ _ _ ->
          error (comp (shows_string
                        ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                          'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                          'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                  shows_nl))
    | Arc_mat_carrier (n, i) ->
        (fun _ _ _ _ ->
          error (comp (shows_string
                        ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                          'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                          'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                  shows_nl))
    | Int_mat_carrier (n, sd, i) ->
        create_mono_redpair_proc_comp
          (eq_lista (eq_lista eq_int), shows_listb (shows_listb shows_inta))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components)
          (mat_mono sd int_mono) i
    | Rat_mat_carrier (d, n, sd, i) ->
        create_mono_redpair_proc_comp
          (eq_lista (eq_lista eq_rata), shows_listb (shows_listb shows_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components d))
          (mat_mono sd rat_mono) i
    | Lpo (pri, pi) ->
        (fun p rP r rR ->
          let m = max ord_nat (get_max_arity _A1 r) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              checkMonoRedPairProc _A1 _B1 (check_mono_afs _A2 pi)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (shows_filtereda _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (shows_filtereda _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) p rP r rR));;

let rec checkAllMIndex
  g f xs =
    snd (foldl (fun (n, mon) x -> (g n x, bind mon (fun _ -> f n x)))
          (one_nat, return ()) xs);;

let rec checkDepGraphProc (_A1, _A2) (_B1, _B2)
  p dps rm rrm =
    bind (catch
           (checkSubset (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) p
             (concat_map snd dps))
           (fun x ->
             error (comp (shows_string
                           ['D'; 'e'; 'p'; 'e'; 'n'; 'd'; 'e'; 'n'; 'c'; 'y';
                             ' '; 'P'; 'a'; 'i'; 'r'; ' '])
                     (comp (shows_rule _A2 _B2 x)
                       (comp (shows_string
                               [' '; 'i'; 's'; ' '; 'm'; 'i'; 's'; 's'; 'i';
                                 'n'; 'g'; ' '; 'i'; 'n'; ' '; 'd'; 'e'; 'c';
                                 'o'; 'm'; 'p'; 'o'; 's'; 'i'; 't'; 'i'; 'o';
                                 'n'])
                         shows_nl)))))
      (fun _ ->
        catch (check_graph_decomp
                (comp (showsa (shows_terma _A2 _B2) (shows_terma _A2 _B2)) fst)
                (fun (a, b) ->
                  let (aa, ba) = a in
                  let (_, t) = aa in
                  (fun (_, ct) (ab, bb) ->
                    let (u, _) = ab in
                    (fun (cu, _) ->
                      gcMatchI _A1 _B1 ct u && gcMatchI _A1 _B1 cu t)
                      bb)
                    ba
                    b)
                (map (fun (real, cs) ->
                       (real,
                         map (fun (s, t) ->
                               ((s, t),
                                 (capRM _A1 _B1 rrm s, capRM _A1 _B1 rm t)))
                           cs))
                  dps))
          (fun x ->
            error (comp (shows_string
                          ['o'; 'u'; 'r'; ' '; 'e'; 's'; 't'; 'i'; 'm'; 'a';
                            't'; 'i'; 'o'; 'n'; ' '; '('; 'E'; 'D'; 'G'; '*';
                            '*'; '*'; ')'; ' '; 'c'; 'o'; 'u'; 'l'; 'd'; ' ';
                            'n'; 'o'; 't'; ' '; 's'; 'h'; 'o'; 'w'; ' '; 't';
                            'h'; 'a'; 't'; ' '; 'y'; 'o'; 'u'; ' '; 'g'; 'a';
                            'v'; 'e'; ' '; 'a'; ' '; 'v'; 'a'; 'l'; 'i'; 'd';
                            ' '; 'd'; 'e'; 'c'; 'o'; 'm'; 'p'; 'o'; 's'; 'i';
                            't'; 'i'; 'o'; 'n'; ' '; 'd'; 'u'; 'e'; ' '; 't';
                            'o'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o'; 'l'; 'l';
                            'o'; 'w'; 'i'; 'n'; 'g'; ' '; 'r'; 'e'; 'a'; 's';
                            'o'; 'n'])
                    (comp shows_nl x))));;

let rec checkRedPairUrAfProc (_A1, _A2) (_B1, _B2)
  checkSS checkNSS name representation p rP r rm ur pi =
    let (_, nSp) = partition (fun foo -> isValue (checkSS foo)) p in
    bind (bind (bind (catch
                       (check_ur_P_closed_rm_af (_A1, _A2) (_B1, _B2) r rm ur p
                         pi)
                       (fun x ->
                         error (comp (shows_string
                                       ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
 'w'; 'i'; 't'; 'h'; ' '; 'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p'; 'r'; 'o';
 'p'; 'e'; 'r'; 't'; 'i'; 'e'; 's'; ' '; 'o'; 'f'; ' '; 'u'; 's'; 'a'; 'b'; 'l';
 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'; ' '; 'w'; '.'; 'r'; '.'; 't'; '.'; ' '; 'a';
 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; ' '; 'f'; 'i'; 'l'; 't'; 'e'; 'r'])
                                 (comp shows_nl x))))
                 (fun _ ->
                   catch (catch (checkAllMI checkNSS ur)
                           (fun x -> error (snd x)))
                     (fun x ->
                       error (comp (shows_string
                                     ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
                                       'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
                                       'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'u';
                                       's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u';
                                       'l'; 'e'; 's'])
                               (comp shows_nl x)))))
           (fun _ ->
             catch (catch (checkAllMI checkNSS nSp) (fun x -> error (snd x)))
               (fun x ->
                 error (comp (shows_string
                               ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i'; 'e'; 'n';
                                 't'; 'i'; 'n'; 'g'; ' '; 'D'; 'P'; 's'])
                         (comp shows_nl x)))))
      (fun _ ->
        catch (catch
                (checkSubset (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) nSp rP)
                (fun x ->
                  error (match checkSS x with Left err -> err
                          | Right _ -> internal_error)))
          (fun x ->
            error (comp (shows_string
                          ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' ';
                            'a'; 'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' ';
                            'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' ';
                            'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r'; 'o'; 'c'; 'e';
                            's'; 's'; 'o'; 'r'; ' '; 'w'; 'i'; 't'; 'h'; ' ';
                            't'; 'h'; 'e'; ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w';
                            'i'; 'n'; 'g'])
                    (comp shows_nl
                      (comp representation
                        (comp shows_nl
                          (comp (shows_string
                                  ['a'; 'n'; 'd'; ' '; 't'; 'h'; 'e'; ' '; 'f';
                                    'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                                    'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u';
                                    'l'; 'e'; 's'])
                            (comp shows_nl
                              (comp (shows_rules _A2 _B2 ur)
                                (comp shows_nl x))))))))));;

let rec create_redpair_ur_af_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c i p rP r rm ur =
    let pi = to_lpoly_inter _B1 (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        checkRedPairUrAfProc (_B1, _B2) (_C1, _C2)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) p rP r rm ur
          (create_afI _A1 _B1 (get_comp_0 c) i));;

let rec create_redpair_ur_af_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_redpair_ur_af_proc_comp (eq_int, shows_inta) (_A1, _A2) (_B1, _B2)
        int_components i
    | Nat_carrier i ->
        create_redpair_ur_af_proc_comp (eq_nata, shows_nata) (_A1, _A2)
          (_B1, _B2) nat_components i
    | Rat_carrier (delta, i) ->
        create_redpair_ur_af_proc_comp (eq_rata, shows_rata) (_A1, _A2)
          (_B1, _B2) (rat_components delta) i
    | Arctic_carrier i ->
        create_redpair_ur_af_proc_comp (eq_arctica, shows_arctica) (_A1, _A2)
          (_B1, _B2) arctic_components i
    | Int_mat_carrier (n, sd, i) ->
        create_redpair_ur_af_proc_comp
          (eq_lista (eq_lista eq_int), shows_listb (shows_listb shows_inta))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components) i
    | Rat_mat_carrier (delta, n, sd, i) ->
        create_redpair_ur_af_proc_comp
          (eq_lista (eq_lista eq_rata), shows_listb (shows_listb shows_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components delta)) i
    | Arc_mat_carrier (n, i) ->
        create_redpair_ur_af_proc_comp
          (eq_lista (eq_lista eq_arctica),
            shows_listb (shows_listb shows_arctica))
          (_A1, _A2) (_B1, _B2) (mat_arctic_components n) i
    | Lpo (pri, pi) ->
        (fun p rP r rm ur ->
          let m = max ord_nat (get_max_arity _A1 r) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              checkRedPairUrAfProc (_A1, _A2) (_B1, _B2)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (shows_filtereda _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (shows_filtereda _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) p rP r rm ur
                (afs_list_to_af _A1 pi)));;

let rec create_redpair_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c i p rP r =
    let pi = to_lpoly_inter _B1 (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        checkRedPairProc _B1 _C1
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) p rP r);;

let rec create_redpair_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_redpair_proc_comp (eq_int, shows_inta) (_A1, _A2) (_B1, _B2)
        int_components i
    | Nat_carrier i ->
        create_redpair_proc_comp (eq_nata, shows_nata) (_A1, _A2) (_B1, _B2)
          nat_components i
    | Rat_carrier (delta, i) ->
        create_redpair_proc_comp (eq_rata, shows_rata) (_A1, _A2) (_B1, _B2)
          (rat_components delta) i
    | Arctic_carrier i ->
        create_redpair_proc_comp (eq_arctica, shows_arctica) (_A1, _A2)
          (_B1, _B2) arctic_components i
    | Int_mat_carrier (n, sd, i) ->
        create_redpair_proc_comp
          (eq_lista (eq_lista eq_int), shows_listb (shows_listb shows_inta))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components) i
    | Rat_mat_carrier (delta, n, sd, i) ->
        create_redpair_proc_comp
          (eq_lista (eq_lista eq_rata), shows_listb (shows_listb shows_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components delta)) i
    | Arc_mat_carrier (n, i) ->
        create_redpair_proc_comp
          (eq_lista (eq_lista eq_arctica),
            shows_listb (shows_listb shows_arctica))
          (_A1, _A2) (_B1, _B2) (mat_arctic_components n) i
    | Lpo (pri, pi) ->
        (fun p rP r ->
          let m = max ord_nat (get_max_arity _A1 r) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              checkRedPairProc _A1 _B1
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (shows_filtereda _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (shows_filtereda _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) p rP r));;

let rec checkDPProof (_A1, _A2) (_B1, _B2)
  i p r m rm rrm x6 = match i, p, r, m, rm, rrm, x6 with
    i, p, r, m, rm, rrm, PisEmpty ->
      debug (i []) ['P'; ' '; 'i'; 's'; ' '; 'e'; 'm'; 'p'; 't'; 'y']
        (match p with [] -> return ()
          | _ :: _ ->
            error (comp i
                    (comp (shows_string
                            [':'; ' '; 'P'; ' '; 'i'; 's'; ' '; 'n'; 'o'; 't';
                              ' '; 'e'; 'm'; 'p'; 't'; 'y'; ' '; 'i'; 'n'; ' ';
                              't'; 'h'; 'e'; ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w';
                              'i'; 'n'; 'g'; ' '; 'D'; 'P'; '-'; 'p'; 'r'; 'o';
                              'b'; 'l'; 'e'; 'm'; ' '])
                      (comp shows_nl
                        (shows_dpp (shows_laba _A2 _B2)
                          (shows_listb shows_chara) p r)))))
    | i, p, r, m, rm, rrm, RedPairProc (redp, rP, prf) ->
        debug (i []) ['R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (create_redpair_proc (eq_laba _A1 _B1, shows_laba _A2 _B2)
                    (eq_lista eq_char, shows_listb shows_chara) redp p rP r)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 't'; 'h';
                                      'e'; ' '; 'r'; 'e'; 'd'; 'u'; 'c'; 't';
                                      'i'; 'o'; 'n'; ' '; 'p'; 'a'; 'i'; 'r';
                                      ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's';
                                      'o'; 'r'; ' '; 't'; 'o'; ' '; 's'; 'w';
                                      'i'; 't'; 'c'; 'h'; ' '; 'f'; 'r'; 'o';
                                      'm'])
                              (comp shows_nl
                                (comp (shows_dpp (shows_laba _A2 _B2)
(shows_listb shows_chara) p r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l';
      'e'; 'm'])
                                      (comp shows_nl
(comp (shows_dpp (shows_laba _A2 _B2) (shows_listb shows_chara) rP r)
  (comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP r m rm rrm prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n';
                                    ' '; 'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r'; 'o';
                                    'c'; 'e'; 's'; 's'; 'o'; 'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, RedPairUrProc (redp, rP, ur, prf) ->
        debug (i [])
          ['R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o'; 'c'; ' '; '('; 'u';
            's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'; ' '; '+';
            ' '; 'a'; 'f'; ')']
          (bind (catch
                  (create_redpair_ur_af_proc
                    (eq_laba _A1 _B1, shows_laba _A2 _B2)
                    (eq_lista eq_char, shows_listb shows_chara) redp p rP r rm
                    ur)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 't'; 'h';
                                      'e'; ' '; 'r'; 'e'; 'd'; 'u'; 'c'; 't';
                                      'i'; 'o'; 'n'; ' '; 'p'; 'a'; 'i'; 'r';
                                      ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's';
                                      'o'; 'r'; ' '; '('; 'w'; 'i'; 't'; 'h';
                                      ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' ';
                                      'r'; 'u'; 'l'; 'e'; 's'; ' '; 'a'; 'n';
                                      'd'; ' '; 'a'; 'f'; ')'; ' '; 't'; 'o';
                                      ' '; 's'; 'w'; 'i'; 't'; 'c'; 'h'; ' ';
                                      'f'; 'r'; 'o'; 'm'; ' '])
                              (comp shows_nl
                                (comp (shows_dpp (shows_laba _A2 _B2)
(shows_listb shows_chara) p r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l';
      'e'; 'm'; ' '])
                                      (comp shows_nl
(comp (shows_dpp (shows_laba _A2 _B2) (shows_listb shows_chara) rP r)
  (comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP r m rm rrm prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n';
                                    ' '; 'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r'; 'o';
                                    'c'; 'e'; 's'; 's'; 'o'; 'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, DepGraphProc edpts ->
        debug (i [])
          ['D'; 'e'; 'p'; 'G'; 'r'; 'a'; 'p'; 'h'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (checkDepGraphProc (eq_laba _A1 _B1, shows_laba _A2 _B2)
                    (eq_lista eq_char, shows_listb shows_chara) p
                    (map (fun x ->
                           ((match fst x with None -> false | Some _ -> true),
                             snd x))
                      edpts)
                    rm rrm)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'i'; 'l'; 'e'; ' '; 't'; 'r';
                                      'y'; 'i'; 'n'; 'g'; ' '; 't'; 'o'; ' ';
                                      'p'; 'e'; 'r'; 'f'; 'o'; 'r'; 'm'; ' ';
                                      'S'; 'C'; 'C'; '-'; 'd'; 'e'; 'c'; 'o';
                                      'm'; 'p'; 'o'; 's'; 'i'; 't'; 'i'; 'o';
                                      'n'; ' '; ' '; 'o'; 'n'; ' '])
                              (comp shows_nl
                                (comp (shows_dpp (shows_laba _A2 _B2)
(shows_listb shows_chara) p r)
                                  (comp shows_nl x)))))))
            (fun _ ->
              catch (checkAllMIndex
                      (fun j rp ->
                        (match fst rp with None -> j | Some _ -> Suc j))
                      (fun j rp ->
                        (match fst rp with None -> return ()
                          | Some a ->
                            checkDPProof (_A1, _A2) (_B1, _B2)
                              (comp i (comp (shows_string ['.']) (shows_nat j)))
                              (snd rp) r m rm rrm a))
                      edpts)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'd'; 'e'; 'p'; 'e'; 'n'; 'd'; 'e'; 'n'; 'c';
                                    'y'; ' '; 'g'; 'r'; 'a'; 'p'; 'h'; ' '; 'p';
                                    'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, MonoRedPairProc (redp, rP, rR, prf) ->
        debug (i [])
          ['M'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o';
            'c']
          (bind (catch
                  (create_mono_redpair_proc
                    (eq_laba _A1 _B1, shows_laba _A2 _B2)
                    (eq_lista eq_char, shows_listb shows_chara) redp p rP r rR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 't'; 'h';
                                      'e'; ' '; 'm'; 'o'; 'n'; 'o'; 't'; 'o';
                                      'n'; 'i'; 'c'; ' '; 'r'; 'e'; 'd'; 'u';
                                      'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a';
                                      'i'; 'r'; ' '; 'p'; 'r'; 'o'; 'c'; 'e';
                                      's'; 's'; 'o'; 'r'; ' '; 't'; 'o'; ' ';
                                      's'; 'w'; 'i'; 't'; 'c'; 'h'; ' '; 'f';
                                      'r'; 'o'; 'm'; ' '])
                              (comp shows_nl
                                (comp (shows_dpp (shows_laba _A2 _B2)
(shows_listb shows_chara) p r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l';
      'e'; 'm'; ' '])
                                      (comp shows_nl
(comp (shows_dpp (shows_laba _A2 _B2) (shows_listb shows_chara) rP rR)
  (comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP rR m
                      (computeRuleMap (eq_laba _A1 _B1) rR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs rR)) prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c';
                                    ' '; 'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o';
                                    'n'; ' '; 'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r';
                                    'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, MonoRedPairUrProc (redp, rP, rR, ur, prf) ->
        debug (i [])
          ['M'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o';
            'c'; ' '; '('; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l';
            'e'; 's'; ')']
          (bind (catch
                  (create_mono_redpair_ur_proc
                    (eq_laba _A1 _B1, shows_laba _A2 _B2)
                    (eq_lista eq_char, shows_listb shows_chara) redp p rP r rm
                    rR ur)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 't'; 'h';
                                      'e'; ' '; 'm'; 'o'; 'n'; 'o'; 't'; 'o';
                                      'n'; 'i'; 'c'; ' '; 'r'; 'e'; 'd'; 'u';
                                      'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a';
                                      'i'; 'r'; ' '; 'p'; 'r'; 'o'; 'c'; 'e';
                                      's'; 's'; 'o'; 'r'; ' '; '('; 'w'; 'i';
                                      't'; 'h'; ' '; 'u'; 's'; 'a'; 'b'; 'l';
                                      'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'; ')';
                                      ' '; 't'; 'o'; ' '; 's'; 'w'; 'i'; 't';
                                      'c'; 'h'; ' '; 'f'; 'r'; 'o'; 'm'; ' '])
                              (comp shows_nl
                                (comp (shows_dpp (shows_laba _A2 _B2)
(shows_listb shows_chara) p r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l';
      'e'; 'm'])
                                      (comp shows_nl
(comp (shows_dpp (shows_laba _A2 _B2) (shows_listb shows_chara) rP rR)
  (comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP rR m
                      (computeRuleMap (eq_laba _A1 _B1) rR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs rR)) prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c';
                                    ' '; 'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o';
                                    'n'; ' '; 'p'; 'a'; 'i'; 'r'; ' '; '('; 'w';
                                    'i'; 't'; 'h'; ' '; 'u'; 's'; 'a'; 'b'; 'l';
                                    'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'; ')'; ' ';
                                    'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                                    'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, Suc m, rm, rrm, UnlabProc (ulP, ulR, prf) ->
        debug (i []) ['U'; 'n'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
          (bind (check_unlab_proc (_A1, _A2) (_B1, _B2)
                  (eq_lista eq_char, shows_listb shows_chara) p r ulP ulR)
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) ulP ulR m
                      (computeRuleMap (eq_laba _A1 _B1) ulR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs ulR)) prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                                    'g'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's';
                                    'o'; 'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, Zero_nat, rm, rrm, UnlabProc (ulP, ulR, prf) ->
        error (comp (shows_string
                      ['u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'; ' ';
                        'n'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w'; 'e'; 'd';
                        ' '; 'd'; 'u'; 'e'; ' '; 't'; 'o'; ' '; 'l'; 'a'; 'b';
                        'e'; 'l'; '-'; 'd'; 'e'; 'p'; 't'; 'h'; ' '; 'o'; 'f';
                        ' '; '0'])
                shows_nl)
    | i, p, r, m, rm, rrm, SemlabProc (sli, l, lP, lR, prf) ->
        debug (i []) ['S'; 'e'; 'm'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (create_sem_lab_proc (_A1, _A2) (_B1, _B2)
                    (eq_lista eq_char, shows_listb shows_chara) sli l p r lP lR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 't'; 'h';
                                      'e'; ' '; 's'; 'e'; 'm'; 'a'; 'n'; 't';
                                      'i'; 'c'; ' '; 'l'; 'a'; 'b'; 'e'; 'l';
                                      'i'; 'n'; 'g'; ' '; 'p'; 'r'; 'o'; 'c';
                                      'e'; 's'; 's'; 'o'; 'r'])
                              (comp shows_nl x)))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) lP lR (Suc m)
                      (computeRuleMap (eq_laba _A1 _B1) lR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs lR)) prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    's'; 'e'; 'm'; 'a'; 'n'; 't'; 'i'; 'c'; ' ';
                                    'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'; ' ';
                                    'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                                    'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, RootlabProc (lP, lR, prf) ->
        debug (i []) ['R'; 'o'; 'o'; 't'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
          (bind (catch (rl_proc (_A1, _A2) (_B1, _B2) p r lP lR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 't'; 'h';
                                      'e'; ' '; 'r'; 'o'; 'o'; 't'; ' '; 'l';
                                      'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'; ' ';
                                      'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                                      'r'])
                              (comp shows_nl x)))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) lP lR (Suc m)
                      (computeRuleMap (eq_laba _A1 _B1) lR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs lR)) prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'r'; 'o'; 'o'; 't'; '-'; 'l'; 'a'; 'b'; 'e';
                                    'l'; 'i'; 'n'; 'g'; ' '; 'p'; 'r'; 'o'; 'c';
                                    'e'; 's'; 's'; 'o'; 'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, SpscProc (sp, rP, prf) ->
        debug (i []) ['S'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (create_spsc_proc (eq_laba _A1 _B1, shows_laba _A2 _B2)
                    (eq_lista eq_char, shows_listb shows_chara) sp p rP r rm)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 't'; 'h';
                                      'e'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r';
                                      'm'; ' '; 'c'; 'r'; 'i'; 't'; 'e'; 'r';
                                      'i'; 'o'; 'n'; ' '; 'p'; 'r'; 'o'; 'c';
                                      'e'; 's'; 's'; 'o'; 'r'])
                              (comp shows_nl x)))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP r m rm rrm prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; ' '; 'c';
                                    'r'; 'i'; 't'; 'e'; 'r'; 'i'; 'o'; 'n'; ' ';
                                    'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                                    'r'])
                            (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, SizeChangeSubTerm graphs ->
        debug (i [])
          ['S'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'S'; 'u'; 'b'; 'T';
            'e'; 'r'; 'm']
          (check_sct_subterm_proc (eq_laba _A1 _B1, shows_laba _A2 _B2)
            (eq_lista eq_char, shows_listb shows_chara) p graphs rm rrm)
    | i, p, r, m, rm, rrm, SizeChangeRedPair (redp, ur, graphs) ->
        debug (i [])
          ['S'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'R'; 'e'; 'd'; 'P';
            'a'; 'i'; 'r']
          (create_sct_ur_af_proc (eq_laba _A1 _B1, shows_laba _A2 _B2)
            (eq_lista eq_char, shows_listb shows_chara) redp p rm rrm ur
            graphs);;

let rec checkTRSProof (_A1, _A2) (_B1, _B2)
  i r m x3 = match i, r, m, x3 with
    i, r, m, DPTrans (p, prf) ->
      debug (i []) ['D'; 'P'; ' '; 't'; 'r'; 'a'; 'n'; 's']
        (bind (catch
                (check_DP (eq_laba _A1 _B1, shows_laba _A2 _B2)
                  (eq_lista eq_char, shows_listb shows_chara) r p)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 's'; 'w'; 'i'; 't'; 'c';
                                    'h'; 'i'; 'n'; 'g'; ' '; 'f'; 'r'; 'o'; 'm';
                                    ' '; 't'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'])
                            (comp shows_nl
                              (comp (shows_trs (shows_laba _A2 _B2)
                                      (shows_listb shows_chara) r)
                                (comp shows_nl
                                  (comp (shows_string
  ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'i'; 'n'; 'i'; 't'; 'i'; 'a'; 'l'; ' ';
    'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'])
                                    (comp (shows_dpp
    (shows_shpa (shows_laba _A2 _B2)) (shows_listb shows_chara) p (iD_list r))
                                      (comp shows_nl x))))))))))
          (fun _ ->
            catch (checkDPProof
                    (eq_shpa (eq_laba _A1 _B1), shows_shpa (shows_laba _A2 _B2))
                    (_B1, _B2) (comp i (shows_string ['.'; '1']))
                    (map (map_rule init_lab) p)
                    (map (map_rule init_lab) (iD_list r)) Zero_nat
                    (computeRuleMap (eq_laba (eq_shpa (eq_laba _A1 _B1)) _B1)
                      (map (map_rule init_lab) (iD_list r)))
                    (computeRuleMap (eq_laba (eq_shpa (eq_laba _A1 _B1)) _B1)
                      (reverse_trs (map (map_rule init_lab) (iD_list r))))
                    prf)
              (fun x ->
                error (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 's'; 'w'; 'i'; 't';
                                  'c'; 'h'; ' '; 't'; 'o'; ' '; 'd'; 'e'; 'p';
                                  'e'; 'n'; 'd'; 'e'; 'n'; 'c'; 'y'; ' '; 'p';
                                  'a'; 'i'; 'r'; 's'])
                          (comp shows_nl (indent x)))))))
    | i, r, m, RuleRemoval (rp, rR, prf) ->
        debug (i [])
          ['r'; 'u'; 'l'; 'e'; ' '; 'r'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l']
          (bind (catch
                  (create_rule_removal_proc
                    (eq_laba _A1 _B1, shows_laba _A2 _B2)
                    (eq_lista eq_char, shows_listb shows_chara) rp r rR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'i'; 'n'; ' '; 'r'; 'u'; 'l'; 'e'; ' ';
                                      'r'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 's'; 'w'; 'i';
                                      't'; 'c'; 'h'; 'i'; 'n'; 'g'; ' '; 'f';
                                      'r'; 'o'; 'm'; ' '; 't'; 'h'; 'e'; ' ';
                                      'T'; 'R'; 'S'])
                              (comp shows_nl
                                (comp (shows_trs (shows_laba _A2 _B2)
(shows_listb shows_chara) r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'])
                                      (comp shows_nl
(comp (shows_trs (shows_laba _A2 _B2) (shows_listb shows_chara) rR)
  (comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rR m prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'r'; 'u'; 'l'; 'e'; ' '; 'r'; 'e'; 'm'; 'o';
                                    'v'; 'a'; 'l'; ' '; 'p'; 'r'; 'o'; 'c'; 'e';
                                    's'; 's'; 'o'; 'r'])
                            (comp shows_nl (indent x)))))))
    | i, r, Suc m, Unlab (uR, prf) ->
        debug (i []) ['u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l']
          (bind (catch
                  (unlab_trs_tt (_A1, _A2) (_B1, _B2)
                    (eq_lista eq_char, shows_listb shows_chara) r uR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'i'; 'n'; ' '; 'u'; 'n'; 'l'; 'a'; 'b';
                                      'e'; 'l'; 'i'; 'n'; 'g'; ' '; 'o'; 'f';
                                      ' '; 'T'; 'R'; 'S'; ' '; 'w'; 'h'; 'e';
                                      'n'; ' '; 's'; 'w'; 'i'; 't'; 'c'; 'h';
                                      'i'; 'n'; 'g'; ' '; 'f'; 'r'; 'o'; 'm';
                                      ' '; 't'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'])
                              (comp shows_nl
                                (comp (shows_trs (shows_laba _A2 _B2)
(shows_listb shows_chara) r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'])
                                      (comp shows_nl
(comp (shows_trs (shows_laba _A2 _B2) (shows_listb shows_chara) uR)
  (comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) uR m prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                    'u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                                    'g'])
                            (comp shows_nl (indent x)))))))
    | i, r, Zero_nat, Unlab (uR, prf) ->
        error (comp (shows_string
                      ['u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'; ' ';
                        'n'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w'; 'e'; 'd';
                        ' '; 'f'; 'o'; 'r'; ' '; 'l'; 'a'; 'b'; 'e'; 'l'; ' ';
                        'd'; 'e'; 'p'; 't'; 'h'; ' '; '0'])
                shows_nl)
    | i, r, m, RisEmpty ->
        debug (i []) ['R'; ' '; 'i'; 's'; ' '; 'e'; 'm'; 'p'; 't'; 'y']
          (checkB (null r)
            (comp i
              (comp (shows_string
                      [':'; ' '; 'R'; ' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' ';
                        'e'; 'm'; 'p'; 't'; 'y'; ','; ' '; 'b'; 'u'; 't'; ' ';
                        'i'; 't'; ' '; 'i'; 's'; ' '; 't'; 'h'; 'e'; ' '; 'T';
                        'R'; 'S'])
                (comp (shows_trs (shows_laba _A2 _B2) (shows_listb shows_chara)
                        r)
                  shows_nl))))
    | i, r, m, Semlab (sli, l, lR, prf) ->
        debug (i []) ['S'; 'e'; 'm'; 'l'; 'a'; 'b']
          (bind (catch
                  (create_sem_lab_tt (_A1, _A2) (_B1, _B2)
                    (eq_lista eq_char, shows_listb shows_chara) sli l r lR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 's'; 'e';
                                      'm'; 'a'; 'n'; 't'; 'i'; 'c'; ' '; 'l';
                                      'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'])
                              (comp shows_nl x)))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) lR (Suc m) prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 's'; 'e'; 'm'; 'a';
                                    'n'; 't'; 'i'; 'c'; ' '; 'l'; 'a'; 'b'; 'e';
                                    'l'; 'i'; 'n'; 'g'])
                            (comp shows_nl (indent x)))))))
    | i, r, m, Rootlab (lR, prf) ->
        debug (i []) ['R'; 'o'; 'o'; 't'; 'l'; 'a'; 'b']
          (bind (catch (rl_tt (_A1, _A2) (_B1, _B2) r lR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 'r'; 'o';
                                      'o'; 't'; '-'; 'l'; 'a'; 'b'; 'e'; 'l';
                                      'i'; 'n'; 'g'])
                              (comp shows_nl x)))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) lR (Suc m) prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 'r'; 'o'; 'o'; 't';
                                    '-'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                                    'g'])
                            (comp shows_nl (indent x)))))))
    | i, r, m, Fcc (fcs, rR, prf) ->
        debug (i []) ['F'; 'c'; 'c']
          (bind (catch (fcc_tt (eq_laba _A1 _B1, shows_laba _A2 _B2) r fcs rR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 'f'; 'l';
                                      'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e';
                                      'x'; 't'; ' '; 'c'; 'l'; 'o'; 's'; 'u';
                                      'r'; 'e'])
                              (comp shows_nl x)))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rR Zero_nat prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 'f'; 'l'; 'a'; 't';
                                    ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; ' ';
                                    'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'])
                            (comp shows_nl (indent x)))))))
    | i, r, m, StringReversal (rR, prf) ->
        debug (i [])
          ['S'; 't'; 'r'; 'i'; 'n'; 'g'; ' '; 'R'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
            'l']
          (bind (catch
                  (check_string_rev (eq_laba _A1 _B1, shows_laba _A2 _B2)
                    (eq_lista eq_char, shows_listb shows_chara) r rR)
                  (fun x ->
                    error (comp i
                            (comp (shows_string
                                    [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p';
                                      'l'; 'y'; 'i'; 'n'; 'g'; ' '; 's'; 't';
                                      'r'; 'i'; 'n'; 'g'; ' '; 'r'; 'e'; 'v';
                                      'e'; 'r'; 's'; 'a'; 'l'])
                              (comp shows_nl x)))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rR Zero_nat prf)
                (fun x ->
                  error (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                    'e'; 'l'; 'o'; 'w'; ' '; 's'; 't'; 'r'; 'i';
                                    'n'; 'g'; ' '; 'r'; 'e'; 'v'; 'e'; 'r'; 's';
                                    'a'; 'l'])
                            (comp shows_nl (indent x)))))));;

let rec certifyCertProblem (_A1, _A2) (_B1, _B2)
  = function
    CertificationProblem (true, trs, Left prf) ->
      (match
        checkTRSProof (_A1, _A2) (_B1, _B2) (shows_string ['1'])
          (map (map_rule init_lab) trs) Zero_nat prf
        with Left err ->
          Error (comp (shows_string
                        ['t'; 'e'; 'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n';
                          ' '; 'p'; 'r'; 'o'; 'o'; 'f'; ' '; 'n'; 'o'; 't'; ' ';
                          'a'; 'c'; 'c'; 'e'; 'p'; 't'; 'e'; 'd'])
                  (comp shows_nl err) [])
        | Right _ -> Certified true)
    | CertificationProblem (false, trs, Right prf) ->
        (match
          checkTRSDisproof (_A1, _A2)
            (eq_lista eq_char, shows_listb shows_chara) (shows_string ['1']) trs
            prf
          with Left err ->
            Error (comp (shows_string
                          ['n'; 'o'; 'n'; '-'; 't'; 'e'; 'r'; 'm'; 'i'; 'n';
                            'a'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'r'; 'o'; 'o';
                            'f'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'c'; 'c'; 'e';
                            'p'; 't'; 'e'; 'd'])
                    (comp shows_nl err) [])
          | Right _ -> Certified false)
    | CertificationProblem (false, va, Left v) ->
        Error ['N'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w'; 'e'; 'd'; ' '; 't';
                'o'; ' '; 'g'; 'i'; 'v'; 'e'; ' '; 'a'; ' '; 't'; 'e'; 'r'; 'm';
                'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'r'; 'o'; 'o'; 'f';
                ' '; 'f'; 'o'; 'r'; ' '; 'a'; ' '; 'n'; 'o'; 'n'; '-'; 't'; 'e';
                'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n'; ' '; 'g'; 'o'; 'a';
                'l'; ' '; 'o'; 'r'; ' '; 'v'; 'i'; 'c'; 'e'; '-'; 'v'; 'e'; 'r';
                's'; 'a']
    | CertificationProblem (true, va, Right vc) ->
        Error ['N'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w'; 'e'; 'd'; ' '; 't';
                'o'; ' '; 'g'; 'i'; 'v'; 'e'; ' '; 'a'; ' '; 't'; 'e'; 'r'; 'm';
                'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'r'; 'o'; 'o'; 'f';
                ' '; 'f'; 'o'; 'r'; ' '; 'a'; ' '; 'n'; 'o'; 'n'; '-'; 't'; 'e';
                'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n'; ' '; 'g'; 'o'; 'a';
                'l'; ' '; 'o'; 'r'; ' '; 'v'; 'i'; 'c'; 'e'; '-'; 'v'; 'e'; 'r';
                's'; 'a'];;

let rec xml_certificationProblem _A _C
  (CertificationProblem (term, rls, prf)) =
    let inp =
      Xml (['i'; 'n'; 'p'; 'u'; 't'], [],
            [Xml (['t'; 'r'; 's'; 'I'; 'n'; 'p'; 'u'; 't'], [],
                   [Xml (['t'; 'r'; 's'], [], [xml_rules _A _C rls], None)],
                   None)],
            None)
      in
    let goal = (if isLeft prf then terminationGoal else nonTerminationGoal) in
    Xml (['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'], [], [goal; inp], None);;

let rec eqWhiteSpace
  s1 s2 =
    debug ['0']
      ['c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g'; ' '; 'w'; 'h'; 'e'; 't';
        'h'; 'e'; 'r'; ' '; 'p'; 'a'; 'r'; 's'; 'e'; 'd'; ' '; 'i'; 'n'; 'p';
        'u'; 't'; ' '; 'c'; 'o'; 'r'; 'r'; 'e'; 's'; 'p'; 'o'; 'n'; 'd'; 's';
        ' '; 't'; 'o'; ' '; 'r'; 'e'; 'a'; 'l'; ' '; 'i'; 'n'; 'p'; 'u'; 't']
      (eq_list eq_char (filter (fun c -> not (member eq_char c wspace)) s1)
        (filter (fun c -> not (member eq_char c wspace)) s2));;

let rec certifyProof
  problemString =
    (match parseCertProblem problemString
      with Left err ->
        Error (append
                ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'i'; 'l'; 'e'; ' ';
                  'p'; 'a'; 'r'; 's'; 'i'; 'n'; 'g']
                (append [newline] err))
      | Right cp ->
        (match extract_goal_input problemString
          with None ->
            Error ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'i';
                    'n'; 'd'; ' '; 'g'; 'o'; 'a'; 'l'; ' '; 'a'; 'n'; 'd'; ' ';
                    'i'; 'n'; 'p'; 'u'; 't'; ' '; 'i'; 'n'; ' '; 's'; 't'; 'r';
                    'i'; 'n'; 'g'; ' '; 'f'; 'o'; 'r'; ' '; 'c'; 'o'; 'm'; 'p';
                    'a'; 'r'; 'i'; 's'; 'o'; 'n'; ' '; 'w'; 'i'; 't'; 'h'; ' ';
                    'p'; 'a'; 'r'; 's'; 'e'; 'd'; ' '; 'r'; 'e'; 's'; 'u'; 'l';
                    't']
          | Some inpProblem ->
            (if eqWhiteSpace inpProblem
                  (show shows_xmla
                    (xml_certificationProblem (shows_listb shows_chara)
                      (shows_listb shows_chara) cp))
              then certifyCertProblem
                     (eq_lista eq_char, shows_listb shows_chara)
                     (eq_lista eq_nata, shows_listb shows_nata) cp
              else Error (concat
                           [['p'; 'a'; 'r'; 's'; 'e'; 'd'; ' '; 'p'; 'r'; 'o';
                              'b'; 'l'; 'e'; 'm'; ' '; 'd'; 'o'; 'e'; 's'; ' ';
                              'n'; 'o'; 't'; ' '; 'c'; 'o'; 'r'; 'r'; 'e'; 's';
                              'p'; 'o'; 'n'; 'd'; ' '; 't'; 'o'; ' '; 'i'; 'n';
                              'p'; 'u'; 't'];
                             [newline]; ['i'; 'n'; 'p'; 'u'; 't'; ':'];
                             [newline];
                             filter (fun c -> not (member eq_char c wspace))
                               inpProblem;
                             [newline]; [newline];
                             ['p'; 'a'; 'r'; 's'; 'e'; 'd'; ':']; [newline];
                             filter (fun c -> not (member eq_char c wspace))
                               (show shows_xmla
                                 (xml_certificationProblem
                                   (shows_listb shows_chara)
                                   (shows_listb shows_chara) cp));
                             [newline]]))));;

end;; (*struct Ceta*)
